/* -*-c++-*- */
#ifndef BACKPOINTER_H
#define BACKPOINTER_H
/* Backpointer : A backpointer object capable of holding an index and
   label.  This object can also be useful for representing a
   half-dependency; e.g., for pruning. */

#include "Macros.h"
#include "Bits.h"

#include <math.h>


class backp {
public:
  /* construct a backpointer holding a word and label */
  explicit backp(const idx_t w, const lab_t l);
  /* construct a backpointer holding a single item */
  explicit backp(const uint16_t w);
  /* construct an uninitialized backpointer */
  backp();

  /* obtain the index and label */
  idx_t i() const;
  lab_t l() const;
  /* obtain the index, accounting for the possibility of null */
  int i_null() const;
  /* obtain the raw value */
  uint16_t raw() const;

  /* for insertion into map */
  bool operator<(const backp& that) const;
  /* for hashtable value removal */
  bool operator==(const backp& that) const;

private:
  uint16_t _bits;
};




/*** inlined implementation ***/

/* lifecycle */
inline backp::backp(const idx_t ww, const lab_t ll)
  : _bits((uint16_t)(ww | (((uint16_t)ll) << IBITS))) {}
inline backp::backp(const uint16_t r) : _bits(r) {}
inline backp::backp()                 : _bits((uint16_t)0xffffU) {}

/* accessors */
inline idx_t backp::i() const      { return (idx_t)(_bits &  IMASK); }
inline lab_t backp::l() const      { return (lab_t)(_bits >> IBITS); }
inline uint16_t backp::raw() const { return _bits; }
inline int backp::i_null() const   {
  const idx_t ii = i();
  return (ii == INULL ? -1 : ii);
}

/* for usage in map<>.  NB: the precedence of i() and l() is reversed
   w.r.t. their placement in the bitstring.  this is intentional. */
inline bool backp::operator<(const backp& that) const {
  return (i() < that.i()) || ((i() == that.i()) && l() < that.l());
}
/* for usage in hashtable */
inline bool backp::operator==(const backp& that) const {
  return (_bits == that._bits);
}

#endif /* BACKPOINTER_H */
