/* -*-c++-*- */
#ifndef BITS_H
#define BITS_H

/* first include, so that __STDC_CONSTANT_MACROS takes effect */
#define __STDC_CONSTANT_MACROS
#include <stdint.h>
#undef __STDC_CONSTANT_MACROS

#include <assert.h>


/* an integral type for a feature bitstring */
typedef uint64_t fbits;



/* word and tag identities */
typedef uint16_t word_t;
typedef uint8_t  tag_t;

/* bit-lengths */
#define WBITS 16U
#define TBITS 8U

/* bitmasks */
#define WMASK ((word_t)0xffffU)
#define TMASK ((tag_t)0xffU)

/* limits of normal values: 0 <= v < MAX */
#define WMAX WMASK
#define TMAX TMASK

/* null values: equivalent to -1/sentence-exterior */
#define WNULL WMASK
#define TNULL TMASK

/* conversions with possible checks */
#ifdef CAUTIOUS
#define WCONV(w) (assert((w) >= 0), assert((w) < WMAX), ((word_t)(w)))
#define TCONV(t) (assert((t) >= 0), assert((t) < TMAX), ((tag_t)(t)))
#else /* CAUTIOUS */
#define WCONV(w) ((word_t)(w))
#define TCONV(t) ((tag_t)(t))
#endif /* CAUTIOUS */




/* sentence index and label index */
typedef uint8_t idx_t;
typedef uint8_t lab_t;

/* bit-lengths */
#define IBITS ((idx_t)8U)
#define LBITS ((lab_t)8U)

/* bitmasks */
#define IMASK ((idx_t)0xffU)
#define LMASK ((lab_t)0xffU)

/* limits of normal values: 0 <= v < MAX */
#define IMAX ((idx_t)0xfeU)
#define LMAX ((lab_t)0xfeU)

/* null values: equivalent to -1/root */
#define INULL IMAX
#define LNULL LMAX

/* invalid values: never used, not even for null/root indices */
#define IINV IMASK
#define LINV LMASK

/* conversions with possible checks */
#ifdef CAUTIOUS
#define ICONV(i) (assert((i) >= 0), assert((i) < IMAX), ((idx_t)(i)))
#define LCONV(l) (assert((l) >= 0), assert((l) < LMAX), ((lab_t)(l)))
#else /* CAUTIOUS */
#define ICONV(i) ((idx_t)(i))
#define LCONV(l) ((lab_t)(l))
#endif /* CAUTIOUS */
#define ICONV_NULL(i)   ((i) == -1 ? INULL : ICONV(i))
#define IUNCONV_NULL(i) ((i) == INULL ? -1 : (i))
#define LUNCONV_NULL(l) ((l) == LNULL ? -1 : (l))

#endif /* BITS_H */
