/* -*-c++-*- */
#ifndef CORPUS_H
#define CORPUS_H
/* Corpus : A dataset of examples with facilities for reading examples
   from a file. */

#include "Macros.h"

#include <stdio.h>

#include <vector>
#include <string>

using std::vector;
using std::string;

/* forward declarations */
class ex;
class active_deps;

/* a corpus of examples */
class corpus {
public:
  ~corpus();

  /* create a new corpus from the named file */
  static corpus* load(const string& fname, const int L);

  /* get the number of examples in this */
  int N() const;
  /* get a short string describing the data ordering */
  string ostr() const;
  /* get a long string describing the data ordering */
  string ostr_long() const;

  /* set the style of ordering: "det", "unif", "perm" */
  void ordering(const string& type);
  /* signal the start of an enumeration */
  void start();
  /* obtain the next example in the enumeration, or NULL if the
     enumeration is complete (N examples have been received) */
  const ex* next();

  /* re-label the corpus.  for each example, find a tree with the
     maximal number of correct dependencies.  in normal parsing, this
     is not interesting as it simply recovers the original labels.
     however, if the parser is pruned or applied to non-projective
     data, then it is often desirable to relabel the corpus before
     training.  otherwise, the parser may be unable to recover the
     correct labels (due to non-projectivity or one of the correct
     dependencies being pruned away), causing spurious errors during
     training.  this relabeling should only be done to the training
     data.  if A is NULL then all dependencies will be considered,
     otherwise the pruned dependencies in A are used.  if fout is
     non-NULL, then the new labels are written to fout for the
     purposes of repeatability. */
  void relabel(const int L, const bool multiroot,
               const active_deps* const restrict A = NULL,
               FILE* const fout = NULL);

private:
  vector<ex*> _data; /* list of training examples */
  size_t _i;         /* position in enumeration */
  int _order;        /* ordering of examples */

  corpus();
};


/* inlined methods */
inline int corpus::N() const { return (int)_data.size(); }

#endif /* CORPUS_H */
