#include "Example.h"

#include "Bits.h"

#include <assert.h>
#include <ctype.h>
#include <string.h>

#include <iostream>

using std::cerr;
using std::endl;


/* buffer size for reading lines from files */
#define LINELEN 65536

/* lifecycle */
ex::ex() : words(), tags(), heads(), labels(), _id(-1) {}
ex::~ex() {}


/* take the next identifier and advance */
void ex::set_id(const int newid) { assert(_id == -1); _id = newid; }


/* read an example from the given file */
ex* ex::fread(FILE* const restrict f, const int L) {
  ex* const restrict e = new ex(); /* the example to create */
  assert(e != NULL);

  char line[LINELEN];        /* line-reading buffer */
  char buf1[256], buf2[256]; /* string column buffers */
  while(1) {
    line[0] = '\0';
    fgets(line, LINELEN, f);
    const size_t len = strlen(line);
    if(len == 0) { break; }
    { /* check for whitespace-only lines */
      bool allspace = true;
      for(size_t i = 0; i < len; ++i) {
        if(!isspace(line[i])) { allspace = false; break; }
      }
      if(allspace) { break; }
    }
    if(line[len - 1] != '\n') {
      cerr << "detected line overflow when reading data" << endl;
      assert(0);
    }

    /* split along whitespace and create a token */
    int m = -2, w = -2, ct = -2, ft = -2, hh = -2, ll = -2;
    int nsc = sscanf(line, " %d %d %255s %d %d %255s %d %d",
                     &m, &w, buf1, &ct, &ft, buf2, &hh, &ll);
    assert(nsc == 8);
    assert(w >= 0); assert(w < WMAX);
    assert(ft >= 0); assert(ft < TMAX);
    e->words.push_back(w);
    e->tags.push_back(ft);  /* NB: fine tag */

    /* read the parent dependency */
    const int h = hh - 1; /* NB: 1-origin to 0-origin */
    assert(h >= -1);
    const int l = (L == 1 ? 0 : ll); /* NB: unlabeled parsing */
    assert(l >= 0); assert(l < L);
    e->heads.push_back(h);
    e->labels.push_back(l);
  }

  assert(e->heads.size() == e->words.size());
  assert(e->heads.size() == e->heads.size());
  assert(e->heads.size() == e->labels.size());

  const int N = (int)e->heads.size();
  for(int i = 0; i < N; ++i) { assert(e->heads[i] <= N); }

  /* empty sentence: no more sentences left */
  if(e->heads.empty()) { delete e; return NULL; }

  return e;
}


/* write an example to the given file with the alternate label
   provided */
void ex::fwrite(FILE* const restrict f,
                const vector<int>& restrict alt_heads,
                const vector<int>& restrict alt_labels) const {
  const int N = (int)alt_heads.size();
  assert(N > 0);
  assert(alt_heads.size() == alt_labels.size());
  assert(heads.size() == alt_heads.size());
  for(int i = 0; i < N; ++i) {
    fprintf(f, "%d %d %d %d\n", _id, i + 1,
            alt_heads[i] + 1, alt_labels[i]);
  }
  fprintf(f, "\n");
}
