/* -*-c++-*- */
#ifndef EXAMPLE_H
#define EXAMPLE_H
/* Example : A training example for parsing, consisting of a sentence,
   represented as lists of words and tags, and a tree, represented as
   lists of heads and labels. */

#include "Macros.h"

#include <stdio.h>

#include <vector>

using std::vector;

/* forward declaration */
class corpus;

class ex {
public:
  /* public access to data; it's okay because we'll only ever be using
     const ex* pointers */
  vector<int> words;       /* words of the sentence */
  vector<int> tags;        /* POS tags of the input sentence */
  vector<int> heads;       /* the correct heads */
  vector<int> labels;      /* the correct labels */

  int id() const;     /* example-index of this */

  /* set the example-id; this can only be done once */
  void set_id(const int newid);

  /* write a labeled example to the given file */
  void fwrite(FILE* const restrict f,
              const vector<int>& restrict heads,
              const vector<int>& restrict labels) const;

private:
  int _id;

  /* can only be created or destroyed by the corpus */
  friend class corpus;
  ex();
  ~ex();

  /* read a training example from the given file, returning NULL if no
     more examples are available.  this is the only method for
     constructing sensible examples. */
  static ex* fread(FILE* const restrict f, const int L);
};


/* inlined methods */
inline int ex::id() const { return _id; }

#endif /* EXAMPLE_H */
