/* -*-c++-*- */
#ifndef FEATFUNCTORS_H
#define FEATFUNCTORS_H

#include "O3ParamsAvg.h"
#include "O3ParamsLin.h"

/* feature-handler classes */
class featf_avg_dot {
public:
  typedef const o3params_avg params;
  typedef const params::primals primals;
  /* return the product of the params and an indicator */
  double operator()(const fbits b) const {
    return _W.get(b, avg_param()).val();
  }

  featf_avg_dot(const params& restrict V, const primals& restrict W,
                const double alpha) : _W(W) {}

private:
  const primals& restrict _W;
};


class featf_avg_adot {
public:
  typedef const o3params_avg params;
  typedef const params::primals primals;
  /* return the product with the summed parameter */
  double operator()(const fbits b) const {
    return _W.get(b, avg_param()).sum();
  }

  featf_avg_adot(const params& restrict V, const primals& restrict W,
                 const double alpha) : _W(W) {}

private:
  const primals& restrict _W;
};


class featf_avg_add {
public:
  typedef o3params_avg params;
  typedef params::primals primals;
  /* add _alpha to the parameters */
  double operator()(const fbits b) {
    /* insert a zeroed avg_param if the key does not exist */
    _W.mget(b, avg_param()).add(_now, _alpha);
    return 0; /* return a dummy value */
  }

  featf_avg_add(params& restrict V, primals& restrict W,
                const double alpha)
    : _W(W), _alpha(alpha), _now(V.now()) {}

private:
  primals& restrict _W;
  const double _alpha; /* the adjustment to make to each feature */
  const uint64_t _now; /* the current timestamp value */
};


class featf_lin_dot {
public:
  typedef const o3params_lin params;
  typedef const params::primals primals;
  /* return the product of the Params and an indicator */
  double operator()(const fbits b) const {
    /* FIXME: scale with scaling factor?  no point for linear argmax,
       but for log-linear it makes a difference */
    return _W.get(b, 0);
  }

  featf_lin_dot(const params& restrict V, const primals& restrict W,
                const double alpha) : _W(W) {}

private:
  const primals& restrict _W;
};

class featf_lin_add {
public:
  typedef o3params_lin params;
  typedef params::primals primals;
  /* add _alpha to the parameters */
  double operator()(const fbits b) {
    double& restrict d = _W.mget(b, 0);
    /* scale_add computes scaled adjustment */
    d += params::scale_add(_alpha, _inv_scale);
    return 0; /* return a dummy value */
  }

  featf_lin_add(params& restrict V, primals& restrict W,
                const double alpha)
    : _W(W), _alpha(alpha), _inv_scale(V._inv_scale) {}

private:
  primals& restrict _W;
  const double _alpha;     /* the adjustment to make to each feature */
  const double _inv_scale; /* the inverse scaling factor */
};

#endif /* FEATFUNCTORS_H */
