/* -*-c++-*- */
#ifndef FEATTYPES_H
#define FEATTYPES_H

struct fbi {
  enum types {
    /* dependency feature types */
    HWMW = 0,   /* word bigram */

    HTMT,       /* tag bigram */
    HWMT,       /* word->tag */
    HTMW,       /* tag->word */
    HBMT,       /* tag bigram + head word */
    HTMB,       /* tag bigram + mod word */
    HBMW,       /* word bigram + head tag */
    HWMB,       /* word bigram + mod tag */
    HBMB,       /* word and tag bigram */

    HQMQ,       /* ctag bigram */
    HWMQ,       /* word->ctag */
    HQMW,       /* ctag->word */
    HRMQ,       /* ctag bigram + head word */
    HQMR,       /* ctag bigram + mod word */
    HRMW,       /* word bigram + head ctag */
    HWMR,       /* word bigram + mod ctag */
    HRMR,       /* word and ctag bigram */

    HWMW_HNW,   /* word bigram + next word of head */
    HWMW_HPW,   /* word bigram + prev word of head */
    HWMW_MNW,   /* word bigram + next word of mod */
    HWMW_MPW,   /* word bigram + prev word of mod */

    HBMB_HNB,   /* word+tag bigram + next word+tag of head */
    HBMB_HPB,   /* word+tag bigram + prev word+tag of head */
    HBMB_MNB,   /* word+tag bigram + next word+tag of mod */
    HBMB_MPB,   /* word+tag bigram + prev word+tag of mod */

    HRMR_HNR,   /* word+ctag bigram + next word+ctag of head */
    HRMR_HPR,   /* word+ctag bigram + prev word+ctag of head */
    HRMR_MNR,   /* word+ctag bigram + next word+ctag of mod */
    HRMR_MPR,   /* word+ctag bigram + prev word+ctag of mod */

    HTMT_HNT,   /* tag bigram + next tag of head */
    HTMT_HPT,   /* tag bigram + prev tag of head */
    HTMT_MNT,   /* tag bigram + next tag of mod */
    HTMT_MPT,   /* tag bigram + prev tag of mod */
    HWMW_HNT,   /* word bigram + next tag of head */
    HWMW_HPT,   /* word bigram + prev tag of head */
    HWMW_MNT,   /* word bigram + next tag of mod */
    HWMW_MPT,   /* word bigram + prev tag of mod */
    HBMB_HNT,   /* word+tag bigram + next tag of head */
    HBMB_HPT,   /* word+tag bigram + prev tag of head */
    HBMB_MNT,   /* word+tag bigram + next tag of mod */
    HBMB_MPT,   /* word+tag bigram + prev tag of mod */

    HTMT_NNTT,  /* tag 4-gram, next of head/mod */
    HTMT_NPTT,  /* tag 4-gram, mixed */
    HTMT_PNTT,  /* tag 4-gram, mixed */
    HTMT_PPTT,  /* tag 4-gram, prev of head/mod */
    HWMW_NNTT,  /* word bigram + next tag of head/mod */
    HWMW_NPTT,  /* word bigram + mixed */
    HWMW_PNTT,  /* word bigram + mixed */
    HWMW_PPTT,  /* word bigram + prev tag of head/mod */
    HBMB_NNTT,  /* word+tag bigram + next tag of head/mod */
    HBMB_NPTT,  /* word+tag bigram + mixed */
    HBMB_PNTT,  /* word+tag bigram + mixed */
    HBMB_PPTT,  /* word+tag bigram + prev tag of head/mod */

    HQMQ_HNQ,   /* ctag bigram + next ctag of head */
    HQMQ_HPQ,   /* ctag bigram + prev ctag of head */
    HQMQ_MNQ,   /* ctag bigram + next ctag of mod */
    HQMQ_MPQ,   /* ctag bigram + prev ctag of mod */
    HWMW_HNQ,   /* word bigram + next ctag of head */
    HWMW_HPQ,   /* word bigram + prev ctag of head */
    HWMW_MNQ,   /* word bigram + next ctag of mod */
    HWMW_MPQ,   /* word bigram + prev ctag of mod */
    HRMR_HNQ,   /* word+ctag bigram + next ctag of head */
    HRMR_HPQ,   /* word+ctag bigram + prev ctag of head */
    HRMR_MNQ,   /* word+ctag bigram + next ctag of mod */
    HRMR_MPQ,   /* word+ctag bigram + prev ctag of mod */

    HQMQ_NNQQ,  /* ctag 4-gram, next of head/mod */
    HQMQ_NPQQ,  /* ctag 4-gram, mixed */
    HQMQ_PNQQ,  /* ctag 4-gram, mixed */
    HQMQ_PPQQ,  /* ctag 4-gram, prev of head/mod */
    HWMW_NNQQ,  /* word bigram + next ctag of head/mod */
    HWMW_NPQQ,  /* word bigram + mixed */
    HWMW_PNQQ,  /* word bigram + mixed */
    HWMW_PPQQ,  /* word bigram + prev ctag of head/mod */
    HRMR_NNQQ,  /* word+ctag bigram + next ctag of head/mod */
    HRMR_NPQQ,  /* word+ctag bigram + mixed */
    HRMR_PNQQ,  /* word+ctag bigram + mixed */
    HRMR_PPQQ,  /* word+ctag bigram + prev ctag of head/mod */

    HTMT_HNW,   /* tag bigram + next word of head */
    HTMT_HPW,   /* tag bigram + prev word of head */
    HTMT_MNW,   /* tag bigram + next word of mod */
    HTMT_MPW,   /* tag bigram + prev word of mod */
    HTMT_NNWW,  /* tag 4-gram, next of head/mod */
    HTMT_NPWW,  /* tag 4-gram, mixed */
    HTMT_PNWW,  /* tag 4-gram, mixed */
    HTMT_PPWW,  /* tag 4-gram, prev of head/mod */

    HQMQ_HNW,   /* ctag bigram + next word of head */
    HQMQ_HPW,   /* ctag bigram + prev word of head */
    HQMQ_MNW,   /* ctag bigram + next word of mod */
    HQMQ_MPW,   /* ctag bigram + prev word of mod */
    HQMQ_NNWW,  /* ctag 4-gram, next of head/mod */
    HQMQ_NPWW,  /* ctag 4-gram, mixed */
    HQMQ_PNWW,  /* ctag 4-gram, mixed */
    HQMQ_PPWW,  /* ctag 4-gram, prev of head/mod */

    DISTEXACT,  /* distance of dependency, exact value */
    DBIN,       /* distance of dependency, binned */
    DBIN_HT,    /* distance bin plus head tag */
    DBIN_HQ,    /* distance bin plus head ctag */
    DBIN_MT,    /* distance bin plus mod tag */
    DBIN_MQ,    /* distance bin plus mod ctag */
    DBIN_HWMW,  /* distance bin plus word bigram */
    DBIN_HTMT,  /* distance bin plus tag bigram */
    DBIN_HQMQ,  /* distance bin plus ctag bigram */

    BT,         /* between tag */
    BQ,         /* between ctag */
    BW,         /* between word */
    HTMT_BT,    /* tag bigram with between tag */
    HQMQ_BQ,    /* ctag bigram with between ctag */
    HTMT_BW,    /* tag bigram with between word */
    HQMQ_BW,    /* ctag bigram with between word */
    HWMW_BT,    /* word bigram with between tag */
    HWMW_BQ,    /* word bigram with between ctag */
    HWMW_BW,    /* word bigram with between word */

    BCC,        /* between-#coords */
    BIN,        /* between-#preps */
    BPU,        /* between-#puncs */
    BVB,        /* between-#verbs */
    HTMT_BCC,   /* tag bigram with #coords between */
    HQMQ_BCC,   /* ctag bigram with #coords between */
    HWMW_BCC,   /* word bigram with #coords between */
    HTMT_BIN,   /* tag bigram with #preps between */
    HQMQ_BIN,   /* ctag bigram with #preps between */
    HWMW_BIN,   /* word bigram with #preps between */
    HTMT_BPU,   /* tag bigram with #puncs between */
    HQMQ_BPU,   /* ctag bigram with #puncs between */
    HWMW_BPU,   /* word bigram with #puncs between */
    HTMT_BVB,   /* tag bigram with #verbs between */
    HQMQ_BVB,   /* ctag bigram with #verbs between */
    HWMW_BVB,   /* word bigram with #verbs between */

    QUOTELEV,
    QUOTELEV_HTMT,
    QUOTELEV_HQMQ,
    PARENLEV,
    PARENLEV_HTMT,
    PARENLEV_HQMQ,
    QUOTELEV_CROSS,
    QUOTELEV_CROSS_HTMT,
    QUOTELEV_CROSS_HQMQ,
    PARENLEV_CROSS,
    PARENLEV_CROSS_HTMT,
    PARENLEV_CROSS_HQMQ,

    /* token features */
    ROOT,       /* root token */

    TW0,        /* current word */
    TWP1,       /* previous word */
    TWP2,       /* 2-previous word */
    TWN1,       /* next word */
    TWN2,       /* 2-next word */

    TT0,        /* current tag and context tags */
    TTP1,
    TTP2,
    TTN1,
    TTN2,

    TQ0,        /* current ctag and context */
    TQP1,
    TQP2,
    TQN1,
    TQN2,

    TB0,        /* word+tag context */
    TBP1,
    TBP2,
    TBN1,
    TBN2,

    TR0,        /* word+ctag context */
    TRP1,
    TRP2,
    TRN1,
    TRN2,

    TWP01,      /* forward/backward word bigram/trigram */
    TWN01,
    TWP012,
    TWN012,

    TTP01,      /* tag n-grams */
    TTN01,
    TTP012,
    TTN012,

    TQP01,      /* ctag n-grams */
    TQN01,
    TQP012,
    TQN012,

    TBP01,      /* word+tag n-grams */
    TBN01,
    TBP012,
    TBN012,

    TRP01,      /* word+ctag n-grams */
    TRN01,
    TRP012,
    TRN012,

    /* auxiliary dependency features */
    AUXD,

    /* end (count) token */
    END
  };
};


struct ftri {
  enum types {
    /* all combinations of tag/word for h,m,c */
    WWW = 0,

    TTT,
    QQQ,

    BBB,
    RRR,

    BTT,
    TBT,
    TTB,
    RQQ,
    QRQ,
    QQR,

    /* trigram plus context */
    TTT_HNT,
    TTT_HPT,
    TTT_MNT,
    TTT_MPT,
    TTT_NNTT,
    TTT_NPTT,
    TTT_PNTT,
    TTT_PPTT,
    WWW_HNT,
    WWW_HPT,
    WWW_MNT,
    WWW_MPT,

    QQQ_HNQ,
    QQQ_HPQ,
    QQQ_MNQ,
    QQQ_MPQ,
    QQQ_NNQQ,
    QQQ_NPQQ,
    QQQ_PNQQ,
    QQQ_PPQQ,
    WWW_HNQ,
    WWW_HPQ,
    WWW_MNQ,
    WWW_MPQ,

    /* backed-off bigrams */
    HWMW, /* head+mod */
    HTMT,
    HQMQ,
    HBMB,
    HRMR,

    HWCW, /* head+child */
    HTCT,
    HTCW,
    HWCT,
    HQCQ,
    HQCW,
    HWCQ,
    HBCB,
    HRCR,

    MWCW, /* mod+child */
    MTCT,
    MTCW,
    MWCT,
    MQCQ,
    MQCW,
    MWCQ,
    MBCB,
    MRCR,

    /* backed-off unigrams */
    HW, /* head */
    HT,
    HQ,
    HB,
    HR,

    MW, /* mod */
    MT,
    MQ,
    MB,
    MR,

    CW, /* child */
    CT,
    CQ,
    CB,
    CR,

    /* always-1, to allow bias on labels themselves */
    BIAS,

    /* end (count) token */
    END
  };
};


struct fquad {
  enum types {
    /* simple word/tag combinations */
    TTTT = 0,

    WTTT,
    TWTT,
    TTWT,
    TTTW,

    TTWW,
    TWTW,
    TWWT,
    WWTT,
    WTWT,
    WTTW,


    QQQQ,

    QQWW,
    QWQW,
    QWWQ,
    WWQQ,
    WQWQ,
    WQQW,

    WQQQ,
    QWQQ,
    QQWQ,
    QQQW,


    SCC_WWW,

    SCC_TTT,
    SCC_WTT,
    SCC_TWT,
    SCC_TTW,
    SCC_TWW,
    SCC_WTW,
    SCC_WWT,

    SCC_QQQ,
    SCC_WQQ,
    SCC_QWQ,
    SCC_QQW,
    SCC_QWW,
    SCC_WQW,
    SCC_WWQ,


    SCC_SAMEW,
    SCC_SAMEW_HW,
    SCC_SAMET,
    SCC_SAMET_HT,
    SCC_SAMEB,
    SCC_SAMEB_HB,
    SCC_SAMEQ,
    SCC_SAMEQ_HQ,
    SCC_SAMER,
    SCC_SAMER_HR,


    SCC_HWMW,

    SCC_HTMT,
    SCC_HWMT,
    SCC_HTMW,

    SCC_HQMQ,
    SCC_HWMQ,
    SCC_HQMW,


    SCC_HWCW,

    SCC_HTCT,
    SCC_HWCT,
    SCC_HTCW,
    SCC_HTCT_NN,
    SCC_HTCT_PP,

    SCC_HQCQ,
    SCC_HWCQ,
    SCC_HQCW,
    SCC_HQCQ_NN,
    SCC_HQCQ_PP,


    SCC_MWCW,

    SCC_MTCT,
    SCC_MWCT,
    SCC_MTCW,
    SCC_MTCT_NN,
    SCC_MTCT_PP,

    SCC_MQCQ,
    SCC_MWCQ,
    SCC_MQCW,
    SCC_MQCQ_NN,
    SCC_MQCQ_PP,


    SCM_ADJ,

    SCM_TTT,
    SCM_WTT,
    SCM_TWT,
    SCM_TTW,
    SCM_TWW,
    SCM_WTW,
    SCM_WWT,

    SCM_QQQ,
    SCM_WQQ,
    SCM_QWQ,
    SCM_QQW,
    SCM_QWW,
    SCM_WQW,
    SCM_WWQ,



    /* tag 4-grams with context */
    TTTT_NNN,
    TTTT_PPP,
    TTTT_NN,
    TTTT_PP,
    TTXT_NNN,
    TTXT_PPP,
    TXTT_NNN,
    TXTT_PPP,
    XTTT_NNN,
    XTTT_PPP,
    TXTX_NN,
    TXTX_PP,
    TXXT_NN,
    TXXT_PP,
    XTXT_NN,
    XTXT_PP,
    XTTX_NN,
    XTTX_PP,
    XXTT_NN,
    XXTT_PP,

    QQQQ_NNN,
    QQQQ_PPP,
    QQQQ_NN,
    QQQQ_PP,
    QQXQ_NNN,
    QQXQ_PPP,
    QXQQ_NNN,
    QXQQ_PPP,
    XQQQ_NNN,
    XQQQ_PPP,
    QXQX_NN,
    QXQX_PP,
    QXXQ_NN,
    QXXQ_PP,
    XQXQ_NN,
    XQXQ_PP,
    XQQX_NN,
    XQQX_PP,
    XXQQ_NN,
    XXQQ_PP,

    /* backed-off trigrams */
    TTXT,
    WTXT,
    TWXT,
    TTXW,
    TWXW,
    WTXW,
    WWXT,

    QQXQ,
    WQXQ,
    QWXQ,
    QQXW,
    QWXW,
    WQXW,
    WWXQ,

    TXTT,
    WXTT,
    TXWT,
    TXTW,
    TXWW,
    WXTW,
    WXWT,

    QXQQ,
    WXQQ,
    QXWQ,
    QXQW,
    QXWW,
    WXQW,
    WXWQ,

    XTTT,
    XWTT,
    XTWT,
    XTTW,
    XTWW,
    XWTW,
    XWWT,

    XQQQ,
    XWQQ,
    XQWQ,
    XQQW,
    XQWW,
    XWQW,
    XWWQ,

    /* backed-off bigrams */
    HWCW,
    HTCT,
    HTCW,
    HWCT,
    HQCQ,
    HQCW,
    HWCQ,

    HWSW,
    HTST,
    HTSW,
    HWST,
    HQSQ,
    HQSW,
    HWSQ,

    MWCW,
    MTCT,
    MTCW,
    MWCT,
    MQCQ,
    MQCW,
    MWCQ,

    MWSW,
    MTST,
    MTSW,
    MWST,
    MQSQ,
    MQSW,
    MWSQ,

    CWSW,
    CTST,
    CTSW,
    CWST,
    CQSQ,
    CQSW,
    CWSQ,

    /* backed-off unigrams */

    /* always-1, to allow bias on labels themselves */
    BIAS,

    /* word 4-grams are handled by putting them in their own feature
       space, so there is no type for them */

    /* end (count) token */
    END
  };
};

#endif /* FEATTYPES_H */
