/* -*-c++-*- */
#ifndef FEATURE_H
#define FEATURE_H

#include "Bits.h"
#include "Macros.h"


typedef uint64_t fbits;

static inline fbits f0(const uint8_t type) {
  return (((fbits)type) << 56);
}


/* combinations of tags */
static inline fbits ft(const uint8_t type,
                       const tag_t t) {
  return (f0(type) | ((fbits)t));
}

static inline fbits ftt(const uint8_t type,
                        const tag_t t1, const tag_t t2) {
  return (ft(type, t1) |
          (((fbits)t2) << TBITS));
}

static inline fbits fttt(const uint8_t type,
                         const tag_t t1, const tag_t t2,
                         const tag_t t3) {
  return (ftt(type, t1, t2) |
          (((fbits)t3) << (TBITS + TBITS)));
}

static inline fbits ftttt(const uint8_t type,
                          const tag_t t1, const tag_t t2,
                          const tag_t t3, const tag_t t4) {
  return (fttt(type, t1, t2, t3) |
          (((fbits)t4) << (TBITS + TBITS + TBITS)));
}

static inline fbits fttttt(const uint8_t type,
                           const tag_t t1, const tag_t t2,
                           const tag_t t3, const tag_t t4,
                           const tag_t t5) {
  return (ftttt(type, t1, t2, t3, t4) |
          (((fbits)t5) << (TBITS + TBITS + TBITS + TBITS)));
}

static inline fbits ftttttt(const uint8_t type,
                            const tag_t t1, const tag_t t2,
                            const tag_t t3, const tag_t t4,
                            const tag_t t5, const tag_t t6) {
  return (fttttt(type, t1, t2, t3, t4, t5) |
          (((fbits)t6) << (TBITS + TBITS + TBITS + TBITS + TBITS)));
}

static inline fbits fttttttt(const uint8_t type,
                             const tag_t t1, const tag_t t2,
                             const tag_t t3, const tag_t t4,
                             const tag_t t5, const tag_t t6,
                             const tag_t t7) {
  return (ftttttt(type, t1, t2, t3, t4, t5, t6) |
          (((fbits)t7) << (TBITS + TBITS + TBITS + TBITS + TBITS + TBITS)));
}


/* combinations of words */
static inline fbits fw(const uint8_t type,
                       const word_t w) {
  return (f0(type) | ((fbits)w));
}

static inline fbits fww(const uint8_t type,
                        const word_t w1, const word_t w2) {
  return (fw(type, w1) |
          (((fbits)w2) << WBITS));
}

static inline fbits fwww(const uint8_t type,
                         const word_t w1, const word_t w2,
                         const word_t w3) {
  return (fww(type, w1, w2) |
          (((fbits)w3) << (WBITS + WBITS)));
}

static inline fbits fwwww(const word_t w1, const word_t w2,
                          const word_t w3, const word_t w4) {
  return (((fbits)w1) |
          (((fbits)w2) << WBITS) |
          (((fbits)w3) << (WBITS + WBITS)) |
          (((fbits)w4) << (WBITS + WBITS + WBITS)));
}



/*** combinations of words and tags; by convention tags are first ***/

/* one word, multiple tags */
static inline fbits ftw(const uint8_t type,
                        const tag_t t1,
                        const word_t w1) {
  return (ft(type, t1) |
          (((fbits)w1) << TBITS));
}
static inline fbits fttw(const uint8_t type,
                         const tag_t t1, const tag_t t2,
                         const word_t w1) {
  return (ftt(type, t1, t2) |
          (((fbits)w1) << (TBITS + TBITS)));
}
static inline fbits ftttw(const uint8_t type,
                          const tag_t t1, const tag_t t2,
                          const tag_t t3,
                          const word_t w1) {
  return (fttt(type, t1, t2, t3) |
          (((fbits)w1) << (TBITS + TBITS + TBITS)));
}
static inline fbits fttttw(const uint8_t type,
                           const tag_t t1, const tag_t t2,
                           const tag_t t3, const tag_t t4,
                           const word_t w1) {
  return (ftttt(type, t1, t2, t3, t4) |
          (((fbits)w1) << (TBITS + TBITS + TBITS + TBITS)));
}

/* one tag, multiple words */
static inline fbits ftww(const uint8_t type,
                         const tag_t t1,
                         const word_t w1, const word_t w2) {
  return (ftw(type, t1, w1) |
          (((fbits)w2) << (TBITS + WBITS)));
}
static inline fbits ftwww(const uint8_t type,
                          const tag_t t1,
                          const word_t w1, const word_t w2,
                          const word_t w3) {
  return (ftww(type, t1, w1, w2) |
          (((fbits)w3) << (TBITS + WBITS + WBITS)));
}

/* other combinations */
static inline fbits fttww(const uint8_t type,
                          const tag_t t1, const tag_t t2,
                          const word_t w1, const word_t w2) {
  return (fttw(type, t1, t2, w1) |
          (((fbits)w2) << (TBITS + TBITS + WBITS)));
}

static inline fbits ftttww(const uint8_t type,
                           const tag_t t1, const tag_t t2,
                           const tag_t t3,
                           const word_t w1, const word_t w2) {
  CHECK(assert(TBITS + TBITS + TBITS + WBITS + WBITS <= 56));
  return (ftttw(type, t1, t2, t3, w1) |
          (((fbits)w2) << (TBITS + TBITS + TBITS + WBITS)));
}


#endif /* FEATURE_H */
