#include "GZIO.h"

#include <string.h>
#include <assert.h>
#include <stdlib.h>

#include <iostream>
#include <sstream>

#define myname "GZIO"

using std::cerr;
using std::endl;

FILE* gzio::gzopen(const char* const fn, const char* const mode) {
  FILE* f = NULL;
  const size_t n = strlen(fn);
  if(n > 3 &&
     (fn[n - 1] == 'z' || fn[n - 1] == 'Z') &&
     (fn[n - 2] == 'g' || fn[n - 2] == 'G') &&
     fn[n - 3] == '.') {
    char* const cmd = (char*)malloc((n + 64)*sizeof(char));
    if(mode[0] == 'r') {
      snprintf(cmd, n + 64, "gunzip -c '%s'", fn);
    } else if(mode[0] == 'w') {
      snprintf(cmd, n + 64, "gzip -c > '%s'", fn);
    } else {
      cerr << myname << " : unrecognized mode \"" << mode << "\"" << endl;
      assert(0);
    }
    f = popen(cmd, mode);
    free(cmd);
  } else {
    f = fopen(fn, mode);
  }
  if(f == NULL) {
    cerr << myname << " : error opening \"" << fn << "\"" << endl;
    assert(0);
  }
  return f;
}

void gzio::gzclose(const char* const fn, FILE* const f) {
  const size_t n = strlen(fn);
  if(n > 3 &&
     (fn[n - 1] == 'z' || fn[n - 1] == 'Z') &&
     (fn[n - 2] == 'g' || fn[n - 2] == 'G') &&
     fn[n - 3] == '.') {
    const int ret = pclose(f);
    if(ret == -1) {
      cerr << myname << " : error closing \"" << fn << "\"" << endl;
      assert(0);
    }
  } else {
    const int ret = fclose(f);
    if(ret != 0) {
      cerr << myname << " : error closing \"" << fn << "\"" << endl;
      assert(0);
    }
  }
}
