/* -*-c++-*- */
#ifndef HASHFUNCTION_H
#define HASHFUNCTION_H
/* Hashfunction : Default hash functions for a hashtable */

#include "Macros.h"
#include "Bits.h"

#include <assert.h>


/* the output of a hash function */
typedef uint32_t hashcode;

template<class Key>
struct hashfunction {
  /* default implementation throws an error */
  static hashcode hash(const Key& restrict k) { assert(0); return 0; }
  static bool eq(const Key& restrict k1,
                 const Key& restrict k2) {
    assert(0); return false;
  }
};




/* useful macros (for the Jenkins hash) */
#define u32rot(x, k) ((x << k) | (x >> (32 - k)))
#define u32mix(a, b, c) {                  \
      a -= c; a ^= u32rot(c, 4);  c += b;  \
      b -= a; b ^= u32rot(a, 6);  a += c;  \
      c -= b; c ^= u32rot(b, 8);  b += a;  \
      a -= c; a ^= u32rot(c, 16); c += b;  \
      b -= a; b ^= u32rot(a, 19); a += c;  \
      c -= b; c ^= u32rot(b, 4);  b += a;  \
    }
#define u32final(a,b,c) {          \
      c ^= b; c -= u32rot(b, 14);  \
      a ^= c; a -= u32rot(c, 11);  \
      b ^= a; b -= u32rot(a, 25);  \
      c ^= b; c -= u32rot(b, 16);  \
      a ^= c; a -= u32rot(c, 4);   \
      b ^= a; b -= u32rot(a, 14);  \
      c ^= b; c -= u32rot(b, 24);  \
    }


/* template specializations for various important key types */

/* hash function for 64-bit integers, which are useful for, e.g.,
   feature bit-strings */
template<>
struct hashfunction<uint64_t> {
  /* The "dreaded" Jenkins hash */
  static hashcode hash(const uint64_t& restrict k) {
    /* grab high- and low-order bits; add salt */
    hashcode aa = 0x31415926 + ((hashcode)(k & 0xffffffff));
    hashcode bb = 0x27182818 + ((hashcode)(k >> 32));
    hashcode cc = 0xdeadbeef; /* nothing to grab */
    u32final(aa, bb, cc); /* there are only two words, final mixing only */
    return cc;
  }
  /* simple equaliy */
  static bool eq(const uint64_t& restrict k1,
                 const uint64_t& restrict k2) {
    return (k1 == k2);
  }
};

#undef u32rot
#undef u32mix
#undef u32final

#endif /* HASHFUNCTION_H */
