/* -*-c++-*- */
#ifndef MATH_UTIL_H
#define MATH_UTIL_H
/* MathUtil : Various mathematical utilities. */

#include <math.h>
#include <stdlib.h>
#include <assert.h>


/* add two numbers in log-space */
static inline double log_add(const double x, const double y) {
  return ((x > y)
          ? (x + log(1.0 + exp(y - x)))
          : (y + log(1.0 + exp(x - y))));
}

/* compute the relative difference between two quantities */
static inline double reldiff(const double x, const double y) {
  const double absx = (x < 0 ? -x : x);
  const double absy = (y < 0 ? -y : y);
  const double tmp_denom = (absx + absy)*0.5;
  /* be forgiving of small values */
  const double denom = (tmp_denom <= 1 ? 1 : tmp_denom);
  const double diff = x - y;
  const double absdiff = (diff < 0 ? -diff : diff);
  return absdiff/denom;
}

/* compute a random integer in [0,N) */
static inline size_t rand_int(const size_t N) {
  /* random value in [0,1) */
  const double rand_01 = ((double)random())/(1.0 + (double)RAND_MAX);
  /* random integer in [0,N) */
  const size_t rand_0N = (size_t)(((double)N)*rand_01);
  assert(rand_0N < N);
  return rand_0N;
}

#endif /* MATH_UTIL_H */
