/* -*-c++-*- */
#ifndef O1PARSER_H
#define O1PARSER_H
/* O1Parser : First-order parser. */

#include "Macros.h"
#include "O3Active.h"
#include "Bits.h"

#include <vector>

using std::vector;


class o1parser {
public:
  o1parser(const int L, const bool multiroot,
           const bool projective = true);
  ~o1parser();

  template<class PartScorer>
  double argmax(const int N, const PartScorer& restrict S,
                vector<int>& restrict ret_heads,
                vector<int>& restrict ret_labels);

  template<class PartScorer>
  double argmax(const int N, const PartScorer& restrict S,
                const o3active& restrict act,
                vector<int>& restrict ret_heads,
                vector<int>& restrict ret_labels);

private:
  struct scbp;
  const int _L;
  const bool _multiroot;
  const bool _projective;

  int _N;           /* size of the current sentence */
  double* _partsc;  /* an internally-held array of scores */
  lab_t* _bestlab;  /* the best label for any given (h,m) */
  scbp* _TC;        /* the trap chart */
  scbp* _WC;        /* the wedge chart */

  /* these are private to prevent usage */
  o1parser(const o1parser& that);
  o1parser& operator=(const o1parser& that);

  /* allocate/deallocate memory */
  void alloc(const int N);
  void dealloc();

  /* given that _partsc and _C are initialized, parse the sentence */
  double argmax_proj(const int N,
                     vector<int>& restrict ret_heads,
                     vector<int>& restrict ret_labels);
  double argmax_mst(const int N,
                    vector<int>& restrict ret_heads,
                    vector<int>& restrict ret_labels);

  /* bottom-up dynamic-programming */
  void sroot(const int N);
  void trap(const int h, const int m);
  void wedge(const int h, const int e);

  /* unraveling methods */
  void unravel_sroot(const int N,
                     vector<int>& restrict ret_heads) const;
  void unravel_trap(const int h, const int m,
                    vector<int>& restrict ret_heads) const;
  void unravel_wedge(const int h, const int e,
                     vector<int>& restrict ret_heads) const;
};



#ifndef EXPLICIT_TEMPLATES
#include "O1Parser.tcc"
#endif /* EXPLICIT_TEMPLATES */

#endif /* O1PARSER_H */
