/* allocate and populate the matrices */

template<class PartScorer>
double o1parser::argmax(const int N, const PartScorer& restrict S,
                        vector<int>& restrict ret_heads,
                        vector<int>& restrict ret_labels) {
  alloc(N); _N = N;
  const idx_t NN = ICONV(N);
  const lab_t LL = LCONV(_L);
  for(idx_t hh = 0; hh < NN; ++hh) {
    for(idx_t m = 0; m < NN; ++m) {
      const int hm = hh + N*m;
      double& restrict s_hm = _partsc[hm];
      lab_t& restrict l_hm = _bestlab[hm];
      const idx_t h = (hh == m ? INULL : hh);
      for(lab_t l = 0; l < LL; ++l) {
        const double sc = S.dep1(h, m, l);
        if(sc > s_hm) { s_hm = sc; l_hm = l; }
      }
    }
  }
  const double score = (_projective
                        ? argmax_proj(N, ret_heads, ret_labels)
                        : argmax_mst(N,  ret_heads, ret_labels));
  dealloc();
  return score;
}

template<class PartScorer>
double o1parser::argmax(const int N, const PartScorer& restrict S,
                        const o3active& restrict act,
                        vector<int>& restrict ret_heads,
                        vector<int>& restrict ret_labels) {
  alloc(N); _N = N;
  for(int hh = -1; hh < N; ++hh) {
    const idx_t h = ICONV_NULL(hh);
    o3active::rodeps rm(act.romods(hh));
    backp mod;
    while(rm.next(mod)) {
      const idx_t m = mod.i();
      const lab_t l = mod.l();
      const double sc = S.dep1(h, m, l);
      const int hm = (hh == -1 ? m : hh) + N*m;
      if(sc > _partsc[hm]) { _partsc[hm] = sc; _bestlab[hm] = l; }
    }
    o3active::lodeps lm(act.lomods(hh));
    while(lm.next(mod)) {
      const idx_t m = mod.i();
      const lab_t l = mod.l();
      const double sc = S.dep1(h, m, l);
      const int hm = (hh == -1 ? m : hh) + N*m;
      if(sc > _partsc[hm]) { _partsc[hm] = sc; _bestlab[hm] = l; }
    }
  }
  const double score = (_projective
                        ? argmax_proj(N, ret_heads, ret_labels)
                        : argmax_mst(N,  ret_heads, ret_labels));
  dealloc();
  return score;
}
