/* -*-c++-*- */
#ifndef O3ARGMAX_H
#define O3ARGMAX_H
/* O3Argmax : Abstract functor that performs an argmax (parsing)
   operation. */

#include "Macros.h"
#include "PartScoreCache.h"
#include "O3Chart.h"
#include "O3Active.h"

template<class PartScorer>
class o3argmax {
public:
  /* nested classes */
  struct state; /* a struct holding the necessary state */
  typedef part_score_cache<PartScorer> cached_scorer;
  /* perform the argmax */
  double operator()(const int N, const int L, const bool multiroot,
                    const o3active& restrict act,
                    achart& restrict C,
                    const PartScorer& restrict S,
                    o3part::dmap& restrict cache) const;

protected:
  o3argmax();
  virtual ~o3argmax();

  /* inference implementation with virtual dispatch */
  virtual double impl(const state& restrict X,
                      const bool multiroot) const = 0;

private:
  /* these are private to prevent usage */
  o3argmax(const o3argmax& that);
  o3argmax& operator=(const o3argmax& that);
};



/* nested class containing the state for the argmax problem */
class accum_argmax; /* forward declaration */
template<class PartScorer>
struct o3argmax<PartScorer>::state {
  typedef accum_argmax accum;
  const int N;
  const int L;
  const o3active& restrict act;
  achart& restrict C;
  const cached_scorer S;

  state(const int inN, const int inL,
        const o3active& restrict inact,
        achart& restrict inC,
        const PartScorer& restrict inS,
        o3part::dmap& restrict incache);
};



/*** inline implementation ***/

/* nested state class */
template<class PartScorer> inline o3argmax<PartScorer>::state::
state(const int inN, const int inL, const o3active& restrict inact,
      achart& restrict inC, const PartScorer& restrict inS,
      o3part::dmap& restrict incache)
  : N(inN), L(inL), act(inact), C(inC), S(inS, incache) {}

/* nested argmax operator */
template<class PartScorer> inline double
o3argmax<PartScorer>::operator()(const int N, const int L,
                                 const bool multiroot,
                                 const o3active& restrict act,
                                 achart& restrict C,
                                 const PartScorer& restrict S,
                                 o3part::dmap& restrict cache) const {
  return impl(state(N, L, act, C, S, cache), multiroot);
}

#ifndef EXPLICIT_TEMPLATES
#include "O3Argmax.tcc"
#endif /* EXPLICIT_TEMPLATES */

#endif /* O3ARGMAX_H */
