#include "O3AuxPerc.h"
#include "O3Part.h"
#include "O3Tree.h"
#include "O1Parser.h"

#include <stdio.h>

#include <vector>
#include <sstream>

using std::vector;
using std::ostringstream;


class o3auxperc::o1collect {
public:
  o1collect(o3part::vec& restrict ret) : _ret(ret) {}

  /* convert all dependencies to dep1 */
  double dep1(const idx_t h, const idx_t m, const lab_t l_hm) const {
    _ret.push_back(fac::mkdep1(h, m, l_hm));
    return 0;
  }
  double dep2(const idx_t h, const idx_t m, const lab_t l_hm) const {
    _ret.push_back(fac::mkdep1(h, m, l_hm));
    return 0;
  }
  double dep3(const idx_t h, const idx_t m, const lab_t l_hm) const {
    _ret.push_back(fac::mkdep1(h, m, l_hm));
    return 0;
  }

  /* ignore all other parts */
  double sib1(const idx_t h, const idx_t m, const idx_t s,
              const lab_t l_hm, const lab_t l_hs) const {
    return 0;
  }
  double sib2(const idx_t h, const idx_t m, const idx_t s,
              const lab_t l_hm, const lab_t l_hs) const {
    return 0;
  }
  double gci1(const idx_t g, const idx_t h, const idx_t m,
              const lab_t l_gh, const lab_t l_hm) const {
    return 0;
  }
  double gci2(const idx_t g, const idx_t h, const idx_t m,
              const lab_t l_gh, const lab_t l_hm) const {
    return 0;
  }
  double gco1(const idx_t g, const idx_t h, const idx_t m,
              const lab_t l_gh, const lab_t l_hm) const {
    return 0;
  }
  double gco2(const idx_t g, const idx_t h, const idx_t m,
              const lab_t l_gh, const lab_t l_hm) const {
    return 0;
  }
  double gcis(const idx_t g, const idx_t h, const idx_t m, const idx_t s,
              const lab_t l_gh, const lab_t l_hm, const lab_t l_hs) const {
    return 0;
  }
  double gcos(const idx_t g, const idx_t h, const idx_t m, const idx_t s,
              const lab_t l_gh, const lab_t l_hm, const lab_t l_hs) const {
    return 0;
  }
  double sibs(const idx_t h, const idx_t m, const idx_t s, const idx_t c,
              const lab_t l_hm, const lab_t l_hs, const lab_t l_hc) const {
    return 0;
  }

private:
  class fac : protected o3part { friend class o1collect; };
  o3part::vec& restrict _ret;
};


/* lifecycle */
o3auxperc::o3auxperc()
  : o3trainer("AuxP"), _mistakes(0),
    _traux(), _valaux(), _FOP(NULL) {}
o3auxperc::~o3auxperc() {}


/* print usage information */
void o3auxperc::usage_impl() const {
  cerr << "AuxPerc options:" << endl
       << " --proj           : projective parsing" << endl
       << " --mst            : non-projective parsing" << endl
       << " --traux=<path>   : training auxiliary dependencies" << endl
       << " --valaux=<path>  : validation auxiliary dependencies" << endl;
}

/* initialize the trainer */
void o3auxperc::init_impl(const o3trainer::strmap& opts) {
  /* recover regularization constant */
  o3trainer::strmap::const_iterator it;

  if((opts.find("proj") == opts.end()) ==
     (opts.find("mst") == opts.end())) {
    usage("please provide only one of --proj or --mst");
  }

  const bool projective = (opts.find("proj") != opts.end());

  _FOP = new o1parser(_P->L(), _P->multiroot(), projective);

  if((it = opts.find("traux")) == opts.end()) {
    usage("please provide auxiliary dependencies --traux=...");
  }
  _traux.load(it->second.c_str(), _P->L());

  if((it = opts.find("valaux")) == opts.end()) {
    usage("please provide auxiliary dependencies --valaux=...");
  }
  _valaux.load(it->second.c_str(), _P->L());
}


/* training interface */
void o3auxperc::train_start() {
  assert(_traux.N()  == _trdata->N());
  assert(_valaux.N() == _valdata->N());
}
void o3auxperc::iter_start() { _mistakes = 0; }
void o3auxperc::iter_end(string& logline, string& repline) {
  ostringstream os;
  os << _mistakes;
  logline = os.str();
  os.str("");
  os << "mistakes = " << _mistakes;
  repline = os.str();
  /* flush averaged updates */
  _F->Wavg().avg_flush();
}


/* process an example */
void o3auxperc::process(const ex* const restrict e) {
  const int N = (int)e->heads.size();
  assert(_F->L() == _P->L());
  assert(_F->averaged());

  /* compute the true part-vector */
  o3part::vec rstar;
  {
    o1collect get(rstar);
    o3tree::score(_P->multiroot(), false /* sibparser */,
                  e->heads, e->labels, get);
  }

  /* compute the parsed part-vector */
  o3part::vec rhat;
  { /* argmax label */
    vector<int> hhat, lhat;
    _FOP->argmax(N,
                 _F->aux_scorer(e->words, e->tags, _traux.get(e->id())),
                 hhat, lhat);
    o1collect get(rhat);
    o3tree::score(_P->multiroot(), false /* sibparser */,
                  hhat, lhat, get);
  }

  assert(N == (int)rstar.size());
  assert(N == (int)rhat.size());

  /* collect the part-wise contributions */
  o3part::bmap rplus;
  o3part::bmap rminus;

  /* set initial positive and negative contributions */
  for(size_t i = 0; i < rstar.size(); ++i) { rplus.ins(rstar[i], true); }
  for(size_t i = 0; i < rhat.size();  ++i) { rminus.ins(rhat[i], true); }
  /* forbid duplicate parts */
  assert(rstar.size() == rplus.size());
  assert(rhat.size() == rminus.size());

  /* eliminate negative and positive contributions */
  for(size_t i = 0; i < rstar.size(); ++i) { rminus.delq(rstar[i]); }
  for(size_t i = 0; i < rhat.size();  ++i) { rplus.delq(rhat[i]); }

  if(rplus.size() > 0 || rminus.size() > 0) {
    ++_mistakes; /* part difference indicates mistake */
    /* create a scorer for adding values to the parameters */
    o3features::faux_add& adder =
      _F->aux_adder(e->words, e->tags, _traux.get(e->id()));
    /* add 1 to the positive features */
    adder.alpha(1);
    o3part::score(rplus, adder);
    /* subtract 1 from negative features */
    adder.alpha(-1);
    o3part::score(rminus, adder);
  }

  /* advance the averaging timestamps */
  _F->Wavg().avg_tick();
}


/* classify an example */
void o3auxperc::classify(const ex* const restrict e,
                         vector<int>& restrict ret_h,
                         vector<int>& restrict ret_l) const {
  _FOP->argmax((int)(e->words.size()),
               _F->aux_scavg(e->words, e->tags, _valaux.get(e->id())),
               ret_h, ret_l);
}
