#include "O3Tree.h"
#include "TreeFunctors.h"
#include "MathUtil.h"


/* brute-force inference algorithms */
template<class PartScorer> inline double
o3bruteforce::argmax(const int N, const PartScorer& restrict S,
                     vector<int>& restrict ret_heads,
                     vector<int>& restrict ret_labels) {
  _N = N;
  treef_argmax amax;
  foralltrees(S, amax);
  ret_heads  = amax.best_p();
  ret_labels = amax.best_l();
  return amax.best_sc();
}

template<class PartScorer> inline double
o3bruteforce::kargmax_next(const PartScorer& restrict S,
                           double& restrict ret_score,
                           vector<int>& restrict ret_heads,
                           vector<int>& restrict ret_labels) {
  assert(0);
  return -1;
}

template<class PartScorer> inline double
o3bruteforce::partition(const int N, const PartScorer& restrict S) {
  _N = N;
  treef_partition part;
  foralltrees(S, part);
  return part.logZ();
}

template<class PartScorer> inline double
o3bruteforce::marginals(const int N, const PartScorer& restrict S,
                        o3part::dmap& restrict M) {
  _N = N;
  treef_marginals marg(M);
  foralltrees(S, marg);
  marg.mkprobs();
  return marg.logZ();
}





/* enumerate all trees, applying a functor to each */
template<class PartScorer, class TreeFunc> void
o3bruteforce::foralltrees(const PartScorer& restrict S,
                          TreeFunc& restrict f) const {
  vector<int> p(_N, -1); /* N copies of -1 */
  vector<int> l(_N, 0);  /* N copies of 0 */
  vector<int> state;     /* initially empty */
  int nroots = 0;

  if(_multiroot) {
    /* in the multiroot case, the first enumeration is the all-root
       tree, as it is the first tree in the enumeration and is
       valid */
    for(int i = 0; i < _N; ++i) { p[i] = -1; l[i] = 0; }
    nroots = _N;
  } else {
    /* in the singleroot case, the first valid structure has all
       zeros except the last word, which points to root, and the
       first word, which points to the last word (imagine starting
       from all-root and advancing until there is one root; that
       root would be the last word) */
    for(int i = 0; i < _N; ++i) { p[i] = 0; l[i] = 0; }
    p[0] = _N - 1; /* NB: this doesn't create a self-loop for N=1
                      thanks to the next line */
    p[_N - 1] = -1; /* set the last word as root */
    nroots = 1;
  }

  if(_pruned) { /* initialize pruned enumeration */
    const bool tree_exists =
      o3tree::next_tree_pruned(_L, _multiroot, _act, p, l, nroots, state);
    CHECK(assert(tree_exists));
  }

  /* enumerate the valid trees */
  do {
    /* print the tree */
    // cerr << "bruteforce: tree =" << endl;
    // for(int i = 0; i < _N; ++i) {
    //   cerr << "  " << i << " " << p[i] << " " << l[i] << endl;
    // }
    // cerr << endl;
    /* convert the parent/label lists to a part-list */
    o3part::vec y;
    o3part::parts(_L, _multiroot, _sibparser, p, l, y);
    const double score = o3part::score(y, S);
    /* pass the scored tree to the functor */
    f(p, l, y, score);
    /* advance to the next tree */
  } while(_pruned
          ? o3tree::next_tree_pruned(_L, _multiroot, _act,
                                     p, l, nroots, state)
          : o3tree::next_tree_full(_L, _multiroot,
                                   p, l, nroots));
}
