/* -*-c++-*- */
#ifndef O3CHART_H
#define O3CHART_H
/* O3Chart : A sparse table suitable for holding values encountered in
   third-order dynamic-programming parsers.  This class is currently
   implemented as a typedef'd hashtable. */

#include "O3Signature.h"
#include "Backpointer.h"
#include "KQueue.h"
#include "Hashtable.h"

#include <math.h>


/* chart of real values */
typedef hashtable<o3sig,double,o3sig::hashf>  dchart;
/* argmax chart: pairs of scores and backpointers */
struct scobkp; /* pair of (score,backp) */
typedef hashtable<o3sig,scobkp,o3sig::hashf>  achart;
/* k-best parsing chart: queues of candidates */
typedef hashtable<o3sig,kqueue,o3sig::hashf>  qchart;




/* pair of score and backpointer */
struct scobkp {
  /* public access to data */
  double sc;
  backp bp;
  /* create a score-backpointer pairing with the given values */
  explicit scobkp(const double s, const backp b);
  /* create a null score-backpointer pairing with -inf score and
     invalid backpointer; appropriate for initialization before a loop
     in a dynamic-programming algorithm */
  explicit scobkp();
  /* equality comparison */
  bool operator==(const scobkp& that) const;
};



/*** inlined implementation ***/

inline scobkp::scobkp(const double s, const backp b) : sc(s), bp(b) {}
inline scobkp::scobkp() : sc(-HUGE_VAL), bp() {}
inline bool scobkp::operator==(const scobkp& that) const {
  return ((sc == that.sc) && (bp == that.bp));
}

#endif /* O3CHART_H */
