#include "O3FGen.h"

#include <assert.h>
#include <stdlib.h>

#include <iostream>

using std::cerr;
using std::endl;


/* static initializer */
bool o3fgen::USECTAG = false;
bool o3fgen::USEFTAG = false;
bool o3fgen::USECOORD = false;
bool o3fgen::USECOORD_BI = false;
bool o3fgen::USECOMMA = false;
bool o3fgen::DEPMASH = false;
double o3fgen::SCALE_4GRAM = 1.0;

void o3fgen::config(const string& str) {
  const char* p = str.c_str();
  while(*p != '\0') {
    if(p[0] == '+' && p[1] == 'c' && p[2] == 't') {
      cerr << "O3FGen : using ctag features" << endl;
      USECTAG = true; p += 3;
    } else if(p[0] == '+' && p[1] == 'f' && p[2] == 't') {
      cerr << "O3FGen : using ftag features" << endl;
      USEFTAG = true; p += 3;
    } else if(p[0] == '+' && p[1] == 'c' && p[2] == 'c') {
      cerr << "O3FGen : using coordination features" << endl;
      USECOORD = true; p += 3;
    } else if(p[0] == '+' && p[1] == 'c' && p[2] == '2') {
      cerr << "O3FGen : using coordination bigram features" << endl;
      USECOORD_BI = true; p += 3;
    } else if(p[0] == '+' && p[1] == 'c' && p[2] == 'm') {
      cerr << "O3FGen : using comma features" << endl;
      USECOMMA = true; p += 3;
    } else if(p[0] == '+' && p[1] == 'd' && p[2] == 'm') {
      cerr << "O3FGen : mashing dependency features" << endl;
      DEPMASH = true; p += 3;
    } else if(p[0] == '+' && p[1] == '4' && p[2] == 's') {
      char* retp = NULL; p += 3;
      const double scale = strtod(p, &retp);
      cerr << "O3FGen : setting 4-gram scaling to: " << scale << endl;
      SCALE_4GRAM = scale; p = retp;
    } else {
      cerr << "O3FGen : malformed feature conf: " + str
           << " [" << p << "]" << endl;
      assert(0);
    }
  }
  /* report on feature types */
  cerr << "O3FGen : feature types:"
       << " [" << ((unsigned)fbi::END) << " o1]"
       << " [" << ((unsigned)ftri::END) << " o2]"
       << " [" << ((unsigned)fquad::END) << " o3]" << endl;
}
