#include "Mappings.h"
#include "FeatTypes.h"
#include "Feature.h"
#include "Macros.h"

template<class FeatFunc> double o3fgen::
aux(FeatFunc& restrict ffun, const padsent& restrict x,
    const vector<bool>& restrict agree,
    const int h, const int m) {
  CHECK(assert(h >= -1); assert(h < x.N));
  CHECK(assert(m >= 0);  assert(m < x.N));

  const tag_t ht = x.tags[h];
  const tag_t mt = x.tags[m];
  const tag_t hq = mappings::ctag(ht);
  const tag_t mq = mappings::ctag(mt);

  double ret = 0;

  /* convert bool into 0xff (agree) or 0x00 (disagree) */
  unsigned num_agree = 0;
  vector<tag_t> aflags;
  for(unsigned i = 0; i < agree.size(); ++i) {
    num_agree += (agree[i] ? 1 : 0);
    const tag_t aflag = (agree[i] ? TNULL : ((tag_t)0));
    const tag_t idx = TCONV(i);
    ret += ffun(ftt(fbi::AUXD, aflag, idx));
    if(USEFTAG) { ret += ffun(ftttt(fbi::AUXD, aflag, idx, ht, mt)); }
    if(USECTAG) { ret += ffun(ftttt(fbi::AUXD, aflag, idx, hq, mq)); }
  }

  /* add features for unanimity (agreement or disagreement) */
  if((num_agree == 0 || num_agree == agree.size()) &&
     agree.size() > 1) { /* only if there is more than one voter */
    const tag_t aflag = (num_agree == 0 ? ((tag_t)0) : TNULL);
    const tag_t idx = TNULL; /* unanimous index */
    ret += ffun(ftt(fbi::AUXD, aflag, idx));
    if(USEFTAG) { ret += ffun(ftttt(fbi::AUXD, aflag, idx, ht, mt)); }
    if(USECTAG) { ret += ffun(ftttt(fbi::AUXD, aflag, idx, hq, mq)); }
  }

  return ret;
}
