#include "Mappings.h"
#include "FeatTypes.h"
#include "Feature.h"
#include "Macros.h"

template<class FeatFunc> double o3fgen::
gsib(FeatFunc& restrict ffun, FeatFunc& restrict ffunw4,
     FeatFunc& restrict ffunb4, FeatFunc& restrict ffunr4,
     const padsent& restrict x,
     const int h, const int m, const int c, const int s) {
  CHECK(assert(h >= -1); assert(h < x.N));
  CHECK(assert(m >= 0);  assert(m < x.N));
  CHECK(assert(c >= 0);  assert(c < x.N));
  CHECK(assert(s >= -1); assert(s < x.N));

  /* quote- and paren-level features */
  // const word_t qlev_hmcs_norm =
  //   levels4(x.quotelev[h], x.quotelev[m], x.quotelev[c], x.quotelev[s]);
  // const word_t plev_hmcs_norm =
  //   levels4(x.parenlev[h], x.parenlev[m], x.parenlev[c], x.parenlev[s]);

  const word_t hw = x.words[h];
  const word_t mw = x.words[m];
  const word_t cw = x.words[c];
  const word_t sw = x.words[s];

  /* lexical-only features */
  double ret =
    /* word 4-grams */
    // ffunw4(fwwww(hw, mw, cw, sw)) +

    /* backed-off trigrams */
    // ffun(fwww(fquad::HWCWSW, hw, cw, sw)) +

    /* backed-off bigrams */
    ffun(fww(fquad::HWSW, hw, sw)) +
    ffun(fww(fquad::HWCW, hw, cw)) +
    ffun(fww(fquad::CWSW, cw, sw)) +

    /* bias feature */
    ffun(f0(fquad::BIAS));// +

    /* normalized level features */
    // ffun(fw(fquad::QUOTELEV, qlev_hmcs_norm)) +
    // ffun(fw(fquad::PARENLEV, plev_hmcs_norm));


  /* check for a coordination-style 4-gram */
  const bool coord = (USECOORD &&
                      (s != -1) &&
                      ((/* inner child is coord */
                        mappings::iscoord(x.tags[s])) ||
                       (/* inner child is comma... */
                        mappings::iscomma(x.words[s]) &&
                        /* ...with subsequent coord (", and") */
                        mappings::iscoord(x.tags[s + 1]))));

  /* check for a comma-style 4-gram */
  const bool comma = (USECOMMA &&
                      (s != -1) &&
                      (m < s) &&
                      (c == s + 1) &&
                      (/* inner child is comma */
                       mappings::iscomma(x.words[s])));

  // const bool coord = ((s != -1) &&
  //                     ((/* inner child is coord */
  //                       mappings::iscoord(x.tags[s]) &&
  //                       /* outer child is not comma (", and") */
  //                       (!mappings::iscomma(x.words[c]))) ||
  //                      (/* inner child is comma... */
  //                       mappings::iscomma(x.words[s]) &&
  //                       /* ...with subsequent coord (", and") */
  //                       mappings::iscoord(x.tags[s + 1]))));
  if(coord) {
    ret +=
      // ffun(fwww(fquad::SCC_WWW, hw, mw, cw)) +
      ffun(fww(fquad::SCC_HWCW, hw, cw));// +
    if(USECOORD_BI) {
      ret +=
        ffun(fww(fquad::SCC_HWMW, hw, mw)) +
        ffun(fww(fquad::SCC_MWCW, mw, cw));
    }
    if(mw == cw) {
      CHECK(assert(mw != WNULL));
      ret +=
        ffun(ftw(fquad::SCC_SAMEW, 0, WNULL)) +
        ffun(ftw(fquad::SCC_SAMEW, 0, mw)) +
        ffun(ftww(fquad::SCC_SAMEW_HW, 0, mw, hw));
    // } else {
    //   ret +=
    //     ffun(ftw(fquad::SCC_SAMEW, TNULL, WNULL));
    }
  }

  if(comma) {
    const tag_t adjflag = (m == s - 1 ? TNULL : ((tag_t)0));
    ret += ffun(ft(fquad::SCM_ADJ, adjflag));
  }


  if(USEFTAG) { /* tag 4-gram features */
    const tag_t ht = x.tags[h];
    const tag_t mt = x.tags[m];
    const tag_t ct = x.tags[c];
    const tag_t st = x.tags[s];

    // const word_t hb = mappings::wft(hw, ht);
    // const word_t mb = mappings::wft(mw, mt);
    // const word_t cb = mappings::wft(cw, ct);
    // const word_t sb = mappings::wft(sw, st);

    const tag_t hnt = x.tags[h + 1];
    const tag_t hpt = x.tags[h - 1];
    const tag_t mnt = x.tags[m + 1];
    const tag_t mpt = x.tags[m - 1];
    const tag_t cnt = x.tags[c + 1];
    const tag_t cpt = x.tags[c - 1];
    const tag_t snt = x.tags[s + 1];
    const tag_t spt = x.tags[s - 1];

    ret +=
      /* word 4-grams */
      ffun(ftttt(fquad::TTTT, ht, mt, ct, st)) +
      // ffunb4(fwwww(hb, mb, cb, sb)) +

      ffun(ftttw(fquad::WTTT,     mt, ct, st, hw)) +
      ffun(ftttw(fquad::TWTT, ht,     ct, st, mw)) +
      ffun(ftttw(fquad::TTWT, ht, mt,     st, cw)) +
      ffun(ftttw(fquad::TTTW, ht, mt, ct,     sw)) +


      /* 4-gram context features */
      ffun(fttttttt(fquad::TTTT_NNN, ht, mt, ct, st, hnt, mnt, cnt)) +
      ffun(fttttttt(fquad::TTTT_PPP, ht, mt, ct, st, hpt, mpt, cpt)) +
      ffun(ftttttt(fquad::TTTT_NN,   ht, mt, ct, st, hnt, mnt)) +
      ffun(ftttttt(fquad::TTTT_PP,   ht, mt, ct, st, hpt, mpt)) +
      /* trigram context features */
      ffun(ftttttt(fquad::TXTT_NNN,  ht,     ct, st, hnt, cnt, snt)) +
      ffun(ftttttt(fquad::TXTT_PPP,  ht,     ct, st, hpt, cpt, spt)) +
      /* bigram context features */
      ffun(ftttt(fquad::TXTX_NN,     ht,     ct,     hnt, cnt)) +
      ffun(ftttt(fquad::TXTX_PP,     ht,     ct,     hpt, cpt)) +
      ffun(ftttt(fquad::TXXT_NN,     ht,         st, hnt, snt)) +
      ffun(ftttt(fquad::TXXT_PP,     ht,         st, hpt, spt)) +
      ffun(ftttt(fquad::XXTT_NN,             ct, st, cnt, snt)) +
      ffun(ftttt(fquad::XXTT_PP,             ct, st, cpt, spt)) +

      ffun(fttww(fquad::TTWW, ht, mt, cw, sw)) +
      ffun(fttww(fquad::TWTW, ht, ct, mw, sw)) +
      ffun(fttww(fquad::TWWT, ht, st, mw, cw)) +
      ffun(fttww(fquad::WWTT, ct, st, hw, mw)) +
      ffun(fttww(fquad::WTWT, mt, st, hw, cw)) +
      ffun(fttww(fquad::WTTW, mt, ct, hw, sw)) +

      /* backed-off trigrams */
      ffun(fttt(fquad::TXTT, ht, ct, st)) +
      ffun(fttw(fquad::WXTT,     ct, st, hw)) +
      ffun(fttw(fquad::TXWT, ht,     st, cw)) +
      ffun(fttw(fquad::TXTW, ht, ct,     sw)) +
      ffun(ftww(fquad::TXWW, ht,     cw, sw)) +
      ffun(ftww(fquad::WXTW, ct, hw,     sw)) +
      ffun(ftww(fquad::WXWT, st, hw, cw    )) +

      /* backed-off bigrams */
      ffun(ftt(fquad::HTST, ht, st)) +
      ffun(ftt(fquad::HTCT, ht, ct)) +
      ffun(ftt(fquad::CTST, ct, st)) +

      ffun(ftw(fquad::HTSW, ht, sw)) +
      ffun(ftw(fquad::HWST, st, hw)) +
      ffun(ftw(fquad::HTCW, ht, cw)) +
      ffun(ftw(fquad::HWCT, ct, hw)) +
      ffun(ftw(fquad::CTSW, ct, sw)) +
      ffun(ftw(fquad::CWST, st, cw));

    if(coord) {
      ret +=
        ffun(fttt(fquad::SCC_TTT, ht, mt, ct)) +
        ffun(fttw(fquad::SCC_WTT,     mt, ct, hw)) +
        ffun(fttw(fquad::SCC_TWT, ht,     ct, mw)) +
        ffun(fttw(fquad::SCC_TTW, ht, mt,     cw)) +
        ffun(ftww(fquad::SCC_TWW, ht,     mw, cw)) +
        ffun(ftww(fquad::SCC_WTW, mt, hw,     cw)) +
        ffun(ftww(fquad::SCC_WWT, ct, hw, mw    )) +

        // ffun(ftttt(fquad::SCC_HTCT_NN, ht, ct, hnt, cnt)) +
        // ffun(ftttt(fquad::SCC_HTCT_PP, ht, ct, hpt, cpt)) +

        ffun(ftt(fquad::SCC_HTCT, ht, ct)) +
        ffun(ftw(fquad::SCC_HTCW, ht, cw)) +
        ffun(ftw(fquad::SCC_HWCT, ct, hw)) +

        // ffun(ftttt(fquad::SCC_MTCT_NN, mt, ct, mnt, cnt)) +
        // ffun(ftttt(fquad::SCC_MTCT_PP, mt, ct, mpt, cpt)) +

        ffun(ftt(fquad::SCC_MTCT, mt, ct)) +
        ffun(ftw(fquad::SCC_MTCW, mt, cw)) +
        ffun(ftw(fquad::SCC_MWCT, ct, mw));

        if(USECOORD_BI) {
          ret +=
            ffun(ftt(fquad::SCC_HTMT, ht, mt)) +
            ffun(ftw(fquad::SCC_HTMW, ht, mw)) +
            ffun(ftw(fquad::SCC_HWMT, mt, hw));
        }

      if(mt == ct) {
        CHECK(assert(mt != TNULL));
        ret +=
          ffun(ftt(fquad::SCC_SAMET, 0, TNULL)) +
          ffun(ftt(fquad::SCC_SAMET, 0, mt)) +
          ffun(fttt(fquad::SCC_SAMET_HT, 0, mt, ht));
        if(mw == cw) {
          ret +=
            ffun(fttw(fquad::SCC_SAMEB, 0, TNULL, WNULL)) +
            ffun(fttw(fquad::SCC_SAMEB, 0, mt, mw)) +
            ffun(ftttww(fquad::SCC_SAMEB_HB, 0, mt, ht, mw, hw));
        // } else {
        //   ret +=
        //     ffun(fttw(fquad::SCC_SAMEB, TNULL, TNULL, WNULL));
        }
        // if(mnt == cnt) { ret += ffun(f0(fquad::SCC_SAMET_NN)); }
        // if(mpt == cpt) { ret += ffun(f0(fquad::SCC_SAMET_PP)); }
      // } else {
      //   ret +=
      //     ffun(ftt(fquad::SCC_SAMET, TNULL, TNULL)) +
      //     ffun(fttw(fquad::SCC_SAMEB, TNULL, TNULL, WNULL));
      }


      if(comma) {
        ret +=
          ffun(fttt(fquad::SCM_TTT, ht, mt, ct)) +
          ffun(fttw(fquad::SCM_WTT,     mt, ct, hw)) +
          ffun(fttw(fquad::SCM_TWT, ht,     ct, mw)) +
          ffun(fttw(fquad::SCM_TTW, ht, mt,     cw)) +
          ffun(ftww(fquad::SCM_TWW, ht,     mw, cw)) +
          ffun(ftww(fquad::SCM_WTW, mt, hw,     cw)) +
          ffun(ftww(fquad::SCM_WWT, ct, hw, mw    ));
      }
    }
  }


  if(USECTAG) { /* tag 4-gram features */
    const tag_t hq = mappings::ctag(x.tags[h]);
    const tag_t mq = mappings::ctag(x.tags[m]);
    const tag_t cq = mappings::ctag(x.tags[c]);
    const tag_t sq = mappings::ctag(x.tags[s]);

    // const word_t hr = mappings::wct(hw, hq);
    // const word_t mr = mappings::wct(mw, mq);
    // const word_t cr = mappings::wct(cw, cq);
    // const word_t sr = mappings::wct(sw, sq);

    const tag_t hnq = mappings::ctag(x.tags[h + 1]);
    const tag_t hpq = mappings::ctag(x.tags[h - 1]);
    const tag_t mnq = mappings::ctag(x.tags[m + 1]);
    const tag_t mpq = mappings::ctag(x.tags[m - 1]);
    const tag_t cnq = mappings::ctag(x.tags[c + 1]);
    const tag_t cpq = mappings::ctag(x.tags[c - 1]);
    const tag_t snq = mappings::ctag(x.tags[s + 1]);
    const tag_t spq = mappings::ctag(x.tags[s - 1]);

    ret +=
      /* word 4-grams */
      ffun(ftttt(fquad::QQQQ, hq, mq, cq, sq)) +
      // ffunr4(fwwww(hr, mr, cr, sr)) +

      ffun(ftttw(fquad::WQQQ,     mq, cq, sq, hw)) +
      ffun(ftttw(fquad::QWQQ, hq,     cq, sq, mw)) +
      ffun(ftttw(fquad::QQWQ, hq, mq,     sq, cw)) +
      ffun(ftttw(fquad::QQQW, hq, mq, cq,     sw)) +


      /* 4-gram context features */
      ffun(fttttttt(fquad::QQQQ_NNN, hq, mq, cq, sq, hnq, mnq, cnq)) +
      ffun(fttttttt(fquad::QQQQ_PPP, hq, mq, cq, sq, hpq, mpq, cpq)) +
      ffun(ftttttt(fquad::QQQQ_NN,   hq, mq, cq, sq, hnq, mnq)) +
      ffun(ftttttt(fquad::QQQQ_PP,   hq, mq, cq, sq, hpq, mpq)) +
      /* trigram context features */
      ffun(ftttttt(fquad::QXQQ_NNN,  hq,     cq, sq, hnq, cnq, snq)) +
      ffun(ftttttt(fquad::QXQQ_PPP,  hq,     cq, sq, hpq, cpq, spq)) +
      /* bigram context features */
      ffun(ftttt(fquad::QXQX_NN,     hq,     cq,     hnq, cnq)) +
      ffun(ftttt(fquad::QXQX_PP,     hq,     cq,     hpq, cpq)) +
      ffun(ftttt(fquad::QXXQ_NN,     hq,         sq, hnq, snq)) +
      ffun(ftttt(fquad::QXXQ_PP,     hq,         sq, hpq, spq)) +
      ffun(ftttt(fquad::XXQQ_NN,             cq, sq, cnq, snq)) +
      ffun(ftttt(fquad::XXQQ_PP,             cq, sq, cpq, spq)) +

      ffun(fttww(fquad::QQWW, hq, mq, cw, sw)) +
      ffun(fttww(fquad::QWQW, hq, cq, mw, sw)) +
      ffun(fttww(fquad::QWWQ, hq, sq, mw, cw)) +
      ffun(fttww(fquad::WWQQ, cq, sq, hw, mw)) +
      ffun(fttww(fquad::WQWQ, mq, sq, hw, cw)) +
      ffun(fttww(fquad::WQQW, mq, cq, hw, sw)) +

      /* backed-off trigrams */
      ffun(fttt(fquad::QXQQ, hq, cq, sq)) +
      ffun(fttw(fquad::WXQQ,     cq, sq, hw)) +
      ffun(fttw(fquad::QXWQ, hq,     sq, cw)) +
      ffun(fttw(fquad::QXQW, hq, cq,     sw)) +
      ffun(ftww(fquad::QXWW, hq,     cw, sw)) +
      ffun(ftww(fquad::WXQW, cq, hw,     sw)) +
      ffun(ftww(fquad::WXWQ, sq, hw, cw    )) +

      /* backed-off bigrams */
      ffun(ftt(fquad::HQSQ, hq, sq)) +
      ffun(ftt(fquad::HQCQ, hq, cq)) +
      ffun(ftt(fquad::CQSQ, cq, sq)) +

      ffun(ftw(fquad::HQSW, hq, sw)) +
      ffun(ftw(fquad::HWSQ, sq, hw)) +
      ffun(ftw(fquad::HQCW, hq, cw)) +
      ffun(ftw(fquad::HWCQ, cq, hw)) +
      ffun(ftw(fquad::CQSW, cq, sw)) +
      ffun(ftw(fquad::CWSQ, sq, cw));

    if(coord) {
      ret +=
        ffun(fttt(fquad::SCC_QQQ, hq, mq, cq)) +
        ffun(fttw(fquad::SCC_WQQ,     mq, cq, hw)) +
        ffun(fttw(fquad::SCC_QWQ, hq,     cq, mw)) +
        ffun(fttw(fquad::SCC_QQW, hq, mq,     cw)) +
        ffun(ftww(fquad::SCC_QWW, hq,     mw, cw)) +
        ffun(ftww(fquad::SCC_WQW, mq, hw,     cw)) +
        ffun(ftww(fquad::SCC_WWQ, cq, hw, mw    )) +

        // ffun(ftttt(fquad::SCC_HQCQ_NN, hq, cq, hnq, cnq)) +
        // ffun(ftttt(fquad::SCC_HQCQ_PP, hq, cq, hpq, cpq)) +

        ffun(ftt(fquad::SCC_HQCQ, hq, cq)) +
        ffun(ftw(fquad::SCC_HQCW, hq, cw)) +
        ffun(ftw(fquad::SCC_HWCQ, cq, hw)) +

        // ffun(ftttt(fquad::SCC_MQCQ_NN, mq, cq, mnq, cnq)) +
        // ffun(ftttt(fquad::SCC_MQCQ_PP, mq, cq, mpq, cpq)) +

        ffun(ftt(fquad::SCC_MQCQ, mq, cq)) +
        ffun(ftw(fquad::SCC_MQCW, mq, cw)) +
        ffun(ftw(fquad::SCC_MWCQ, cq, mw));

        if(USECOORD_BI) {
          ret +=
            ffun(ftt(fquad::SCC_HQMQ, hq, mq)) +
            ffun(ftw(fquad::SCC_HQMW, hq, mw)) +
            ffun(ftw(fquad::SCC_HWMQ, mq, hw));
        }

      if(mq == cq) {
        CHECK(assert(mq != TNULL));
        ret +=
          ffun(ftt(fquad::SCC_SAMEQ, 0, TNULL)) +
          ffun(ftt(fquad::SCC_SAMEQ, 0, mq)) +
          ffun(fttt(fquad::SCC_SAMEQ_HQ, 0, mq, hq));
        if(mw == cw) {
          ret +=
            ffun(fttw(fquad::SCC_SAMER, 0, TNULL, WNULL)) +
            ffun(fttw(fquad::SCC_SAMER, 0, mq, mw)) +
            ffun(ftttww(fquad::SCC_SAMER_HR, 0, mq, hq, mw, hw));
        // } else {
        //   ret +=
        //     ffun(fttw(fquad::SCC_SAMER, TNULL, TNULL, WNULL));
        }
        // if(mnq == cnq) { ret += ffun(f0(fquad::SCC_SAMEQ_NN)); }
        // if(mpq == cpq) { ret += ffun(f0(fquad::SCC_SAMEQ_PP)); }
      // } else {
      //   ret +=
      //     ffun(ftt(fquad::SCC_SAMEQ, TNULL, TNULL)) +
      //     ffun(fttw(fquad::SCC_SAMER, TNULL, TNULL, WNULL));
      }


      if(comma) {
        ret +=
          ffun(fttt(fquad::SCM_QQQ, hq, mq, cq)) +
          ffun(fttw(fquad::SCM_WQQ,     mq, cq, hw)) +
          ffun(fttw(fquad::SCM_QWQ, hq,     cq, mw)) +
          ffun(fttw(fquad::SCM_QQW, hq, mq,     cw)) +
          ffun(ftww(fquad::SCM_QWW, hq,     mw, cw)) +
          ffun(ftww(fquad::SCM_WQW, mq, hw,     cw)) +
          ffun(ftww(fquad::SCM_WWQ, cq, hw, mw    ));
      }
    }
  }

  return ret;
}
