#include "Mappings.h"
#include "FeatTypes.h"
#include "Feature.h"
#include "Macros.h"

template<class FeatFunc> double o3fgen::
sib(FeatFunc& restrict ffun, const padsent& restrict x,
    const int h, const int m, const int c) {
  CHECK(assert(h >= -1); assert(h < x.N));
  CHECK(assert(m >= 0);  assert(m < x.N));
  CHECK(assert(c >= 0);  assert(c < x.N));
  const word_t hw = x.words[h];
  const word_t mw = x.words[m];
  const word_t cw = x.words[c];

  /* lexical-only features */
  double ret =
    /* full trigrams */
    ffun(fwww(ftri::WWW, hw, mw, cw)) +

    /* backed-off bigrams */
    ffun(fww(ftri::MWCW, mw, cw)) +

    /* backed-off unigrams */
    ffun(fw(ftri::MW, mw)) +
    ffun(fw(ftri::CW, cw)) +

    /* bias feature */
    ffun(f0(ftri::BIAS));


  if(USEFTAG) { /* tag triple features */
    const tag_t ht = x.tags[h];
    const tag_t mt = x.tags[m];
    const tag_t ct = x.tags[c];

    const word_t hb = mappings::wft(hw, ht);
    const word_t mb = mappings::wft(mw, mt);
    const word_t cb = mappings::wft(cw, ct);

    ret +=
      /* full trigrams */
      ffun(fttt(ftri::TTT, ht, mt, ct)) +
      ffun(fwww(ftri::BBB, hb, mb, cb)) +

      ffun(fttw(ftri::BTT,     mt, ct, hb)) +
      ffun(fttw(ftri::TBT, ht,     ct, mb)) +
      ffun(fttw(ftri::TTB, ht, mt,     cb)) +

      /* backed-off bigrams */
      ffun(ftt(ftri::MTCT, mt, ct)) +
      // ffun(ftw(ftri::MTCW, mt, cw)) +
      // ffun(ftw(ftri::MWCT, ct, mw)) +
      ffun(fww(ftri::MBCB, mb, cb)) +

      /* backed-off unigrams */
      ffun(ft(ftri::MT, mt)) +
      ffun(ft(ftri::CT, ct)) +
      ffun(fw(ftri::MB, mb)) +
      ffun(fw(ftri::CB, cb));
  }


  if(USECTAG) { /* tag triple features */
    const tag_t hq = mappings::ctag(x.tags[h]);
    const tag_t mq = mappings::ctag(x.tags[m]);
    const tag_t cq = mappings::ctag(x.tags[c]);

    const word_t hr = mappings::wct(hw, hq);
    const word_t mr = mappings::wct(mw, mq);
    const word_t cr = mappings::wct(cw, cq);

    ret +=
      /* full trigrams */
      ffun(fttt(ftri::QQQ, hq, mq, cq)) +
      ffun(fwww(ftri::RRR, hr, mr, cr)) +

      ffun(fttw(ftri::RQQ,     mq, cq, hr)) +
      ffun(fttw(ftri::QRQ, hq,     cq, mr)) +
      ffun(fttw(ftri::QQR, hq, mq,     cr)) +

      /* backed-off bigrams */
      ffun(ftt(ftri::MQCQ, mq, cq)) +
      // ffun(ftw(ftri::MQCW, mq, cw)) +
      // ffun(ftw(ftri::MWCQ, cq, mw)) +
      ffun(fww(ftri::MRCR, mr, cr)) +

      /* backed-off unigrams */
      ffun(ft(ftri::MQ, mq)) +
      ffun(ft(ftri::CQ, cq)) +
      ffun(fw(ftri::MR, mr)) +
      ffun(fw(ftri::CR, cr));
  }

  return ret;
}


  // const tag_t  hnt = x.tags[h + 1];
  // const tag_t  hpt = x.tags[h - 1];
  // const tag_t  mnt = x.tags[m + 1];
  // const tag_t  mpt = x.tags[m - 1];

    // ffun(ftttt(ftri::TTT_HNT, ht, mt, ct, hnt)) +
    // ffun(ftttt(ftri::TTT_MNT, ht, mt, ct, mnt)) +
    // ffun(ftttt(ftri::TTT_HPT, ht, mt, ct, hpt)) +
    // ffun(ftttt(ftri::TTT_MPT, ht, mt, ct, mpt)) +

    // ffun(fttttt(ftri::TTT_NNTT, ht, mt, ct, hnt, mnt)) +
    // ffun(fttttt(ftri::TTT_NPTT, ht, mt, ct, hnt, mpt)) +
    // ffun(fttttt(ftri::TTT_PNTT, ht, mt, ct, hpt, mnt)) +
    // ffun(fttttt(ftri::TTT_PPTT, ht, mt, ct, hpt, mpt)) +

    // ffun(fttww(ftri::WWT_HNT, hnt, ct, hw, mw)) +
    // ffun(fttww(ftri::WWT_MNT, mnt, ct, hw, mw)) +
    // ffun(fttww(ftri::WWT_HPT, hpt, ct, hw, mw)) +
    // ffun(fttww(ftri::WWT_MPT, mpt, ct, hw, mw)) +

    // ffun(ftttww(ftri::WWT_NNTT, hnt, mnt, ct, hw, mw)) +
    // ffun(ftttww(ftri::WWT_NPTT, hnt, mpt, ct, hw, mw)) +
    // ffun(ftttww(ftri::WWT_PNTT, hpt, mnt, ct, hw, mw)) +
    // ffun(ftttww(ftri::WWT_PPTT, hpt, mpt, ct, hw, mw)) +

    // ffun(ftwww(ftri::WWW_HNT, hnt, hw, mw, cw)) +
    // ffun(ftwww(ftri::WWW_MNT, mnt, hw, mw, cw)) +
    // ffun(ftwww(ftri::WWW_HPT, hpt, hw, mw, cw)) +
    // ffun(ftwww(ftri::WWW_MPT, mpt, hw, mw, cw)) +
