#include "Mappings.h"
#include "FeatTypes.h"
#include "Feature.h"
#include "Macros.h"

template<class FeatFunc> double o3fgen::
sibs(FeatFunc& restrict ffun, FeatFunc& restrict ffunw4,
     FeatFunc& restrict ffunb4, FeatFunc& restrict ffunr4,
     const padsent& restrict x,
     const int h, const int m, const int c, const int s) {
  CHECK(assert(h >= -1); assert(h < x.N));
  CHECK(assert(m >= 0);  assert(m < x.N));
  CHECK(assert(c >= 0);  assert(c < x.N));
  CHECK(assert(s >= -1); assert(s < x.N));

  /* quote- and paren-level features */
  // const word_t qlev_hmcs_norm =
  //   levels4(x.quotelev[h], x.quotelev[m], x.quotelev[c], x.quotelev[s]);
  // const word_t plev_hmcs_norm =
  //   levels4(x.parenlev[h], x.parenlev[m], x.parenlev[c], x.parenlev[s]);

  const word_t hw = x.words[h];
  const word_t mw = x.words[m];
  const word_t cw = x.words[c];
  const word_t sw = x.words[s];

  /* lexical-only features */
  double ret =
    /* word 4-grams */
    // ffunw4(fwwww(hw, mw, cw, sw)) +

    /* backed-off trigrams */
    // ffun(fwww(fquad::HWMWSW, hw, mw, sw)) +
    // ffun(fwww(fquad::MWCWSW, mw, cw, sw)) +

    /* backed-off bigrams */
    ffun(fww(fquad::MWSW, mw, sw)) +
    ffun(fww(fquad::MWCW, mw, cw)) +
    ffun(fww(fquad::CWSW, cw, sw)) +

    /* bias feature */
    ffun(f0(fquad::BIAS));// +

    /* normalized level features */
    // ffun(fw(fquad::QUOTELEV, qlev_hmcs_norm)) +
    // ffun(fw(fquad::PARENLEV, plev_hmcs_norm));


  // const bool coord = ((s != -1) &&
  //                     mappings::iscoord(x.tags[s]) &&
  //                     mappings::iscomma(x.tags[c]));
  // if(coord) {
  //   ret +=
  //     ffun(fww(fquad::SCC_HWMW, hw, mw));
  //   if(hw == mw) { ret += ffun(f0(fquad::SCC_SAMEW)); }
  // }


  if(USEFTAG) { /* tag 4-gram features */
    const tag_t ht = x.tags[h];
    const tag_t mt = x.tags[m];
    const tag_t ct = x.tags[c];
    const tag_t st = x.tags[s];

    // const word_t hb = mappings::wft(hw, ht);
    // const word_t mb = mappings::wft(mw, mt);
    // const word_t cb = mappings::wft(cw, ct);
    // const word_t sb = mappings::wft(sw, st);

    const tag_t hnt = x.tags[h + 1];
    const tag_t hpt = x.tags[h - 1];
    const tag_t mnt = x.tags[m + 1];
    const tag_t mpt = x.tags[m - 1];
    const tag_t cnt = x.tags[c + 1];
    const tag_t cpt = x.tags[c - 1];
    const tag_t snt = x.tags[s + 1];
    const tag_t spt = x.tags[s - 1];

    ret +=
      /* word 4-grams */
      ffun(ftttt(fquad::TTTT, ht, mt, ct, st)) +
      // ffunb4(fwwww(hb, mb, cb, sb)) +

      ffun(ftttw(fquad::WTTT,     mt, ct, st, hw)) +
      ffun(ftttw(fquad::TWTT, ht,     ct, st, mw)) +
      ffun(ftttw(fquad::TTWT, ht, mt,     st, cw)) +
      ffun(ftttw(fquad::TTTW, ht, mt, ct,     sw)) +


      /* 4-gram context features */
      ffun(fttttttt(fquad::TTTT_NNN, ht, mt, ct, st, hnt, mnt, cnt)) +
      ffun(fttttttt(fquad::TTTT_PPP, ht, mt, ct, st, hpt, mpt, cpt)) +
      ffun(ftttttt(fquad::TTTT_NN,   ht, mt, ct, st, hnt, mnt)) +
      ffun(ftttttt(fquad::TTTT_PP,   ht, mt, ct, st, hpt, mpt)) +
      /* trigram context features */
      ffun(ftttttt(fquad::TTXT_NNN,  ht, mt,     st, hnt, mnt, snt)) +
      ffun(ftttttt(fquad::TTXT_PPP,  ht, mt,     st, hpt, mpt, spt)) +
      ffun(ftttttt(fquad::XTTT_NNN,      mt, ct, st, mnt, cnt, snt)) +
      ffun(ftttttt(fquad::XTTT_PPP,      mt, ct, st, mpt, cpt, spt)) +
      /* bigram context features */
      ffun(ftttt(fquad::XTXT_NN,         mt,     st, mnt, snt)) +
      ffun(ftttt(fquad::XTXT_PP,         mt,     st, mpt, spt)) +
      ffun(ftttt(fquad::XTTX_NN,         mt, ct,     mnt, cnt)) +
      ffun(ftttt(fquad::XTTX_PP,         mt, ct,     mpt, cpt)) +
      ffun(ftttt(fquad::XXTT_NN,             ct, st, cnt, snt)) +
      ffun(ftttt(fquad::XXTT_PP,             ct, st, cpt, spt)) +

      ffun(fttww(fquad::TTWW, ht, mt, cw, sw)) +
      ffun(fttww(fquad::TWTW, ht, ct, mw, sw)) +
      ffun(fttww(fquad::TWWT, ht, st, mw, cw)) +
      ffun(fttww(fquad::WWTT, ct, st, hw, mw)) +
      ffun(fttww(fquad::WTWT, mt, st, hw, cw)) +
      ffun(fttww(fquad::WTTW, mt, ct, hw, sw)) +

      /* backed-off trigrams */
      // ffun(fttt(fquad::TTXT, ht, mt, st)) +
      // ffun(fttw(fquad::WTXT,     mt, st, hw)) +
      // ffun(fttw(fquad::TWXT, ht,     st, mw)) +
      // ffun(fttw(fquad::TTXW, ht, mt,     sw)) +
      // ffun(ftww(fquad::TWXW, ht,     mw, sw)) +
      // ffun(ftww(fquad::WTXW, mt, hw,     sw)) +
      // ffun(ftww(fquad::WWXT, st, hw, mw    )) +

      ffun(fttt(fquad::XTTT, mt, ct, st)) +
      ffun(fttw(fquad::XWTT,     ct, st, mw)) +
      ffun(fttw(fquad::XTWT, mt,     st, cw)) +
      ffun(fttw(fquad::XTTW, mt, ct,     sw)) +
      ffun(ftww(fquad::XTWW, mt,     cw, sw)) +
      ffun(ftww(fquad::XWTW, ct, mw,     sw)) +
      ffun(ftww(fquad::XWWT, st, mw, cw    )) +

      /* backed-off bigrams */
      ffun(ftt(fquad::MTST, mt, st)) +
      ffun(ftt(fquad::MTCT, mt, ct)) +
      ffun(ftt(fquad::CTST, ct, st)) +

      ffun(ftw(fquad::MTSW, mt, sw)) +
      ffun(ftw(fquad::MWST, st, mw)) +
      ffun(ftw(fquad::MTCW, mt, cw)) +
      ffun(ftw(fquad::MWCT, ct, mw)) +
      ffun(ftw(fquad::CTSW, ct, sw)) +
      ffun(ftw(fquad::CWST, st, cw));// +

    // if(coord) {
    //   ret +=
    //     ffun(ftt(fquad::SCC_HTMT, ht, mt)) +
    //     ffun(ftw(fquad::SCC_HTMW, ht, mw)) +
    //     ffun(ftw(fquad::SCC_HWMT, mt, hw));// +
    //     // ffun(fww(fquad::SCC_HBMB, hb, mb));
    //   if(ht == mt) { ret += ffun(f0(fquad::SCC_SAMET)); }
    // }
  }


  if(USECTAG) { /* tag 4-gram features */
    const tag_t hq = mappings::ctag(x.tags[h]);
    const tag_t mq = mappings::ctag(x.tags[m]);
    const tag_t cq = mappings::ctag(x.tags[c]);
    const tag_t sq = mappings::ctag(x.tags[s]);

    // const word_t hr = mappings::wct(hw, hq);
    // const word_t mr = mappings::wct(mw, mq);
    // const word_t cr = mappings::wct(cw, cq);
    // const word_t sr = mappings::wct(sw, sq);

    const tag_t hnq = mappings::ctag(x.tags[h + 1]);
    const tag_t hpq = mappings::ctag(x.tags[h - 1]);
    const tag_t mnq = mappings::ctag(x.tags[m + 1]);
    const tag_t mpq = mappings::ctag(x.tags[m - 1]);
    const tag_t cnq = mappings::ctag(x.tags[c + 1]);
    const tag_t cpq = mappings::ctag(x.tags[c - 1]);
    const tag_t snq = mappings::ctag(x.tags[s + 1]);
    const tag_t spq = mappings::ctag(x.tags[s - 1]);

    ret +=
      /* word 4-grams */
      ffun(ftttt(fquad::QQQQ, hq, mq, cq, sq)) +
      // ffunr4(fwwww(hr, mr, cr, sr)) +

      ffun(ftttw(fquad::WQQQ,     mq, cq, sq, hw)) +
      ffun(ftttw(fquad::QWQQ, hq,     cq, sq, mw)) +
      ffun(ftttw(fquad::QQWQ, hq, mq,     sq, cw)) +
      ffun(ftttw(fquad::QQQW, hq, mq, cq,     sw)) +


      /* 4-gram context features */
      ffun(fttttttt(fquad::QQQQ_NNN, hq, mq, cq, sq, hnq, mnq, cnq)) +
      ffun(fttttttt(fquad::QQQQ_PPP, hq, mq, cq, sq, hpq, mpq, cpq)) +
      ffun(ftttttt(fquad::QQQQ_NN,   hq, mq, cq, sq, hnq, mnq)) +
      ffun(ftttttt(fquad::QQQQ_PP,   hq, mq, cq, sq, hpq, mpq)) +
      /* trigram context features */
      ffun(ftttttt(fquad::QQXQ_NNN,  hq, mq,     sq, hnq, mnq, snq)) +
      ffun(ftttttt(fquad::QQXQ_PPP,  hq, mq,     sq, hpq, mpq, spq)) +
      ffun(ftttttt(fquad::XQQQ_NNN,      mq, cq, sq, mnq, cnq, snq)) +
      ffun(ftttttt(fquad::XQQQ_PPP,      mq, cq, sq, mpq, cpq, spq)) +
      /* bigram context features */
      ffun(ftttt(fquad::XQXQ_NN,         mq,     sq, mnq, snq)) +
      ffun(ftttt(fquad::XQXQ_PP,         mq,     sq, mpq, spq)) +
      ffun(ftttt(fquad::XQQX_NN,         mq, cq,     mnq, cnq)) +
      ffun(ftttt(fquad::XQQX_PP,         mq, cq,     mpq, cpq)) +
      ffun(ftttt(fquad::XXQQ_NN,             cq, sq, cnq, snq)) +
      ffun(ftttt(fquad::XXQQ_PP,             cq, sq, cpq, spq)) +

      ffun(fttww(fquad::QQWW, hq, mq, cw, sw)) +
      ffun(fttww(fquad::QWQW, hq, cq, mw, sw)) +
      ffun(fttww(fquad::QWWQ, hq, sq, mw, cw)) +
      ffun(fttww(fquad::WWQQ, cq, sq, hw, mw)) +
      ffun(fttww(fquad::WQWQ, mq, sq, hw, cw)) +
      ffun(fttww(fquad::WQQW, mq, cq, hw, sw)) +

      /* backed-off trigrams */
      // ffun(fttt(fquad::QQXQ, hq, mq, sq)) +
      // ffun(fttw(fquad::WQXQ,     mq, sq, hw)) +
      // ffun(fttw(fquad::QWXQ, hq,     sq, mw)) +
      // ffun(fttw(fquad::QQXW, hq, mq,     sw)) +
      // ffun(ftww(fquad::QWXW, hq,     mw, sw)) +
      // ffun(ftww(fquad::WQXW, mq, hw,     sw)) +
      // ffun(ftww(fquad::WWXQ, sq, hw, mw    )) +

      ffun(fttt(fquad::XQQQ, mq, cq, sq)) +
      ffun(fttw(fquad::XWQQ,     cq, sq, mw)) +
      ffun(fttw(fquad::XQWQ, mq,     sq, cw)) +
      ffun(fttw(fquad::XQQW, mq, cq,     sw)) +
      ffun(ftww(fquad::XQWW, mq,     cw, sw)) +
      ffun(ftww(fquad::XWQW, cq, mw,     sw)) +
      ffun(ftww(fquad::XWWQ, sq, mw, cw    )) +

      /* backed-off bigrams */
      ffun(ftt(fquad::MQSQ, mq, sq)) +
      ffun(ftt(fquad::MQCQ, mq, cq)) +
      ffun(ftt(fquad::CQSQ, cq, sq)) +

      ffun(ftw(fquad::MQSW, mq, sw)) +
      ffun(ftw(fquad::MWSQ, sq, mw)) +
      ffun(ftw(fquad::MQCW, mq, cw)) +
      ffun(ftw(fquad::MWCQ, cq, mw)) +
      ffun(ftw(fquad::CQSW, cq, sw)) +
      ffun(ftw(fquad::CWSQ, sq, cw));

    // if(coord) {
    //   ret +=
    //     ffun(ftt(fquad::SCC_HQMQ, hq, mq)) +
    //     ffun(ftw(fquad::SCC_HQMW, hq, mw)) +
    //     ffun(ftw(fquad::SCC_HWMQ, mq, hw));// +
    //     // ffun(fww(fquad::SCC_HRMR, hr, mr));
    //   if(hq == mq) { ret += ffun(f0(fquad::SCC_SAMEQ)); }
    // }
  }

  return ret;
}
