#include "Mappings.h"
#include "FeatTypes.h"
#include "Feature.h"
#include "Macros.h"

template<class FeatFunc> double o3fgen::
tok(FeatFunc& restrict ffun, const padsent& restrict x,
    const int t, const tag_t tok_type) {
  CHECK(assert(t >= 0); assert(t < x.N));
  /* word and contextual words */
  const word_t tw0  = x.words[t];
  const word_t twp1 = x.words[t - 1];
  const word_t twp2 = x.words[t - 2];
  const word_t twn1 = x.words[t + 1];
  const word_t twn2 = x.words[t + 2];

  double ret = /* lexical-only features */
    ffun(ftw(fbi::TW0,  tok_type, tw0)) +
    ffun(ftw(fbi::TWP1, tok_type, twp1)) +
    ffun(ftw(fbi::TWP2, tok_type, twp2)) +
    ffun(ftw(fbi::TWN1, tok_type, twn1)) +
    ffun(ftw(fbi::TWN2, tok_type, twn2)) +

    ffun(ftww(fbi::TWP01,   tok_type, tw0, twp1)) +
    ffun(ftww(fbi::TWN01,   tok_type, tw0, twn1)) +
    ffun(ftwww(fbi::TWP012, tok_type, tw0, twp1, twp2)) +
    ffun(ftwww(fbi::TWN012, tok_type, tw0, twn1, twn2));


  if(USEFTAG) { /* tag and contextual tags */
    const tag_t tt0  = x.tags[t];
    const tag_t ttp1 = x.tags[t - 1];
    const tag_t ttn1 = x.tags[t + 1];
    const tag_t ttp2 = x.tags[t - 2];
    const tag_t ttn2 = x.tags[t + 2];
    ret +=
      ffun(ftt(fbi::TT0,  tok_type, tt0)) +
      ffun(ftt(fbi::TTP1, tok_type, ttp1)) +
      ffun(ftt(fbi::TTP2, tok_type, ttp2)) +
      ffun(ftt(fbi::TTN1, tok_type, ttn1)) +
      ffun(ftt(fbi::TTN2, tok_type, ttn2)) +

      ffun(fttt(fbi::TTP01,   tok_type, tt0, ttp1)) +
      ffun(fttt(fbi::TTN01,   tok_type, tt0, ttn1)) +
      ffun(ftttt(fbi::TTP012, tok_type, tt0, ttp1, ttp2)) +
      ffun(ftttt(fbi::TTN012, tok_type, tt0, ttn1, ttn2));

    const word_t tb0  = mappings::wft(tw0,  tt0);
    const word_t tbp1 = mappings::wft(twp1, ttp1);
    const word_t tbn1 = mappings::wft(twn1, ttn1);
    const word_t tbp2 = mappings::wft(twp2, ttp2);
    const word_t tbn2 = mappings::wft(twn2, ttn2);
    ret +=
      ffun(ftw(fbi::TB0,  tok_type, tb0)) +
      ffun(ftw(fbi::TBP1, tok_type, tbp1)) +
      ffun(ftw(fbi::TBP2, tok_type, tbp2)) +
      ffun(ftw(fbi::TBN1, tok_type, tbn1)) +
      ffun(ftw(fbi::TBN2, tok_type, tbn2)) +

      ffun(ftww(fbi::TBP01,   tok_type, tb0, tbp1)) +
      ffun(ftww(fbi::TBN01,   tok_type, tb0, tbn1)) +
      ffun(ftwww(fbi::TBP012, tok_type, tb0, tbp1, tbp2)) +
      ffun(ftwww(fbi::TBN012, tok_type, tb0, tbn1, tbn2));
  }


  if(USECTAG) { /* ctag and contextual ctags */
    const tag_t tq0  = mappings::ctag(x.tags[t]);
    const tag_t tqp1 = mappings::ctag(x.tags[t - 1]);
    const tag_t tqn1 = mappings::ctag(x.tags[t + 1]);
    const tag_t tqp2 = mappings::ctag(x.tags[t - 2]);
    const tag_t tqn2 = mappings::ctag(x.tags[t + 2]);
    ret +=
      ffun(ftt(fbi::TQ0,  tok_type, tq0)) +
      ffun(ftt(fbi::TQP1, tok_type, tqp1)) +
      ffun(ftt(fbi::TQP2, tok_type, tqp2)) +
      ffun(ftt(fbi::TQN1, tok_type, tqn1)) +
      ffun(ftt(fbi::TQN2, tok_type, tqn2)) +

      ffun(fttt(fbi::TQP01,   tok_type, tq0, tqp1)) +
      ffun(fttt(fbi::TQN01,   tok_type, tq0, tqn1)) +
      ffun(ftttt(fbi::TQP012, tok_type, tq0, tqp1, tqp2)) +
      ffun(ftttt(fbi::TQN012, tok_type, tq0, tqn1, tqn2));

    const word_t tr0  = mappings::wct(tw0,  tq0);
    const word_t trp1 = mappings::wct(twp1, tqp1);
    const word_t trn1 = mappings::wct(twn1, tqn1);
    const word_t trp2 = mappings::wct(twp2, tqp2);
    const word_t trn2 = mappings::wct(twn2, tqn2);
    ret +=
      ffun(ftw(fbi::TR0,  tok_type, tr0)) +
      ffun(ftw(fbi::TRP1, tok_type, trp1)) +
      ffun(ftw(fbi::TRP2, tok_type, trp2)) +
      ffun(ftw(fbi::TRN1, tok_type, trn1)) +
      ffun(ftw(fbi::TRN2, tok_type, trn2)) +

      ffun(ftww(fbi::TRP01,   tok_type, tr0, trp1)) +
      ffun(ftww(fbi::TRN01,   tok_type, tr0, trn1)) +
      ffun(ftwww(fbi::TRP012, tok_type, tr0, trp1, trp2)) +
      ffun(ftwww(fbi::TRN012, tok_type, tr0, trn1, trn2));
  }

  return ret;
}
