/* dependency features */
template<class FeatFunc> inline double o3features::fgen<FeatFunc>::
dep1(const idx_t h, const idx_t m, const lab_t l_hm) const {
  const bool hbefm = (h == INULL || h < m);
  const bool hadjm = (h != INULL && ((h == m + 1) || (m == h + 1)));
  if(o3fgen::DEPMASH) {
    FeatFunc ffun0(*_W, (hadjm ? W1(hbefm, dep1adj, l_hm)
                               : W1(hbefm, dep, l_hm)), _alpha);
    FeatFunc ffun1(*_W, W1(hbefm, dep1, l_hm), _alpha);
    return 0.5*(o3fgen::dep(ffun0, _x, IUNCONV_NULL(h), m) +
                o3fgen::dep(ffun1, _x, IUNCONV_NULL(h), m));
  } else {
    FeatFunc ffun1(*_W, (hadjm ? W1(hbefm, dep1adj, l_hm)
                               : W1(hbefm, dep1, l_hm)), _alpha);
    return o3fgen::dep(ffun1, _x, IUNCONV_NULL(h), m);
  }
}
template<class FeatFunc> inline double o3features::fgen<FeatFunc>::
dep2(const idx_t h, const idx_t m, const lab_t l_hm) const {
  const bool hbefm = (h == INULL || h < m);
  const bool hadjm = (h != INULL && ((h == m + 1) || (m == h + 1)));
  assert(!hadjm);
  FeatFunc ffun2(*_W, W1(hbefm, dep2, l_hm), _alpha);
  if(o3fgen::DEPMASH) {
    FeatFunc ffun0(*_W, W1(hbefm, dep,  l_hm), _alpha);
    return 0.5*(o3fgen::dep(ffun0, _x, IUNCONV_NULL(h), m) +
                o3fgen::dep(ffun2, _x, IUNCONV_NULL(h), m));
  } else {
    return o3fgen::dep(ffun2, _x, IUNCONV_NULL(h), m);
  }
}
template<class FeatFunc> inline double o3features::fgen<FeatFunc>::
dep3(const idx_t h, const idx_t m, const lab_t l_hm) const {
  const bool hbefm = (h == INULL || h < m);
  const bool hadjm = (h != INULL && ((h == m + 1) || (m == h + 1)));
  assert(!hadjm);
  FeatFunc ffun3(*_W, W1(hbefm, dep3, l_hm), _alpha);
  if(o3fgen::DEPMASH) {
    FeatFunc ffun0(*_W, W1(hbefm, dep,  l_hm), _alpha);
    return 0.5*(o3fgen::dep(ffun0, _x, IUNCONV_NULL(h), m) +
                o3fgen::dep(ffun3, _x, IUNCONV_NULL(h), m));
  } else {
    return o3fgen::dep(ffun3, _x, IUNCONV_NULL(h), m);
  }
}

// template<class FeatFunc> inline double o3features::fgen<FeatFunc>::
// dep1(const idx_t h, const idx_t m, const lab_t l_hm) const {
//   const bool hbefm = (h == INULL || h < m);
//   const bool hadjm = (h != INULL && ((h == m + 1) || (m == h + 1)));
//   FeatFunc ffun1(*_W, (hadjm ? W1(hbefm, dep1adj, l_hm)
//                              : W1(hbefm, dep1, l_hm)), _alpha);
//   return o3fgen::dep(ffun1, _x, IUNCONV_NULL(h), m);
// }
// template<class FeatFunc> inline double o3features::fgen<FeatFunc>::
// dep2(const idx_t h, const idx_t m, const lab_t l_hm) const {
//   const bool hbefm = (h == INULL || h < m);
//   const bool hadjm = (h != INULL && ((h == m + 1) || (m == h + 1)));
//   assert(!hadjm);
//   FeatFunc ffun1(*_W, W1(hbefm, dep1, l_hm), _alpha);
//   return o3fgen::dep(ffun1, _x, IUNCONV_NULL(h), m);
// }
// template<class FeatFunc> inline double o3features::fgen<FeatFunc>::
// dep3(const idx_t h, const idx_t m, const lab_t l_hm) const {
//   const bool hbefm = (h == INULL || h < m);
//   const bool hadjm = (h != INULL && ((h == m + 1) || (m == h + 1)));
//   assert(!hadjm);
//   FeatFunc ffun1(*_W, W1(hbefm, dep1, l_hm), _alpha);
//   return o3fgen::dep(ffun1, _x, IUNCONV_NULL(h), m);
// }



/* loss-augmented versions */
template<class FeatFunc> inline double o3features::fgen_loss<FeatFunc>::
dep1(const idx_t h, const idx_t m, const lab_t l_hm) const {
  /* augment score with a bonus on the wrong parts */
  return this->fgen<FeatFunc>::dep1(h, m, l_hm) +
    ((_heads[m] != h || _labels[m] != l_hm) ? 1 : 0);
}
template<class FeatFunc> inline double o3features::fgen_loss<FeatFunc>::
dep2(const idx_t h, const idx_t m, const lab_t l_hm) const {
  /* augment score with a bonus on the wrong parts */
  return this->fgen<FeatFunc>::dep2(h, m, l_hm) +
    ((_heads[m] != h || _labels[m] != l_hm) ? 1 : 0);
}
template<class FeatFunc> inline double o3features::fgen_loss<FeatFunc>::
dep3(const idx_t h, const idx_t m, const lab_t l_hm) const {
  /* augment score with a bonus on the wrong parts */
  return this->fgen<FeatFunc>::dep3(h, m, l_hm) +
    ((_heads[m] != h || _labels[m] != l_hm) ? 1 : 0);
}




/* auxiliary dependency versions */
template<class FeatFunc> inline double o3features::fgen_aux<FeatFunc>::
dep1(const idx_t h, const idx_t m, const lab_t l_hm) const {
  const bool hbefm = (h == INULL || h < m);
  const bool hadjm = (h != INULL && ((h == m + 1) || (m == h + 1)));
  //FeatFunc ffun0(*_W, W1(hbefm, dep, l_hm), _alpha);
  FeatFunc ffun1(*_W, (hadjm ? W1(hbefm, dep1adj, l_hm)
                             : W1(hbefm, dep1, l_hm)), _alpha);
  return //o3fgen::dep(ffun0, _x, IUNCONV_NULL(h), m) +
    o3fgen::dep(ffun1, _x, IUNCONV_NULL(h), m) +
    o3fgen::aux(ffun1, _x, _aux.agree(h, m, l_hm), IUNCONV_NULL(h), m);
}




/* sibling features */
template<class FeatFunc> inline double o3features::fgen<FeatFunc>::
sib1(const idx_t h, const idx_t m, const idx_t s,
     const lab_t l_hm, const lab_t l_hs) const {
  const bool hbefm = (h == INULL || h < m);
  const bool madjs = ((m == s + 1) || (s == m + 1));
  FeatFunc ffun(*_W, (madjs ? W2(hbefm, sib1adj, l_hm, l_hs)
                            : W2(hbefm, sib1, l_hm, l_hs)), _alpha);
  return o3fgen::sib(ffun, _x, IUNCONV_NULL(h), m, s);
}
template<class FeatFunc> inline double o3features::fgen<FeatFunc>::
sib2(const idx_t h, const idx_t m, const idx_t s,
     const lab_t l_hm, const lab_t l_hs) const {
  const bool hbefm = (h == INULL || h < m);
  const bool madjs = ((m == s + 1) || (s == m + 1));
  const bool hadjs = (h != INULL && ((h == s + 1) || (s == h + 1)));
  assert(!hadjs);
  FeatFunc ffun(*_W, (madjs ? W2(hbefm, sib2adj, l_hm, l_hs)
                            : W2(hbefm, sib2, l_hm, l_hs)), _alpha);
  return o3fgen::sib(ffun, _x, IUNCONV_NULL(h), m, s);
}
// template<class FeatFunc> inline double o3features::fgen<FeatFunc>::
// sib2(const idx_t h, const idx_t m, const idx_t s,
//      const lab_t l_hm, const lab_t l_hs) const {
//   const bool hbefm = (h == INULL || h < m);
//   const bool madjs = ((m == s + 1) || (s == m + 1));
//   FeatFunc ffun(*_W, (madjs ? W2(hbefm, sib1adj, l_hm, l_hs)
//                             : W2(hbefm, sib1, l_hm, l_hs)), _alpha);
//   return o3fgen::sib(ffun, _x, IUNCONV_NULL(h), m, s);
// }


/* inner grandparent */
template<class FeatFunc> inline double o3features::fgen<FeatFunc>::
gci1(const idx_t g, const idx_t h, const idx_t m,
     const lab_t l_gh, const lab_t ll_hm) const {
  const bool gbefh = (g == INULL || g < h);
  const lab_t l_hm = (m == INULL ? _W->Lnull() : ll_hm);
  FeatFunc ffun(*_W, W2(gbefh, gci1, l_gh, l_hm), _alpha);
  return o3fgen::gch(ffun, _x, IUNCONV_NULL(g), h, IUNCONV_NULL(m));
}
template<class FeatFunc> inline double o3features::fgen<FeatFunc>::
gci2(const idx_t g, const idx_t h, const idx_t m,
     const lab_t l_gh, const lab_t ll_hm) const {
  const bool gbefh = (g == INULL || g < h);
  const lab_t l_hm = (m == INULL ? _W->Lnull() : ll_hm);
  FeatFunc ffun(*_W, W2(gbefh, gci2, l_gh, l_hm), _alpha);
  return o3fgen::gch(ffun, _x, IUNCONV_NULL(g), h, IUNCONV_NULL(m));
}
// template<class FeatFunc> inline double o3features::fgen<FeatFunc>::
// gci2(const idx_t g, const idx_t h, const idx_t m,
//      const lab_t l_gh, const lab_t ll_hm) const {
//   const bool gbefh = (g == INULL || g < h);
//   const lab_t l_hm = (m == INULL ? _W->Lnull() : ll_hm);
//   FeatFunc ffun(*_W, W2(gbefh, gci1, l_gh, l_hm), _alpha);
//   return o3fgen::gch(ffun, _x, IUNCONV_NULL(g), h, IUNCONV_NULL(m));
// }


/* outer grandparent */
template<class FeatFunc> inline double o3features::fgen<FeatFunc>::
gco1(const idx_t g, const idx_t h, const idx_t m,
     const lab_t l_gh, const lab_t ll_hm) const {
  const bool gbefh = (g == INULL || g < h);
  const lab_t l_hm = (m == INULL ? _W->Lnull() : ll_hm);
  FeatFunc ffun(*_W, W2(gbefh, gco1, l_gh, l_hm), _alpha);
  return o3fgen::gch(ffun, _x, IUNCONV_NULL(g), h, IUNCONV_NULL(m));
}
template<class FeatFunc> inline double o3features::fgen<FeatFunc>::
gco2(const idx_t g, const idx_t h, const idx_t m,
     const lab_t l_gh, const lab_t ll_hm) const {
  const bool gbefh = (g == INULL || g < h);
  const lab_t l_hm = (m == INULL ? _W->Lnull() : ll_hm);
  FeatFunc ffun(*_W, W2(gbefh, gco2, l_gh, l_hm), _alpha);
  return o3fgen::gch(ffun, _x, IUNCONV_NULL(g), h, IUNCONV_NULL(m));
}
// template<class FeatFunc> inline double o3features::fgen<FeatFunc>::
// gco2(const idx_t g, const idx_t h, const idx_t m,
//      const lab_t l_gh, const lab_t ll_hm) const {
//   const bool gbefh = (g == INULL || g < h);
//   const lab_t l_hm = (m == INULL ? _W->Lnull() : ll_hm);
//   FeatFunc ffun(*_W, W2(gbefh, gco1, l_gh, l_hm), _alpha);
//   return o3fgen::gch(ffun, _x, IUNCONV_NULL(g), h, IUNCONV_NULL(m));
// }




/* outer grand-siblings */
template<class FeatFunc> inline double o3features::fgen<FeatFunc>::
gcis(const idx_t g, const idx_t h, const idx_t m, const idx_t s,
     const lab_t l_gh, const lab_t l_hm, const lab_t ll_hs) const {
  const bool gbefh = (g == INULL || g < h);
  const bool gadjm = (g != INULL && (g == m + 1 || m == g + 1));
  const bool madjs = (s != INULL && (m == s + 1 || s == m + 1));
  const bool sadjh = (s != INULL && (s == h + 1 || h == s + 1));
  const bool adj3 = (madjs && sadjh);
  const bool adj4 = (gadjm && madjs && sadjh);
  const lab_t l_hs = (s == INULL ? _W->Lnull() : ll_hs);
  double ret = 0;
  // if(adj4) { /* purely adjacent grand-sibs */
  //   FeatFunc ffun(*_W,   W3(gbefh, gcisadj4,   l_gh, l_hm, l_hs), _alpha);
  //   FeatFunc ffunw4(*_W, W3(gbefh, gcisadj4w4, l_gh, l_hm, l_hs), _alpha);
  //   FeatFunc ffunb4(*_W, W3(gbefh, gcisadj4b4, l_gh, l_hm, l_hs), _alpha);
  //   FeatFunc ffunr4(*_W, W3(gbefh, gcisadj4r4, l_gh, l_hm, l_hs), _alpha);
  //   ret = o3fgen::gsib(ffun, ffunw4, ffunb4, ffunr4,
  //                      _x, IUNCONV_NULL(g), h, m, IUNCONV_NULL(s));
  // } else if(adj3) { /* adjacent grand-sibs */
  //   FeatFunc ffun(*_W,   W3(gbefh, gcisadj3,   l_gh, l_hm, l_hs), _alpha);
  //   FeatFunc ffunw4(*_W, W3(gbefh, gcisadj3w4, l_gh, l_hm, l_hs), _alpha);
  //   FeatFunc ffunb4(*_W, W3(gbefh, gcisadj3b4, l_gh, l_hm, l_hs), _alpha);
  //   FeatFunc ffunr4(*_W, W3(gbefh, gcisadj3r4, l_gh, l_hm, l_hs), _alpha);
  //   ret = o3fgen::gsib(ffun, ffunw4, ffunb4, ffunr4,
  //                      _x, IUNCONV_NULL(g), h, m, IUNCONV_NULL(s));
  // } else { /* non-adjacent */
  //   FeatFunc ffun(*_W,   W3(gbefh, gcisnadj,   l_gh, l_hm, l_hs), _alpha);
  //   FeatFunc ffunw4(*_W, W3(gbefh, gcisnadjw4, l_gh, l_hm, l_hs), _alpha);
  //   FeatFunc ffunb4(*_W, W3(gbefh, gcisnadjb4, l_gh, l_hm, l_hs), _alpha);
  //   FeatFunc ffunr4(*_W, W3(gbefh, gcisnadjr4, l_gh, l_hm, l_hs), _alpha);
  //   ret = o3fgen::gsib(ffun, ffunw4, ffunb4, ffunr4,
  //                      _x, IUNCONV_NULL(g), h, m, IUNCONV_NULL(s));
  // }
  /* general features */
  FeatFunc ffun(*_W,   W3(gbefh, gcis,   l_gh, l_hm, l_hs), _alpha);
  FeatFunc ffunw4(*_W, W3(gbefh, gcisw4, l_gh, l_hm, l_hs), _alpha);
  FeatFunc ffunb4(*_W, W3(gbefh, gcisb4, l_gh, l_hm, l_hs), _alpha);
  FeatFunc ffunr4(*_W, W3(gbefh, gcisr4, l_gh, l_hm, l_hs), _alpha);
  const double gen =
    o3fgen::gsib(ffun, ffunw4, ffunb4, ffunr4,
                 _x, IUNCONV_NULL(g), h, m, IUNCONV_NULL(s));
  // return 0.5*(ret + gen);
  return gen;
}


/* inner grand-siblings */
template<class FeatFunc> inline double o3features::fgen<FeatFunc>::
gcos(const idx_t g, const idx_t h, const idx_t m, const idx_t s,
     const lab_t l_gh, const lab_t l_hm, const lab_t ll_hs) const {
  const bool gbefh = (g == INULL || g < h);
  const bool gadjh = (g != INULL && (g == h + 1 || h == g + 1));
  const bool hadjs = (s != INULL && (h == s + 1 || s == h + 1));
  const bool sadjm = (s != INULL && (s == m + 1 || m == s + 1));
  const bool adj3 = (hadjs && sadjm);
  const bool adj4 = (gadjh && hadjs && sadjm);
  const lab_t l_hs = (s == INULL ? _W->Lnull() : ll_hs);
  double ret = 0;
  // if(adj4) { /* purely adjacent grand-sibs */
  //   FeatFunc ffun(*_W,   W3(gbefh, gcosadj4,   l_gh, l_hm, l_hs), _alpha);
  //   FeatFunc ffunw4(*_W, W3(gbefh, gcosadj4w4, l_gh, l_hm, l_hs), _alpha);
  //   FeatFunc ffunb4(*_W, W3(gbefh, gcosadj4b4, l_gh, l_hm, l_hs), _alpha);
  //   FeatFunc ffunr4(*_W, W3(gbefh, gcosadj4r4, l_gh, l_hm, l_hs), _alpha);
  //   ret = o3fgen::gsib(ffun, ffunw4, ffunb4, ffunr4,
  //                      _x, IUNCONV_NULL(g), h, m, IUNCONV_NULL(s));
  // } else if(adj3) { /* adjacent grand-sibs */
  //   FeatFunc ffun(*_W,   W3(gbefh, gcosadj3,   l_gh, l_hm, l_hs), _alpha);
  //   FeatFunc ffunw4(*_W, W3(gbefh, gcosadj3w4, l_gh, l_hm, l_hs), _alpha);
  //   FeatFunc ffunb4(*_W, W3(gbefh, gcosadj3b4, l_gh, l_hm, l_hs), _alpha);
  //   FeatFunc ffunr4(*_W, W3(gbefh, gcosadj3r4, l_gh, l_hm, l_hs), _alpha);
  //   ret = o3fgen::gsib(ffun, ffunw4, ffunb4, ffunr4,
  //                      _x, IUNCONV_NULL(g), h, m, IUNCONV_NULL(s));
  // } else { /* non-adjacent */
  //   FeatFunc ffun(*_W,   W3(gbefh, gcosnadj,   l_gh, l_hm, l_hs), _alpha);
  //   FeatFunc ffunw4(*_W, W3(gbefh, gcosnadjw4, l_gh, l_hm, l_hs), _alpha);
  //   FeatFunc ffunb4(*_W, W3(gbefh, gcosnadjb4, l_gh, l_hm, l_hs), _alpha);
  //   FeatFunc ffunr4(*_W, W3(gbefh, gcosnadjr4, l_gh, l_hm, l_hs), _alpha);
  //   ret = o3fgen::gsib(ffun, ffunw4, ffunb4, ffunr4,
  //                      _x, IUNCONV_NULL(g), h, m, IUNCONV_NULL(s));
  // }
  /* general features */
  FeatFunc ffun(*_W,   W3(gbefh, gcos,   l_gh, l_hm, l_hs), _alpha);
  FeatFunc ffunw4(*_W, W3(gbefh, gcosw4, l_gh, l_hm, l_hs), _alpha);
  FeatFunc ffunb4(*_W, W3(gbefh, gcosb4, l_gh, l_hm, l_hs), _alpha);
  FeatFunc ffunr4(*_W, W3(gbefh, gcosr4, l_gh, l_hm, l_hs), _alpha);
  const double gen =
    o3fgen::gsib(ffun, ffunw4, ffunb4, ffunr4,
                 _x, IUNCONV_NULL(g), h, m, IUNCONV_NULL(s));
  // return 0.5*(ret + gen);
  return gen;
}


/* tri-siblings */
template<class FeatFunc> inline double o3features::fgen<FeatFunc>::
sibs(const idx_t h, const idx_t m, const idx_t c, const idx_t s,
     const lab_t l_hm, const lab_t l_hc, const lab_t ll_hs) const {
  const bool hbefm = (h == INULL || h < m);
  const bool hadjs = (h != INULL && s != INULL &&
                      (h == s + 1 || s == h + 1));
  const bool sadjc = (s != INULL && (s == c + 1 || c == s + 1));
  const bool cadjm = (c == m + 1 || m == c + 1);
  const bool adj3 = false;//(sadjc && cadjm);
  const bool adj4 = false;//(hadjs && sadjc && cadjm);
  const lab_t l_hs = (s == INULL ? _W->Lnull() : ll_hs);
  double ret = 0;
  // if(adj4) { /* purely adjacent grand-sibs */
  //   FeatFunc ffun(*_W,   W3(hbefm, sibsadj4,   l_hm, l_hc, l_hs), _alpha);
  //   FeatFunc ffunw4(*_W, W3(hbefm, sibsadj4w4, l_hm, l_hc, l_hs), _alpha);
  //   FeatFunc ffunb4(*_W, W3(hbefm, sibsadj4b4, l_hm, l_hc, l_hs), _alpha);
  //   FeatFunc ffunr4(*_W, W3(hbefm, sibsadj4r4, l_hm, l_hc, l_hs), _alpha);
  //   ret = o3fgen::sibs(ffun, ffunw4, ffunb4, ffunr4,
  //                      _x, IUNCONV_NULL(h), m, c, IUNCONV_NULL(s));
  // } else if(adj3) { /* adjacent grand-sibs */
  //   FeatFunc ffun(*_W,   W3(hbefm, sibsadj3,   l_hm, l_hc, l_hs), _alpha);
  //   FeatFunc ffunw4(*_W, W3(hbefm, sibsadj3w4, l_hm, l_hc, l_hs), _alpha);
  //   FeatFunc ffunb4(*_W, W3(hbefm, sibsadj3b4, l_hm, l_hc, l_hs), _alpha);
  //   FeatFunc ffunr4(*_W, W3(hbefm, sibsadj3r4, l_hm, l_hc, l_hs), _alpha);
  //   ret = o3fgen::sibs(ffun, ffunw4, ffunb4, ffunr4,
  //                      _x, IUNCONV_NULL(h), m, c, IUNCONV_NULL(s));
  // } else { /* not purely adjacent */
  //   FeatFunc ffun(*_W,   W3(hbefm, sibsnadj,   l_hm, l_hc, l_hs), _alpha);
  //   FeatFunc ffunw4(*_W, W3(hbefm, sibsnadjw4, l_hm, l_hc, l_hs), _alpha);
  //   FeatFunc ffunb4(*_W, W3(hbefm, sibsnadjb4, l_hm, l_hc, l_hs), _alpha);
  //   FeatFunc ffunr4(*_W, W3(hbefm, sibsnadjr4, l_hm, l_hc, l_hs), _alpha);
  //   ret = o3fgen::sibs(ffun, ffunw4, ffunb4, ffunr4,
  //                      _x, IUNCONV_NULL(h), m, c, IUNCONV_NULL(s));
  // }
  /* general features */
  FeatFunc ffun(*_W,   W3(hbefm, sibs,   l_hm, l_hc, l_hs), _alpha);
  FeatFunc ffunw4(*_W, W3(hbefm, sibsw4, l_hm, l_hc, l_hs), _alpha);
  FeatFunc ffunb4(*_W, W3(hbefm, sibsb4, l_hm, l_hc, l_hs), _alpha);
  FeatFunc ffunr4(*_W, W3(hbefm, sibsr4, l_hm, l_hc, l_hs), _alpha);
  const double gen =
    o3fgen::sibs(ffun, ffunw4, ffunb4, ffunr4,
                 _x, IUNCONV_NULL(h), m, c, IUNCONV_NULL(s));
  // return 0.5*(ret + gen);
  return gen;
}
