#include "Bits.h"
#include "O3Signature.h"


template<class State>
void o3gbfinside<State>::
lbox(const State& restrict X, const int i_s, const int i_m) {
  accum A;
  const lab_t LL = LCONV(X.L);
  const idx_t s = ICONV(i_s);
  const idx_t m = ICONV(i_m);
  const bool sbefm = (i_s < i_m);
  const int beg_h  = (sbefm ? -1 : (X.N - 1));
  const int inc_hs = (sbefm ? 1  : -1);
  for(lab_t l_hm = 0; l_hm < LL; ++l_hm) {
    for(lab_t l_hs = 0; l_hs < LL; ++l_hs) {
      for(int i_h = beg_h; i_h != i_s; i_h += inc_hs) {
        const idx_t h = ICONV_NULL(i_h);
        const o3box bb(h, m, s, l_hm, l_hs);
        for(int i_rs = i_s; i_rs != i_m; i_rs += inc_hs) {
          const int i_rm = i_rs + inc_hs;
          const idx_t rs = ICONV(i_rs);
          const idx_t rm = ICONV(i_rm);
          const double sc = X.S.sib1(h, m, s, l_hm, l_hs) +
            (rs == s ? X.S.gco1(h, s, INULL, l_hs, LNULL)
             : accum::get(X.C, o3wedge(bb, rs))) +
            (rm == m ? X.S.gci1(h, m, INULL, l_hm, LNULL)
             : accum::get(X.C, o3wedge(rm, bb)));
          A.accum(sc, backp((uint16_t)rm));
        }
        X.C.ins(bb, A.finish());
      }
    }
  }
}


template<class State>
void o3gbfinside<State>::
ltrap(const State& restrict X, const int i_h, const int i_m) {
  accum A;
  const lab_t LL = LCONV(X.L);
  const idx_t h = ICONV_NULL(i_h);
  const idx_t m = ICONV(i_m);
  const bool hbefm = (i_h < i_m);
  const bool nullh = (i_h == -1);
  const int inc_hm = (hbefm ? 1 : -1);
  const int i_h1 = i_h + inc_hm;
  const idx_t h1 = ICONV(i_h1);
  const int i_gci_beg = (hbefm ? i_m + 1 : -1);
  const int i_gci_end = (hbefm ? X.N     : i_m);
  const int i_gco_beg = (hbefm ? -1                    : i_h + 1);
  const int i_gco_end = (hbefm ? (i_h == -1 ? 0 : i_h) : X.N);
  const lab_t num_l_gh = (nullh ? ((lab_t)1) : LL);
  for(lab_t ll_gh = 0; ll_gh < num_l_gh; ++ll_gh) {
    const lab_t l_gh = (nullh ? LNULL : ll_gh);
    for(lab_t l_hm = 0; l_hm < LL; ++l_hm) {
      /* handle grandparents on gci side */
      for(int i_gci = i_gci_beg; i_gci != i_gci_end; ++i_gci) {
        const idx_t gci = ICONV_NULL(i_gci);
        const o3trap tt(gci, h, m, l_gh, l_hm);
        /* null sibling, innermost dep */
        const double null_sc = X.S.dep1(h, m, l_hm) +
          /* compute grandparent score if necessary */
          (nullh ? 0 : (X.S.gci1(gci, h, m, l_gh, l_hm) +
                        X.S.gcis(gci, h, m, INULL, l_gh, l_hm, LNULL))) +
          /* almost-complete wedge construction */
          (h1 == m ? X.S.gci1(h, m, INULL, l_hm, LNULL)
           : accum::get(X.C, o3wedge(tt, h1)));
        A.accum(null_sc, backp(h1, LINV));
        if(i_h1 != i_m) { /* potential siblings actually exist */
          /* base dependency and grandparent score */
          const double gdep_sc = X.S.dep2(h, m, l_hm) +
            /* compute grandparent score if necessary */
            (nullh ? 0 : X.S.gci2(gci, h, m, l_gh, l_hm));
          for(lab_t l_hs = 0; l_hs < LL; ++l_hs) {
            for(int i_s = h1; i_s != i_m; i_s += inc_hm) {
              const idx_t s = ICONV(i_s);
              const backp sb(s, l_hs);
              const double sc = gdep_sc +
                (nullh ? 0 : X.S.gcis(gci, h, m, s, l_gh, l_hm, l_hs)) +
                accum::get(X.C, o3box(tt, sb)) +
                accum::get(X.C, o3trap(tt, sb));
              A.accum(sc, sb);
            }
          }
        }
        X.C.ins(tt, A.finish());
      }
      /* handle grandparents on gco side */
      for(int i_gco = i_gco_beg; i_gco != i_gco_end; ++i_gco) {
        const idx_t gco = ICONV_NULL(i_gco);
        const o3trap tt(gco, h, m, l_gh, l_hm);
        /* null sibling, innermost dep */
        const double null_sc = X.S.dep1(h, m, l_hm) +
          /* compute grandparent score if necessary */
          (nullh ? 0 : (X.S.gco1(gco, h, m, l_gh, l_hm) +
                        X.S.gcos(gco, h, m, INULL, l_gh, l_hm, LNULL))) +
          /* almost-complete wedge construction */
          (h1 == m ? X.S.gci1(h, m, INULL, l_hm, LNULL)
           : accum::get(X.C, o3wedge(tt, h1)));
        A.accum(null_sc, backp(h1, LINV));
        if(i_h1 != i_m) { /* potential siblings actually exist */
          /* base dependency and grandparent score */
          const double gdep_sc = X.S.dep2(h, m, l_hm) +
            /* compute grandparent score if necessary */
            (nullh ? 0 : X.S.gco2(gco, h, m, l_gh, l_hm));
          for(lab_t l_hs = 0; l_hs < LL; ++l_hs) {
            for(int i_s = h1; i_s != i_m; i_s += inc_hm) {
              const idx_t s = ICONV(i_s);
              const backp sb(s, l_hs);
              const double sc = gdep_sc +
                (nullh ? 0 : X.S.gcos(gco, h, m, s, l_gh, l_hm, l_hs)) +
                accum::get(X.C, o3box(tt, sb)) +
                accum::get(X.C, o3trap(tt, sb));
              A.accum(sc, sb);
            }
          }
        }
        X.C.ins(tt, A.finish());
      }
    }
  }
}


template<class State>
void o3gbfinside<State>::
lwedge(const State& restrict X, const int i_h, const int i_e) {
  accum A;
  const int N1_tmp = X.N - 1;
  const idx_t N1 = ICONV(N1_tmp);
  const lab_t LL = LCONV(X.L);
  const idx_t h = ICONV_NULL(i_h);
  const idx_t e = ICONV(i_e);
  const bool hbefe = (i_h < i_e);
  const bool nullh = (i_h == -1);
  const int inc_he = (hbefe ? 1 : -1);
  const int i_h1 = i_h + inc_he;
  const idx_t h1 = ICONV(i_h1);
  const int i_lg_beg = -1;
  const int i_lg_end = (hbefe ? (i_h == -1 ? 0 : i_h) : i_e);
  const int i_rg_beg = (hbefe ? i_e                   : i_h) + 1;
  const int i_rg_end = X.N;
  const lab_t num_l_gh = (nullh ? ((lab_t)1) : LL);
  for(lab_t ll_gh = 0; ll_gh < num_l_gh; ++ll_gh) {
    const lab_t l_gh = (nullh ? LNULL : ll_gh);
    /* handle grandparents on left side */
    for(int i_lg = i_lg_beg; i_lg != i_lg_end; ++i_lg) {
      const idx_t lg = ICONV_NULL(i_lg);
      const o3wedge ww(lg, h, e, l_gh);
      /* outermost dependency/grandchild deprecated */
      /* X.S.odep(h, m, l_hm) + */
      /* (h == INULL ? 0 : X.S.o ## GP(g, h, m, l_gh, l_hm)) + */
      for(lab_t l_hm = 0; l_hm < LL; ++l_hm) {
        /* non-null outer wedges */
        for(int i_m = h1; i_m != i_e; i_m += inc_he) {
          const idx_t m = ICONV(i_m);
          const backp mb(m, l_hm);
          const double sc =
            accum::get(X.C, o3trap(ww, mb)) +
            accum::get(X.C, o3wedge(ww, mb));
          A.accum(sc, mb);
        }
        /* null outer wedge, m = e */
        const backp mb(e, l_hm);
        const double sc =
          X.S.gco1(h, e, INULL, l_hm, LNULL) +
          accum::get(X.C, o3trap(ww, mb));
        A.accum(sc, mb);
      }
      X.C.ins(ww, A.finish());
    }
    /* handle grandparents on right side */
    for(int i_rg = i_rg_beg; i_rg != i_rg_end; ++i_rg) {
      const idx_t rg = ICONV_NULL(i_rg);
      const o3wedge ww(rg, h, e, l_gh);
      /* outermost dependency/grandchild deprecated */
      /* X.S.odep(h, m, l_hm) + */
      /* (h == INULL ? 0 : X.S.o ## GP(g, h, m, l_gh, l_hm)) + */
      for(lab_t l_hm = 0; l_hm < LL; ++l_hm) {
        /* non-null outer wedges */
        for(int i_m = h1; i_m != i_e; i_m += inc_he) {
          const idx_t m = ICONV(i_m);
          const backp mb(m, l_hm);
          const double sc =
            accum::get(X.C, o3trap(ww, mb)) +
            accum::get(X.C, o3wedge(ww, mb));
          A.accum(sc, mb);
        }
        /* null outer wedge, m = e */
        const backp mb(e, l_hm);
        const double sc =
          X.S.gco1(h, e, INULL, l_hm, LNULL) +
          accum::get(X.C, o3trap(ww, mb));
        A.accum(sc, mb);
      }
      X.C.ins(ww, A.finish());
    }
  }
}


template<class State>
void o3gbfinside<State>::
sroot(const State& restrict X) {
  accum A;
  const int N1_tmp = X.N - 1;
  const idx_t N1 = ICONV(N1_tmp);
  const lab_t LL = LCONV(X.L);
  for(lab_t ll = 0; ll < LL; ++ll) {
    for(idx_t rr = 0; rr <= N1; ++rr) {
      const backp rb(rr, ll);
      const double sc = X.S.dep1(INULL, rr, ll) +
        (rr == N1 ? X.S.gco1(INULL, rr, INULL, ll, LNULL)
         : accum::get(X.C, o3wedge(rb, N1))) +
        (rr == 0  ? X.S.gci1(INULL, rr, INULL, ll, LNULL)
         : accum::get(X.C, o3wedge(rb, 0)));
      A.accum(sc, rb);
    }
  }
  X.C.ins(o3wedge(N1), A.finish());
}




template<class State>
double o3gbfinside<State>::
inside(const State& restrict X, const bool multiroot) {
  /* maximum span width to consider */
  const int max_width = (multiroot ? X.N : (X.N - 1));
  /* consider spans starting from here */
  const int start_idx = (multiroot ? -1 : 0);

  for(int w = 1; w <= max_width; ++w) { /* for small to large spans */
    const int end_idx = X.N - w;
    for(int i_s = start_idx; i_s < end_idx; ++i_s) {
      const int i_e = i_s + w; /* end index of the span */
      if(i_s != -1) {
        lbox(X, i_s, i_e);
        lbox(X, i_e, i_s);
      }
      ltrap(X, i_s, i_e);
      if(i_s != -1) { ltrap(X, i_e, i_s); }
      lwedge(X, i_s, i_e);
      if(i_s != -1) { lwedge(X, i_e, i_s); }
    }
  }

  /* single-root constructions */
  if(!multiroot) { sroot(X); }

  const int N1_tmp = X.N - 1;
  const idx_t N1 = ICONV(N1_tmp);
  return accum::get(X.C, o3wedge(N1));
}
