/* -*-c++-*- */
#ifndef O3GTPARGMAX_H
#define O3GTPARGMAX_H
/* O3GTPArgmax : The third-order grandchild parser, implemented as a
   top-down pruned parser. */

#include "O3Argmax.h"

template<class PartScorer>
class o3gtpargmax : public o3argmax<PartScorer> {
public:
  o3gtpargmax();
  virtual ~o3gtpargmax();

protected:
  typedef class o3argmax<PartScorer>::state state;
  virtual double impl(const state& restrict X,
                      const bool multiroot) const;

private:
  /* top-down recursive inference */

  /* find the best singleroot construction */
  static double singleroot(const state& restrict X);

  /* score a box, assuming an ordering on the indices */
  static double box_hsm(const state& restrict X, const o3box bb);
  static double box_msh(const state& restrict X, const o3box bb);

  /* score a trapezoid, assuming an ordering on the indices */
  static double trap_ghm(const state& restrict X, const o3trap tt);
  static double trap_mhg(const state& restrict X, const o3trap tt);
  static double trap_gmh(const state& restrict X, const o3trap tt);
  static double trap_hmg(const state& restrict X, const o3trap tt);

  /* score a wedge, assuming an ordering on the indices */
  static double wedge_ghe(const state& restrict X, const o3wedge ww);
  static double wedge_ehg(const state& restrict X, const o3wedge ww);
  static double wedge_geh(const state& restrict X, const o3wedge ww);
  static double wedge_heg(const state& restrict X, const o3wedge ww);
};

#ifndef EXPLICIT_TEMPLATES
#include "O3GTPArgmax.tcc"
#endif /* EXPLICIT_TEMPLATES */

#endif /* O3GTPARGMAX_H */
