#include "Macros.h"

#include <assert.h>


/* lifecycle */
template<class PartScorer>
o3gtpargmax<PartScorer>::o3gtpargmax() {}
template<class PartScorer>
o3gtpargmax<PartScorer>::~o3gtpargmax() {}


template<class PartScorer> double
o3gtpargmax<PartScorer>::impl(const state& restrict X,
                              const bool multiroot) const {
  const int N1 = X.N - 1;
  return (multiroot
          /* construct a wedge spanning the sentence */
          ? wedge_ghe(X, o3wedge(ICONV(N1)))
          /* use singleroot construction */
          : singleroot(X));
}


/* find the best singleroot construction */
template<class PartScorer> double
o3gtpargmax<PartScorer>::singleroot(const state& restrict X) {
  const int tmp_N1 = X.N - 1; /* index of last word */
  const idx_t N1 = ICONV(tmp_N1);
  const achart::entry* el; /* return value for pget() */
  const achart::entry* er; /* return value for pget() */
  scobkp best;
  backp rb;
  o3active::rodeps roots(X.act.romods(-1));
  while(roots.next(rb)) {
    const o3wedge lwedge(rb, 0);    /* left wedge */
    const o3wedge rwedge(rb, N1);   /* right wedge */
    const double sc = X.S.dep1(INULL, rb.i(), rb.l()) +
      /* left half */
      (rb.i() == 0  ? X.S.gci1(INULL, rb.i(), INULL, rb.l(), LNULL)
       /* use cached score if e != NULL; otherwise compute */
       : (((el = X.C.pget(lwedge)) != NULL) ? el->val.sc
          : wedge_geh(X, lwedge))) +
      /* right half */
      (rb.i() == N1 ? X.S.gco1(INULL, rb.i(), INULL, rb.l(), LNULL)
       /* use cached score if e != NULL; otherwise compute */
       : (((er = X.C.pget(rwedge)) != NULL) ? er->val.sc
          : wedge_ghe(X, rwedge)));
    if(sc > best.sc) { best.sc = sc; best.bp = rb; }
  }
  X.C.ins(o3wedge(N1), best);
  return best.sc;
}



/* the remaining functions make assumptions about the ordering of the
   input indices, to avoid tiresome if-else nestings.  macros are used
   for code duplication. */

/* find the best box */
template<class PartScorer> double o3gtpargmax<PartScorer>::
box_hsm(const state& restrict X, const o3box bb) {
  /* construct this part */
#define opt_box(INC, INCC, ORDER_SWEDGE, ORDER_MWEDGE)                  \
  CHECK(assert(!X.C.has(bb)));                                          \
  const achart::entry* es; /* return value for pget() */                \
  const achart::entry* em; /* return value for pget() */                \
  scobkp best;                                                          \
  const double sib_sc =                                                 \
    X.S.sib1(bb.h(), bb.m(), bb.s(), bb.l_hm(), bb.l_hs());             \
  for(idx_t rs = bb.s(); rs != bb.m(); INCC rs) {                       \
    const idx_t rm = ICONV(rs INC 1);                                   \
    const o3wedge swedge(bb, rs);  /* s-side wedge */                   \
    const o3wedge mwedge(rm, bb);  /* m-side wedge */                   \
    const double sc = sib_sc +                                          \
      /* s-side half */                                                 \
      (rs == bb.s()                                                     \
       ? X.S.gco1(bb.h(), bb.s(), INULL, bb.l_hs(), LNULL)              \
       /* use cached score if e != NULL; otherwise compute */           \
       : (((es = X.C.pget(swedge)) != NULL) ? es->val.sc                \
          : wedge_ ## ORDER_SWEDGE(X, swedge))) +                       \
      /* m-side half */                                                 \
      (rm == bb.m()                                                     \
       ? X.S.gci1(bb.h(), bb.m(), INULL, bb.l_hm(), LNULL)              \
       /* use cached score if e != NULL; otherwise compute */           \
       : (((em = X.C.pget(mwedge)) != NULL) ? em->val.sc                \
          : wedge_ ## ORDER_MWEDGE(X, mwedge)));                        \
    if(sc > best.sc) { best.sc = sc; best.bp = backp(rm); }             \
  }                                                                     \
  X.C.ins(bb, best);                                                    \
  return best.sc


  CHECK(assert(bb.s() != INULL); assert(bb.m() != INULL);
        assert(bb.h() == INULL || bb.h() < bb.s());
        assert(bb.s() < bb.m()));
  opt_box(+, ++, ghe, geh);
}
template<class PartScorer> double o3gtpargmax<PartScorer>::
box_msh(const state& restrict X, const o3box bb) {
  CHECK(assert(bb.m() != INULL);
        assert(bb.m() < bb.s());
        assert(bb.s() < bb.h()));
  opt_box(-, --, ehg, heg);
}
#undef opt_box


/* find the best trapezoid */
template<class PartScorer> double o3gtpargmax<PartScorer>::
trap_ghm(const state& restrict X, const o3trap tt) {
  /* construct this part */
#define opt_trap(DIR, NULLH, ORDER_TRAP, ORDER_BOX, ORDER_WEDGE, GPART) \
  CHECK(assert(!X.C.has(tt)));                                          \
  const achart::entry* eb; /* return value for pget() */                \
  const achart::entry* et; /* return value for pget() */                \
  /* null sibling: innermost dep */                                     \
  const o3wedge null_wedge(tt, h1);                                     \
  const double null_sc = X.S.dep1(tt.h(), tt.m(), tt.l_hm()) +          \
    /* compute grandparent score if necessary */                        \
    ((NULLH && (tt.h() == INULL)) ? 0                                   \
     : (X.S.GPART ## 1(tt.g(), tt.h(), tt.m(), tt.l_gh(), tt.l_hm()) +  \
        X.S.GPART ## s(tt.g(), tt.h(), tt.m(), INULL,                   \
                       tt.l_gh(), tt.l_hm(), LNULL))) +                 \
    /* almost-complete wedge construction */                            \
    (h1 == tt.m() ? X.S.gci1(tt.h(), tt.m(), INULL, tt.l_hm(), LNULL)   \
     /* use cached score if e != NULL; otherwise compute */             \
     : (((eb = X.C.pget(null_wedge)) != NULL) ? eb->val.sc              \
        : wedge_ ## ORDER_WEDGE(X, null_wedge)));                       \
  scobkp best(null_sc, backp(h1, LINV));                                \
  o3active::DIR ## odeps sibs(X.act.DIR ## omods(hh, tt.m()));          \
  if(sibs.hasnext()) {                                                  \
    /* score for dep and gch in the non-innermost case */               \
    const double gdep_sc = X.S.dep2(tt.h(), tt.m(), tt.l_hm()) +        \
      /* compute grandparent scores if necessary */                     \
      ((NULLH && (tt.h() == INULL)) ? 0                                 \
       : X.S.GPART ## 2(tt.g(), tt.h(), tt.m(), tt.l_gh(), tt.l_hm())); \
    backp sb;                                                           \
    while(sibs.next(sb)) {                                              \
      const o3box sbox(tt, sb);                                         \
      const o3trap strap(tt, sb);                                       \
      /* non-null sibling: second or further dep */                     \
      const double sc = gdep_sc +                                       \
        /* compute grandparent scores if necessary */                   \
        ((NULLH && (tt.h() == INULL)) ? 0                               \
         : X.S.GPART ## s(tt.g(), tt.h(), tt.m(), sb.i(),               \
                          tt.l_gh(), tt.l_hm(), sb.l())) +              \
        /* box half split at s */                                       \
        /* use cached score if e != NULL; otherwise compute */          \
        (((eb = X.C.pget(sbox)) != NULL) ? eb->val.sc                   \
         : box_ ## ORDER_BOX(X, sbox)) +                                \
        /* trapezoid half split at s */                                 \
        /* use cached score if e != NULL; otherwise compute */          \
        (((et = X.C.pget(strap)) != NULL) ? et->val.sc                  \
         : trap_ ## ORDER_TRAP(X, strap));                              \
      if(sc > best.sc) { best.sc = sc; best.bp = sb; }                  \
    }                                                                   \
  }                                                                     \
  X.C.ins(tt, best);                                                    \
  return best.sc


  CHECK(if(tt.g() == tt.h()) { assert(tt.g() == INULL); }
        else { assert(tt.g() == INULL || tt.g() < tt.h()); }
        assert(tt.h() == INULL || tt.h() < tt.m()));
  const int hh = (tt.h() == INULL ? -1 : tt.h());
  const int tmp_h1 = hh + 1;
  const idx_t h1 = ICONV(tmp_h1);
  opt_trap(r, true, ghm, hsm, geh, gco);
}
template<class PartScorer> double o3gtpargmax<PartScorer>::
trap_mhg(const state& restrict X, const o3trap tt) {
  CHECK(assert(tt.g() > tt.h()); assert(tt.h() > tt.m()));
  const idx_t hh = tt.h();
  const int tmp_h1 = ((int)hh) - 1;
  const idx_t h1 = ICONV(tmp_h1);
  opt_trap(l, false, mhg, msh, heg, gco);
}
template<class PartScorer> double o3gtpargmax<PartScorer>::
trap_gmh(const state& restrict X, const o3trap tt) {
  CHECK(assert(tt.g() == INULL || tt.g() < tt.m());
        assert(tt.m() < tt.h()));
  const idx_t hh = tt.h();
  const int tmp_h1 = ((int)hh) - 1;
  const idx_t h1 = ICONV(tmp_h1);
  opt_trap(l, false, gmh, msh, heg, gci);
}
template<class PartScorer> double o3gtpargmax<PartScorer>::
trap_hmg(const state& restrict X, const o3trap tt) {
  CHECK(assert(tt.g() > tt.m()); assert(tt.m() > tt.h()));
  const idx_t hh = tt.h();
  const int tmp_h1 = ((int)hh) + 1;
  const idx_t h1 = ICONV(tmp_h1);
  opt_trap(r, false, hmg, hsm, geh, gci);
}
#undef opt_trap



/* find the best wedge */
template<class PartScorer> double o3gtpargmax<PartScorer>::
wedge_ghe(const state& restrict X, const o3wedge ww) {
  /* construct this part */
#define opt_wedge(DIR, INC, ORDER_TRAP, ORDER_WEDGE, GPART)             \
  CHECK(assert(!X.C.has(ww)));                                          \
  const achart::entry* et; /* return value for pget() */                \
  const achart::entry* ew; /* return value for pget() */                \
  scobkp best;                                                          \
  backp mb;                                                             \
  o3active::DIR ## odeps                                                \
    gcs(X.act.DIR ## omods(hh, ((int)ww.e()) INC 1));                   \
  while(gcs.next(mb)) {                                                 \
    const o3trap mtrap(ww, mb);                                         \
    const o3wedge mwedge(ww, mb);                                       \
    const double sc =                                                   \
      /* outermost dependency/grandchild deprecated */                  \
      /* X.S.odep(h, m, l_hm) + */                                      \
      /* (h == -1 ? 0 : X.S.o ## GPART(g, h, m, l_gh, l_hm)) + */       \
      /* trapezoid half split at m */                                   \
      /* use cached score if e != NULL; otherwise compute */            \
      (((et = X.C.pget(mtrap)) != NULL) ? et->val.sc                    \
       : trap_ ## ORDER_TRAP(X, mtrap)) +                               \
      /* wedge half split at m */                                       \
      (mb.i() == ww.e()                                                 \
       ? X.S.gco1(ww.h(), mb.i(), INULL, mb.l(), LNULL)                 \
       /* use cached score if e != NULL; otherwise compute */           \
       : (((ew = X.C.pget(mwedge)) != NULL) ? ew->val.sc                \
          : wedge_ ## ORDER_WEDGE(X, mwedge)));                         \
    if(sc > best.sc) { best.sc = sc; best.bp = mb; }                    \
  }                                                                     \
  X.C.ins(ww, best);                                                    \
  return best.sc


  CHECK(if(ww.g() == ww.h()) { assert(ww.g() == INULL); }
        else { assert(ww.g() == INULL || ww.g() < ww.h()); }
        assert(ww.h() == INULL || ww.h() < ww.e()));
  const int hh = (ww.h() == INULL ? -1 : ww.h());
  opt_wedge(r, +, ghm, ghe, gco);
}
template<class PartScorer> double o3gtpargmax<PartScorer>::
wedge_ehg(const state& restrict X, const o3wedge ww) {
  CHECK(assert(ww.g() > ww.h()); assert(ww.h() > ww.e()));
  const idx_t hh = ww.h();
  opt_wedge(l, -, mhg, ehg, gco);
}
template<class PartScorer> double o3gtpargmax<PartScorer>::
wedge_geh(const state& restrict X, const o3wedge ww) {
  CHECK(assert(ww.g() == INULL || ww.g() < ww.e());
        assert(ww.e() < ww.h()));
  const idx_t hh = ww.h();
  opt_wedge(l, -, gmh, ehg, gci);
}
template<class PartScorer> double o3gtpargmax<PartScorer>::
wedge_heg(const state& restrict X, const o3wedge ww) {
  CHECK(assert(ww.g() > ww.e()); assert(ww.e() > ww.h()));
  const idx_t hh = ww.h();
  opt_wedge(r, +, hmg, ghe, gci);
}
#undef opt_wedge
