/* -*-c++-*- */
#ifndef O3PARAMS_AVG_H
#define O3PARAMS_AVG_H

/* O3ParamsAvg : A sparse parameter vector with fast averaging */

#include "O3Params.h"
#include "AvgParam.h"

class o3params_avg : public o3params<avg_param,avg_param_util> {
public:
  o3params_avg(const int L);
  ~o3params_avg();

  /* advance to the next time-step in the averaged updates */
  void avg_tick();
  /* bring the params up to date with timestamp now *and* reset the
     timestamps to zero */
  void avg_flush();
  /* recover the current timestamp */
  uint64_t now() const;


private:
  /* additional data members on top of o3params */
  uint64_t _now;  /* current timestamp */
};


/* inlined methods for o3params_avg */
inline void o3params_avg::avg_tick()      { ++_now; }
inline uint64_t o3params_avg::now() const { return _now; }


#endif /* O3PARAMS_AVG_H */
