/* -*-c++-*- */
#ifndef O3PARAMS_DOUBLE_UTIL_H
#define O3PARAMS_DOUBLE_UTIL_H
/* O3ParamsDoubleUtil : Utilities for instantiating parameters with
   double values. */

#include "Macros.h"

class o3params_double_util {
public:
  static void fprint(FILE* const f, const double& restrict d) {
    fprintf(f, "%.17g", d);
  }
  static double fscan(FILE* const f) {
    double d = 0;
    const int nscanned = fscanf(f, " %lf", &d);
    assert(nscanned == 1);
    return d;
  }
  static bool equals(const double& restrict d1,
                     const double& restrict d2,
                     const double tol) {
    const double diff = d1 - d2;
    const double absdiff = (diff < 0 ? -diff : diff);
    return (absdiff <= tol);
  }
  static void send(const double& restrict d,
                   const int write_fd, const double scale) {
    double tmp = (scale*d);
    safe_write(write_fd, &tmp, sizeof(double));
  }
  static void recv(double& restrict d, const int read_fd) {
    double tmp = 0;
    safe_read(read_fd, &tmp, sizeof(double));
    d += tmp;
  }
  static void add(double& restrict d, const double& restrict d2) {
    d += d2;
  }
};

#endif /* O3PARAMS_DOUBLE_UTIL_H */
