#include "O3ParamsLin.h"

#include <assert.h>
#include <string.h>
#include <math.h>

/* automatically rescale when the scaling factor is multiplicatively
   either above or below this ratio */
#define SCALE_TOL 10

o3params_lin::o3params_lin(const int L)
  : o3params<double,o3params_double_util>(L),
    _scale(1.0),
    _inv_scale(1.0) {}
o3params_lin::~o3params_lin() {}



/* define scale() here to keep SCALE_TOL hidden */
void o3params_lin::scale(const double alpha) {
  /* zero-scaling: clear all params */
  if(alpha == 0) { zero(); _scale = _inv_scale = 1.0; return; }
  _scale     *= alpha;
  _inv_scale /= alpha;
  if(_scale > SCALE_TOL || _inv_scale > SCALE_TOL) { rescale(); }
}

void o3params_lin::rescale() {
  for(int i = 0; i < _nW; ++i) {
    primals& W = _W_arr[i];
    primals::iterator it = W.begin();
    const primals::iterator iend = W.end();
    for(; it != iend; ++it) { it->val *= _scale; }
  }
  _scale = _inv_scale = 1.0;
}
