/* -*-c++-*- */
#ifndef O3PARAMS_LIN_H
#define O3PARAMS_LIN_H

/* O3ParamsLin : A sparse parameter vector supporting fast scaling */

#include "O3Params.h"
#include "O3ParamsDoubleUtil.h"

class featf_lin_add;

class o3params_lin : public o3params<double,o3params_double_util> {
public:
  o3params_lin(const int L);
  ~o3params_lin();

  /* scale the parameter vector */
  void scale(const double alpha);
  /* apply the current scale to the parameters and renormalize */
  void rescale();
  /* compute the addition to this resulting from the scale */
  double scale_add(const double d) const;


private:
  friend class featf_lin_add;
  /* a scalar parameter for adjusting the scale of the parameters */
  double _scale;
  double _inv_scale;

  static double scale_add(const double d, const double is);
};

/* inlined methods for o3params_lin */
inline double o3params_lin::scale_add(const double d, const double is) {
  return is*d;
}
inline double o3params_lin::scale_add(const double d) const {
  return scale_add(d, _inv_scale);
}

#endif /* O3PARAMS_LIN_H */
