#include "O3Parser.h"

#include <math.h>
#include <assert.h>
#include <stdlib.h>


/* lifecycle */
o3parser::o3parser(const int inL,
                   const bool inmultiroot,
                   const bool inpruned,
                   const bool intopdown,
                   const bool insibparser)
  : _L(inL),
    _multiroot(inmultiroot),
    _pruned(inpruned),
    _topdown(intopdown),
    _sibparser(insibparser),
    _act(),
    _N(0),
    _K(-1),
    _Kmax(-1),
    _C(8),
    _I(8),
    _O(8),
    _Q(8) {
  /* sanity checks on necessary math ops */
  assert(exp(-HUGE_VAL) <= 0);
  assert(-HUGE_VAL == -HUGE_VAL);
  assert(_L > 0);
  assert(_L <= O3PARSER_MAX_LABELS);
}
/* delete the charts */
o3parser::~o3parser() {}

/* disabled functions */
o3parser::o3parser()
  : _L(0),
    _multiroot(false), _pruned(false),
    _topdown(false), _sibparser(false) { assert(0); }
o3parser::o3parser(const o3parser& that)
  : _L(0),
    _multiroot(false), _pruned(false),
    _topdown(false), _sibparser(false) { assert(0); }
o3parser& o3parser::operator=(const o3parser& that) {
  assert(0); return *this;
}



void o3parser::kargmax_init(const int in_N, const int Kmax) {
  assert(in_N > 0); assert(in_N <= O3PARSER_MAX_SENTLEN);
  _N = in_N;
  assert(Kmax > 0); assert(_Kmax == -1);
  _Kmax = Kmax;

  /* check existence of pruned deps if necessary */
  if(_pruned) { assert(_act.N() == _N); }
  else        { assert(_act.N() == -1); }

  /* set up the k-best rank index */
  assert(_K == -1);
  _K = 0;

  /* queue chart must be empty */
  assert(_cache.size() == 0);
  assert(_Q.size() == 0);
}


void o3parser::kargmax_done() {
  /* unset the rank index */
  _K = -1; _Kmax = -1;

  /* deallocate queue chart */
  _cache.clear();
  _Q.clear();
}
