#include "O3Unravel.h"
#include "O3GTPArgmax.h"
#include "O3GBFArgmax.h"
#include "O3STPArgmax.h"
#include "O3SBFArgmax.h"

#include <assert.h>
#include <stdlib.h>


/* inference routines: set up and tear down only, actual inference is
   handled by o3argmax and related classes */
template<class PartScorer> double o3parser::
argmax(const int in_N, const PartScorer& restrict S,
       vector<int>& restrict ret_heads,
       vector<int>& restrict ret_labels) {
  assert(in_N > 0); assert(in_N <= O3PARSER_MAX_SENTLEN);
  _N = in_N;

  /* check existence of pruned deps if necessary */
  if(_pruned) { assert(_act.N() == _N); }
  else        { assert(_act.N() == -1); }

  /* check argmax chart of max-values and backpointers */
  assert(_cache.size() == 0);
  assert(_C.size() == 0);

  /* parse the sentence... */
  double score = -HUGE_VAL;
  if(_sibparser) { /* sibling parser */
    if(_topdown) { /* top-down sibling parser */
      const o3stpargmax<PartScorer> P;
      score = P(_N, _L, _multiroot, _act, _C, S, _cache);
    } else { /* bottom-up sibling parser */
      if(_pruned) {
        assert(0);
      } else {
        const o3sbfargmax<PartScorer> P;
        score = P(_N, _L, _multiroot, _act, _C, S, _cache);
      }
    }
  } else { /* grandchild parser */
    if(_topdown) { /* top-down grandchild parser */
      const o3gtpargmax<PartScorer> P;
      score = P(_N, _L, _multiroot, _act, _C, S, _cache);
    } else { /* bottom-up grandchild parser */
      if(_pruned) {
        assert(0);
      } else {
        const o3gbfargmax<PartScorer> P;
        score = P(_N, _L, _multiroot, _act, _C, S, _cache);
      }
    }
  }

  /* unravel backpointers */
  if(_sibparser) { /* unravel sibling parser's tree */
    o3unravel::sibdeps(_N, _multiroot, _C, ret_heads, ret_labels);
  } else { /* unravel grandchild parser's tree */
    o3unravel::gchdeps(_N, _multiroot, _C, ret_heads, ret_labels);
  }

  /* clear the chart */
  _cache.clear();
  _C.clear();

  return score;
}


template<class PartScorer> int o3parser::
kargmax_next(const PartScorer& restrict S,
             double& restrict ret_score,
             vector<int>& restrict ret_heads,
             vector<int>& restrict ret_labels) {
  /* make sure things are initialized */
  assert(_K >= 0);

  //return (_K = -1); /* signal parsing failure */

  assert(0); /* FIXME */
  return _K++; /* postincrement... */
}


template<class PartScorer> double o3parser::
partition(const int in_N, const PartScorer& restrict S) {
  assert(in_N > 0); assert(in_N <= O3PARSER_MAX_SENTLEN);
  _N = in_N;

  /* check existence of pruned deps if necessary */
  if(_pruned) { assert(_act.N() == _N); }
  else        { assert(_act.N() == -1); }

  /* check chart */
  assert(_I.size() == 0);

  const double logZ = 0;
  assert(0);

  /* clear the chart */
  _cache.clear();
  _I.clear();

  return logZ;
}


template<class PartScorer> double o3parser::
marginals(const int in_N, const PartScorer& restrict S,
          o3part::dmap& restrict M) {
  assert(in_N > 0); assert(in_N <= O3PARSER_MAX_SENTLEN);
  _N = in_N;

  /* check existence of pruned deps if necessary */
  if(_pruned) { assert(_act.N() == _N); }
  else        { assert(_act.N() == -1); }

  /* make sure charts are initialized */
  assert(_I.size() == 0);
  assert(_O.size() == 0);

  /* virtual dispatch... */
  const double logZ = 0;
  assert(0);

  /* clear the charts */
  _cache.clear();
  _I.clear();
  _O.clear();

  return logZ;
}
