#include "O3Part.h"

#include "PartCollect.h"
#include "O3Tree.h"

#include <iostream>
#include <sstream>

using std::cerr;
using std::endl;
using std::ostringstream;


/* create a part-list for p,l */
void o3part::parts(const int L,
                   const bool multiroot, const bool sibparser,
                   const vector<int>& restrict p,
                   const vector<int>& restrict l,
                   o3part::vec& restrict ret) {
  const o3tree* const root = o3tree::convert(p, l);
  part_collect collect((int)p.size(), L, sibparser, ret);
  o3tree::score(multiroot, sibparser, root, collect);
  delete root;
}



#define IINT(x) IUNCONV_NULL(x)
#define LINT(x) LUNCONV_NULL(x)
string o3part::tostring() const {
  /* dispatch according to type */
  ostringstream os;
  switch(type()) {
  case O3PART_DEP1:
    os << "DEP1 h=" << IINT(i2()) << " m=" << IINT(i3())
       << " l_hm=" << LINT(l2());
    break;
  case O3PART_DEP2:
    os << "DEP2 h=" << IINT(i2()) << " m=" << IINT(i3())
       << " l_hm=" << LINT(l2());
    break;
  case O3PART_DEP3:
    os << "DEP3 h=" << IINT(i2()) << " m=" << IINT(i3())
       << " l_hm=" << LINT(l2());
    break;
  case O3PART_SIB1:
    os << "SIB1 h=" << IINT(i2()) << " m=" << IINT(i3())
       << " c=" << IINT(i4())
       << " l_hm=" << LINT(l2()) << " l_hc=" << LINT(l3());
    break;
  case O3PART_SIB2:
    os << "SIB2 h=" << IINT(i2()) << " m=" << IINT(i3())
       << " c=" << IINT(i4())
       << " l_hm=" << LINT(l2()) << " l_hc=" << LINT(l3());
    break;
  case O3PART_GCI1:
    os << "GCI1 h=" << IINT(i1()) << " m=" << IINT(i2())
       << " c=" << IINT(i3())
       << " l_hm=" << LINT(l1()) << " l_mc=" << LINT(l2());
    break;
  case O3PART_GCI2:
    os << "GCI2 h=" << IINT(i1()) << " m=" << IINT(i2())
       << " c=" << IINT(i3())
       << " l_hm=" << LINT(l1()) << " l_mc=" << LINT(l2());
    break;
  case O3PART_GCO1:
    os << "GCO1 h=" << IINT(i1()) << " m=" << IINT(i2())
       << " c=" << IINT(i3())
       << " l_hm=" << LINT(l1()) << " l_mc=" << LINT(l2());
    break;
  case O3PART_GCO2:
    os << "GCO2 h=" << IINT(i1()) << " m=" << IINT(i2())
       << " c=" << IINT(i3())
       << " l_hm=" << LINT(l1()) << " l_mc=" << LINT(l2());
    break;
  case O3PART_GCIS:
    os << "GCIS h=" << IINT(i1()) << " m=" << IINT(i2())
       << " c=" << IINT(i3()) << " s=" << IINT(i4())
       << " l_hm=" << LINT(l1()) << " l_mc=" << LINT(l2())
       << " l_ms=" << LINT(l3());
    break;
  case O3PART_GCOS:
    os << "GCOS h=" << IINT(i1()) << " m=" << IINT(i2())
       << " c=" << IINT(i3()) << " s=" << IINT(i4())
       << " l_hm=" << LINT(l1()) << " l_mc=" << LINT(l2())
       << " l_ms=" << LINT(l3());
    break;
  case O3PART_SIBS:
    os << "SIBS h=" << IINT(i1()) << " m=" << IINT(i2())
       << " c=" << IINT(i3()) << " s=" << IINT(i4())
       << " l_hm=" << LINT(l1()) << " l_hc=" << LINT(l2())
       << " l_hs=" << LINT(l3());
    break;
  default:
    cerr << "o3part.tostring(): unknown type " << type() << endl;
    assert(0);
  }
  return os.str();
}
#undef IINT
#undef LINT
