#include <assert.h>

/* walk a tree and accumulate scores */
template<class PartScorer> double
o3tree::score(const bool multiroot, const bool sibparser,
              const o3tree* const restrict root,
              PartScorer& restrict S) {
  assert(root->nlc() == 0);
  assert(root->idx() == -1);
  double score = 0;
  const int nr = root->nrc();
  if(multiroot) { assert(nr >= 1); }
  else          { assert(nr == 1); }

  idx_t last_r1 = IINV, last_r2 = IINV; /* previous root (x2) */
  lab_t last_l1 = LINV, last_l2 = LINV; /* previous label (x2) */
  for(int i = 0; i < nr; ++i) {
    const o3tree* const c = root->rc(i);
    const idx_t r = ICONV(c->idx());
    const lab_t l = LCONV(c->label());
    if(multiroot) {
      /* outermost dependency */
      /* if(i == (nr - 1)) { score += S.odep(-1, r, l); } */
      if(sibparser) {
        score += (i > 1
                  /* 3 or more dependencies */
                  ? (S.dep3(INULL, r, l) +
                     S.sib2(INULL, r, last_r1, l, last_l1) +
                     S.sibs(INULL, r, last_r1, last_r2, l, last_l1, last_l2))
                  : (i == 1
                     /* first 2 dependencies */
                     ? (S.dep2(INULL, r, l) +
                        S.sib1(INULL, r, last_r1, l, last_l1) +
                        S.sibs(INULL, r, last_r1, INULL, l, last_l1, LNULL))
                     /* first dependency */
                     : S.dep1(INULL, r, l)));
      } else {
        score += (i > 0
                  /* 2 or more dependencies */
                  ? (S.dep2(INULL, r, l) +
                     S.sib1(INULL, r, last_r1, l, last_l1))
                  /* first dependency */
                  : S.dep1(INULL, r, l));
      }
    } else {
      score += S.dep1(INULL, r, l);
    }
    score += (sibparser ? walk_sib(INULL, c, S) : walk_gch(INULL, c, S));
    last_r2 = last_r1; last_l2 = last_l1;
    last_r1 = r;       last_l1 = l;
  }

  return score;
}


template<class PartScorer>
double o3tree::walk_gch(const idx_t g,
                        const o3tree* const restrict node,
                        PartScorer& restrict S) {
  const idx_t h    = ICONV(node->idx());
  const lab_t l_gh = LCONV(node->label());
  const int nl     = node->nlc();
  const int nr     = node->nrc();
  const bool gbefh = (g == INULL || g < h);
  assert(g != h);
  double score = 0;
  if(nl == 0) {
    if(gbefh) { score += S.gci1(g, h, INULL, l_gh, LNULL); }
    else      { score += S.gco1(g, h, INULL, l_gh, LNULL); }
  }
  if(nr == 0) {
    if(gbefh) { score += S.gco1(g, h, INULL, l_gh, LNULL); }
    else      { score += S.gci1(g, h, INULL, l_gh, LNULL); }
  }
  idx_t last_m = IINV;
  lab_t last_l = LINV;
  for(int i = 0; i < nl; ++i) {
    /* outermost dependency */
    // if(i == (nl - 1)) {
    //   score += S.odep(h, m, l);
    //   score += (gbefh ? S.ogci(g, h, m, l_gh, l)
    //                   : S.ogco(g, h, m, l_gh, l));
    // }
    const o3tree* const c = node->lc(i);
    const idx_t m = ICONV(c->idx());
    const lab_t l = LCONV(c->label());
    score += (i == 0
              /* first dependency */
              ? S.dep1(h, m, l)
              /* 2 or more dependencies */
              : (S.dep2(h, m, l) +
                 S.sib1(h, m, last_m, l, last_l)));
    score += (gbefh
              /* left grandparent */
              ? (i == 0
                 /* first dependency */
                 ? (S.gci1(g, h, m, l_gh, l) +
                    S.gcis(g, h, m, INULL, l_gh, l, LNULL))
                 /* 2 or more dependencies */
                 : (S.gci2(g, h, m, l_gh, l) +
                    S.gcis(g, h, m, last_m, l_gh, l, last_l)))
              /* right grandparent */
              : (i == 0
                 /* first dependency */
                 ? (S.gco1(g, h, m, l_gh, l) +
                    S.gcos(g, h, m, INULL, l_gh, l, LNULL))
                 /* 2 or more dependencies */
                 : (S.gco2(g, h, m, l_gh, l) +
                    S.gcos(g, h, m, last_m, l_gh, l, last_l))));
    score += walk_gch(h, c, S);
    last_m = m; last_l = l;
  }
  last_m = IINV; last_l = LINV;
  for(int i = 0; i < nr; ++i) {
    /* outermost dependency */
    // if(i == (nr - 1)) {
    //   score += S.odep(h, m, l);
    //   score += (gbefh ? S.ogco(g, h, m, l_gh, l)
    //                   : S.ogci(g, h, m, l_gh, l));
    // }
    const o3tree* const c = node->rc(i);
    const idx_t m = ICONV(c->idx());
    const lab_t l = LCONV(c->label());
    score += (i == 0
              /* first dependency */
              ? S.dep1(h, m, l)
              /* 2 or more dependencies */
              : (S.dep2(h, m, l) +
                 S.sib1(h, m, last_m, l, last_l)));
    score += (gbefh
              /* left grandparent */
              ? (i == 0
                 /* first dependency */
                 ? (S.gco1(g, h, m, l_gh, l) +
                    S.gcos(g, h, m, INULL, l_gh, l, LNULL))
                 /* 2 or more dependencies */
                 : (S.gco2(g, h, m, l_gh, l) +
                    S.gcos(g, h, m, last_m, l_gh, l, last_l)))
              /* right grandparent */
              : (i == 0
                 /* first dependency */
                 ? (S.gci1(g, h, m, l_gh, l) +
                    S.gcis(g, h, m, INULL, l_gh, l, LNULL))
                 /* 2 or more dependencies */
                 : (S.gci2(g, h, m, l_gh, l) +
                    S.gcis(g, h, m, last_m, l_gh, l, last_l))));
    score += walk_gch(h, c, S);
    last_m = m; last_l = l;
  }
  return score;
}


template<class PartScorer>
double o3tree::walk_sib(const idx_t g,
                        const o3tree* const restrict node,
                        PartScorer& restrict S) {
  const idx_t h    = ICONV(node->idx());
  const lab_t l_gh = LCONV(node->label());
  const int nl     = node->nlc();
  const int nr     = node->nrc();
  const bool gbefh = (g == INULL || g < h);
  assert(g != h);
  double score = 0;
  if(nl == 0) {
    if(gbefh) { score += S.gci1(g, h, INULL, l_gh, LNULL); }
    else      { score += S.gco1(g, h, INULL, l_gh, LNULL); }
  }
  if(nr == 0) {
    if(gbefh) { score += S.gco1(g, h, INULL, l_gh, LNULL); }
    else      { score += S.gci1(g, h, INULL, l_gh, LNULL); }
  }
  idx_t m1 = IINV, m2 = IINV;
  lab_t l1 = LINV, l2 = LINV;
  for(int i = 0; i < nl; ++i) {
    const o3tree* const c = node->lc(i);
    const idx_t m = ICONV(c->idx());
    const lab_t l = LCONV(c->label());
    score += (i > 1
              /* 3 or more dependencies */
              ? (S.dep3(h, m, l) +
                 S.sib2(h, m, m1, l, l1) +
                 S.sibs(h, m, m1, m2, l, l1, l2))
              : (i == 1
                 /* first 2 dependencies */
                 ? (S.dep2(h, m, l) +
                    S.sib1(h, m, m1, l, l1) +
                    S.sibs(h, m, m1, INULL, l, l1, LNULL))
                 /* first dependency */
                 : S.dep1(h, m, l)));
    if(i == (nl - 1)) { /* outermost dependency */
      score += (gbefh
                /* left grandparent */
                ? (i == 0
                   /* only one dependency */
                   ? (S.gci1(g, h, m, l_gh, l) +
                      S.gci2(g, h, INULL, l_gh, LNULL) +
                      S.gcis(g, h, m, INULL, l_gh, l, LNULL))
                   /* 2 or more dependencies */
                   : (S.gci1(g, h, m, l_gh, l) +
                      S.gci2(g, h, m1, l_gh, l1) +
                      S.gcis(g, h, m, m1, l_gh, l, l1)))
                /* right grandparent */
                : (i == 0
                   /* only one dependency */
                   ? (S.gco1(g, h, m, l_gh, l) +
                      S.gco2(g, h, INULL, l_gh, LNULL) +
                      S.gcos(g, h, m, INULL, l_gh, l, LNULL))
                   /* 2 or more dependencies */
                   : (S.gco1(g, h, m, l_gh, l) +
                      S.gco2(g, h, m1, l_gh, l1) +
                      S.gcos(g, h, m, m1, l_gh, l, l1))));
    }
    score += walk_sib(h, c, S);
    m2 = m1; l2 = l1;
    m1 = m;  l1 = l;
  }
  m2 = m1 = IINV;
  l2 = l1 = LINV;
  for(int i = 0; i < nr; ++i) {
    const o3tree* const c = node->rc(i);
    const idx_t m = ICONV(c->idx());
    const lab_t l = LCONV(c->label());
    score += (i > 1
              /* 3 or more dependencies */
              ? (S.dep3(h, m, l) +
                 S.sib2(h, m, m1, l, l1) +
                 S.sibs(h, m, m1, m2, l, l1, l2))
              : (i == 1
                 /* first 2 dependencies */
                 ? (S.dep2(h, m, l) +
                    S.sib1(h, m, m1, l, l1) +
                    S.sibs(h, m, m1, INULL, l, l1, LNULL))
                 /* first dependency */
                 : S.dep1(h, m, l)));
    if(i == (nr - 1)) { /* outermost dependency */
      score += (gbefh
                /* left grandparent */
                ? (i == 0
                   /* only one dependency */
                   ? (S.gco1(g, h, m, l_gh, l) +
                      S.gco2(g, h, INULL, l_gh, LNULL) +
                      S.gcos(g, h, m, INULL, l_gh, l, LNULL))
                   /* 2 or more dependencies */
                   : (S.gco1(g, h, m, l_gh, l) +
                      S.gco2(g, h, m1, l_gh, l1) +
                      S.gcos(g, h, m, m1, l_gh, l, l1)))
                /* right grandparent */
                : (i == 0
                   /* only one dependency */
                   ? (S.gci1(g, h, m, l_gh, l) +
                      S.gci2(g, h, INULL, l_gh, LNULL) +
                      S.gcis(g, h, m, INULL, l_gh, l, LNULL))
                   /* 2 or more dependencies */
                   : (S.gci1(g, h, m, l_gh, l) +
                      S.gci2(g, h, m1, l_gh, l1) +
                      S.gcis(g, h, m, m1, l_gh, l, l1))));
    }
    score += walk_sib(h, c, S);
    m2 = m1; l2 = l1;
    m1 = m;  l1 = l;
  }
  return score;
}
