#include "PartScoreCache.h"

#include <sstream>
#include <iomanip>

using std::endl;
using std::fixed;
using std::setprecision;
using std::ostringstream;

#ifdef CAUTIOUS
part_score_cache_base::part_score_cache_base()
  : _Ncalls_dep1(0), _Nhits_dep1(0),
    _Ncalls_dep2(0), _Nhits_dep2(0),
    _Ncalls_dep3(0), _Nhits_dep3(0),
    _Ncalls_sib1(0), _Nhits_sib1(0),
    _Ncalls_sib2(0), _Nhits_sib2(0),
    _Ncalls_gci1(0), _Nhits_gci1(0),
    _Ncalls_gci2(0), _Nhits_gci2(0),
    _Ncalls_gco1(0), _Nhits_gco1(0),
    _Ncalls_gco2(0), _Nhits_gco2(0),
    _Ncalls_gcis(0), _Nhits_gcis(0),
    _Ncalls_gcos(0), _Nhits_gcos(0),
    _Ncalls_sibs(0), _Nhits_sibs(0) {}
#endif /* CAUTIOUS */

string part_score_cache_base::report() const {
  ostringstream os;
#ifdef CAUTIOUS
#define REP(type)                               \
  os << "  " #type " "                          \
     << _Nhits_ ## type << " / "                \
     << _Ncalls_ ## type << " = "               \
     << fixed << setprecision(2)                \
     << 100.0*(((double)(_Nhits_ ## type))/     \
               ((double)(_Ncalls_ ## type)))    \
     << endl

  REP(dep1);
  REP(dep2);
  REP(dep3);
  REP(sib1);
  REP(sib2);
  REP(gci1);
  REP(gci2);
  REP(gco1);
  REP(gco2);
  REP(gcis);
  REP(gcos);
  REP(sibs);

#undef REP
#endif /* CAUTIOUS */
  return os.str();
}

#ifdef CAUTIOUS
#define CLEAR(type) _Nhits_ ## type = _Ncalls_ ## type = 0
void part_score_cache_base::clear_counts() const {
  CLEAR(dep1);
  CLEAR(dep2);
  CLEAR(dep3);
  CLEAR(sib1);
  CLEAR(sib2);
  CLEAR(gci1);
  CLEAR(gci2);
  CLEAR(gco1);
  CLEAR(gco2);
  CLEAR(gcis);
  CLEAR(gcos);
  CLEAR(sibs);
}
#undef CLEAR
#endif /* CAUTIOUS */
