#include "PartScoreRandom.h"

#include "O3Part.h"
#include "Hashtable.h"

#include <stdlib.h>


part_score_random::part_score_random(const int N, const int L,
                                     const bool sibparser,
                                     const double minsc,
                                     const double maxsc)
  : _test(N, L, sibparser),
    _min(minsc), _max(maxsc), _range(maxsc - minsc),
    _S(new o3part::dmap()) {}
part_score_random::~part_score_random() { delete _S; }

struct part_score_random::fac : public o3part {
  friend class part_score_random;
};


double part_score_random::
dep1(const idx_t h, const idx_t m, const lab_t l_hm) const {
  _test.dep1(h, m, l_hm);
  return randscore(fac::mkdep1(h, m, l_hm));
}
double part_score_random::
dep2(const idx_t h, const idx_t m, const lab_t l_hm) const {
  _test.dep2(h, m, l_hm);
  return randscore(fac::mkdep2(h, m, l_hm));
}
double part_score_random::
dep3(const idx_t h, const idx_t m, const lab_t l_hm) const {
  _test.dep3(h, m, l_hm);
  return randscore(fac::mkdep3(h, m, l_hm));
}

double part_score_random::
sib1(const idx_t h, const idx_t m, const idx_t c,
     const lab_t l_hm, const lab_t l_hc) const {
  _test.sib1(h, m, c, l_hm, l_hc);
  return randscore(fac::mksib1(h, m, c, l_hm, l_hc));
}
double part_score_random::
sib2(const idx_t h, const idx_t m, const idx_t c,
     const lab_t l_hm, const lab_t l_hc) const {
  _test.sib2(h, m, c, l_hm, l_hc);
  return randscore(fac::mksib2(h, m, c, l_hm, l_hc));
}

double part_score_random::
gci1(const idx_t h, const idx_t m, const idx_t c,
     const lab_t l_hm, const lab_t l_mc) const {
  _test.gci1(h, m, c, l_hm, l_mc);
  return randscore(fac::mkgci1(h, m, c, l_hm, l_mc));
}
double part_score_random::
gci2(const idx_t h, const idx_t m, const idx_t c,
     const lab_t l_hm, const lab_t l_mc) const {
  _test.gci2(h, m, c, l_hm, l_mc);
  return randscore(fac::mkgci2(h, m, c, l_hm, l_mc));
}

double part_score_random::
gco1(const idx_t h, const idx_t m, const idx_t c,
     const lab_t l_hm, const lab_t l_mc) const {
  _test.gco1(h, m, c, l_hm, l_mc);
  return randscore(fac::mkgco1(h, m, c, l_hm, l_mc));
}
double part_score_random::
gco2(const idx_t h, const idx_t m, const idx_t c,
     const lab_t l_hm, const lab_t l_mc) const {
  _test.gco2(h, m, c, l_hm, l_mc);
  return randscore(fac::mkgco2(h, m, c, l_hm, l_mc));
}

double part_score_random::
gcis(const idx_t h, const idx_t m, const idx_t c, const idx_t s,
     const lab_t l_hm, const lab_t l_mc, const lab_t l_ms) const {
  _test.gcis(h, m, c, s, l_hm, l_mc, l_ms);
  return randscore(fac::mkgcis(h, m, c, s, l_hm, l_mc, l_ms));
}

double part_score_random::
gcos(const idx_t h, const idx_t m, const idx_t c, const idx_t s,
     const lab_t l_hm, const lab_t l_mc, const lab_t l_ms) const {
  _test.gcos(h, m, c, s, l_hm, l_mc, l_ms);
  return randscore(fac::mkgcos(h, m, c, s, l_hm, l_mc, l_ms));
}

double part_score_random::
sibs(const idx_t h, const idx_t m, const idx_t c, const idx_t s,
     const lab_t l_hm, const lab_t l_hc, const lab_t l_hs) const {
  _test.sibs(h, m, c, s, l_hm, l_hc, l_hs);
  return randscore(fac::mksibs(h, m, c, s, l_hm, l_hc, l_hs));
}


/* generate a random score and return it */
double part_score_random::
randscore(const o3part& restrict p) const {
  /* insert a hopefully invalid default value */
  const double def = _min - 1e300;
  double& restrict d = _S->mget(p, def);
  if(d == def) {
    d = (_min + _range*((double)random())/((double)RAND_MAX));
  }
  return d;
}
