#include "PartScoreTree.h"

part_score_tree::part_score_tree(const int N, const int L,
                                 const bool sibparser,
                                 const vector<int>& restrict h,
                                 const vector<int>& restrict l)
  :  _test(N, L, sibparser), _h(h), _l(l) {
  assert(N == (int)_h.size());
  assert(N == (int)_l.size());
  /* convert to nulls */
  for(int i = 0; i < N; ++i) {
    if(_h[i] == -1) { _h[i] = INULL; }
  }
}
part_score_tree::~part_score_tree() {}

double part_score_tree::
dep1(const idx_t h, const idx_t m, const lab_t l_hm) const {
  return _test.dep1(h, m, l_hm) +
    ((_h[m] == h && _l[m] == l_hm) ? 1 : 0);
}
double part_score_tree::
dep2(const idx_t h, const idx_t m, const lab_t l_hm) const {
  return _test.dep2(h, m, l_hm) +
    ((_h[m] == h && _l[m] == l_hm) ? 1 : 0);
}
double part_score_tree::
dep3(const idx_t h, const idx_t m, const lab_t l_hm) const {
  return _test.dep3(h, m, l_hm) +
    ((_h[m] == h && _l[m] == l_hm) ? 1 : 0);
}

double part_score_tree::
sib1(const idx_t h, const idx_t m, const idx_t c,
     const lab_t l_hm, const lab_t l_hc) const {
  return _test.sib1(h, m, c, l_hm, l_hc) +
    (((_h[m] == h && _l[m] == l_hm) &&
      (_h[c] == h && _l[c] == l_hc))
     ? 1 : 0);
}
double part_score_tree::
sib2(const idx_t h, const idx_t m, const idx_t c,
     const lab_t l_hm, const lab_t l_hc) const {
  return _test.sib2(h, m, c, l_hm, l_hc) +
    (((_h[m] == h && _l[m] == l_hm) &&
      (_h[c] == h && _l[c] == l_hc))
     ? 1 : 0);
}

double part_score_tree::
gci1(const idx_t h, const idx_t m, const idx_t c,
     const lab_t l_hm, const lab_t l_mc) const {
  return _test.gci1(h, m, c, l_hm, l_mc) +
    (((_h[m] == h && _l[m] == l_hm) &&
      (c == INULL || (_h[c] == m && _l[c] == l_mc)))
     ? 1 : 0);
}
double part_score_tree::
gci2(const idx_t h, const idx_t m, const idx_t c,
     const lab_t l_hm, const lab_t l_mc) const {
  return _test.gci2(h, m, c, l_hm, l_mc) +
    (((_h[m] == h && _l[m] == l_hm) &&
      (c == INULL || (_h[c] == m && _l[c] == l_mc)))
     ? 1 : 0);
}

double part_score_tree::
gco1(const idx_t h, const idx_t m, const idx_t c,
     const lab_t l_hm, const lab_t l_mc) const {
  return _test.gco1(h, m, c, l_hm, l_mc) +
    (((_h[m] == h && _l[m] == l_hm) &&
      (c == INULL || (_h[c] == m && _l[c] == l_mc)))
     ? 1 : 0);
}
double part_score_tree::
gco2(const idx_t h, const idx_t m, const idx_t c,
     const lab_t l_hm, const lab_t l_mc) const {
  return _test.gco2(h, m, c, l_hm, l_mc) +
    (((_h[m] == h && _l[m] == l_hm) &&
      (c == INULL || (_h[c] == m && _l[c] == l_mc)))
     ? 1 : 0);
}

double part_score_tree::
gcis(const idx_t h, const idx_t m, const idx_t c, const idx_t s,
     const lab_t l_hm, const lab_t l_mc, const lab_t l_ms) const {
  return _test.gcis(h, m, c, s, l_hm, l_mc, l_ms) +
    (((_h[m] == h && _l[m] == l_hm) &&
      (_h[c] == m && _l[c] == l_mc) &&
      (_h[s] == m && _l[s] == l_ms))
     ? 1 : 0);
}

double part_score_tree::
gcos(const idx_t h, const idx_t m, const idx_t c, const idx_t s,
     const lab_t l_hm, const lab_t l_mc, const lab_t l_ms) const {
  return _test.gcos(h, m, c, s, l_hm, l_mc, l_ms) +
    (((_h[m] == h && _l[m] == l_hm) &&
      (_h[c] == m && _l[c] == l_mc) &&
      (_h[s] == m && _l[s] == l_ms))
     ? 1 : 0);
}

double part_score_tree::
sibs(const idx_t h, const idx_t m, const idx_t c, const idx_t s,
     const lab_t l_hm, const lab_t l_hc, const lab_t l_hs) const {
  return _test.sibs(h, m, c, s, l_hm, l_hc, l_hs) +
    (((_h[m] == h && _l[m] == l_hm) &&
      (_h[c] == h && _l[c] == l_hc) &&
      (_h[s] == h && _l[s] == l_hs))
     ? 1 : 0);
}
