#include "PartTest.h"

#include "Bits.h"

#include <assert.h>


/* lifecycle */
part_test::part_test(const int N, const int L, const bool sibparser)
  : _N((unsigned int)N), _L((unsigned int)L), _sibparser(sibparser) {
  assert(N >= 0); assert(L >= 0);
}
part_test::~part_test() {}


/* inlined part-scorers */
double part_test::
dep1(const idx_t h, const idx_t m, const lab_t l_hm) const {
  assert(h != m);
  assert(h == INULL || h < _N);
  assert(m < _N);
  assert(l_hm < _L);
  return 0;
}
double part_test::
dep2(const idx_t h, const idx_t m, const lab_t l_hm) const {
  return dep1(h, m, l_hm);
}
double part_test::
dep3(const idx_t h, const idx_t m, const lab_t l_hm) const {
  assert(_sibparser);
  return dep1(h, m, l_hm);
}

double part_test::
sib1(const idx_t h, const idx_t m, const idx_t c,
     const lab_t l_hm, const lab_t l_hc) const {
  assert(h != m);
  assert(h == INULL || h < _N);
  assert(m < _N);
  assert(l_hm < _L);
  assert(l_hc < _L);
  if(h == INULL) { assert(c < m); }
  else if(h < m) { assert(h < c); assert(c < m); }
  else           { assert(h > c); assert(c > m); }
  return 0;
}
double part_test::
sib2(const idx_t h, const idx_t m, const idx_t c,
     const lab_t l_hm, const lab_t l_hc) const {
  assert(_sibparser);
  return sib1(h, m, c, l_hm, l_hc);
}

double part_test::
gci1(const idx_t h, const idx_t m, const idx_t c,
     const lab_t l_hm, const lab_t l_mc) const {
  assert(h != m);
  assert(h == INULL || h < _N);
  assert(m < _N);
  assert(l_hm < _L);
  if(c != INULL) {
    assert(c < _N);
    assert(l_mc < _L);
    if(h == INULL) { assert(c < m); }
    else if(h < m) { assert(h < c); assert(c < m); }
    else           { assert(h > c); assert(c > m); }
  }
  return 0;
}
double part_test::
gci2(const idx_t h, const idx_t m, const idx_t c,
     const lab_t l_hm, const lab_t l_mc) const {
  if(!_sibparser) { assert(c != INULL); }
  return gci1(h, m, c, l_hm, l_mc);
}

double part_test::
gco1(const idx_t h, const idx_t m, const idx_t c,
     const lab_t l_hm, const lab_t l_mc) const {
  assert(h != m);
  assert(h == INULL || h < _N);
  assert(m < _N);
  assert(l_hm < _L);
  if(c != INULL) {
    assert(c < _N);
    assert(l_mc < _L);
    if(h == INULL || h < m) { assert(m < c); }
    else                    { assert(m > c); }
  }
  return 0;
}
double part_test::
gco2(const idx_t h, const idx_t m, const idx_t c,
     const lab_t l_hm, const lab_t l_mc) const {
  if(!_sibparser) { assert(c != INULL); }
  return gco1(h, m, c, l_hm, l_mc);
}

double part_test::
gcis(const idx_t h, const idx_t m, const idx_t c, const idx_t s,
     const lab_t l_hm, const lab_t l_mc, const lab_t l_ms) const {
  assert(h != m);
  assert(h == INULL || h < _N);
  assert(m < _N);
  assert(c < _N);
  assert(s == INULL || s < _N);
  assert(l_hm < _L);
  assert(l_mc < _L);
  if(s != INULL) {
    assert(l_ms < _L);
    if(h == INULL) { assert(c < s); assert(s < m); }
    else if(h < m) { assert(h < c); assert(c < s); assert(s < m); }
    else           { assert(h > c); assert(c > s); assert(s > m); }
  } else {
    if(h == INULL) { assert(c < m); }
    else if(h < m) { assert(h < c); assert(c < m); }
    else           { assert(h > c); assert(c > m); }
  }
  return 0;
}

double part_test::
gcos(const idx_t h, const idx_t m, const idx_t c, const idx_t s,
     const lab_t l_hm, const lab_t l_mc, const lab_t l_ms) const {
  assert(h != m);
  assert(h == INULL || h < _N);
  assert(m < _N);
  assert(c < _N);
  assert(s == INULL || s < _N);
  assert(l_hm < _L);
  assert(l_mc < _L);
  if(s != INULL) {
    assert(l_ms < _L);
    if(h == INULL || h < m) { assert(m < s); assert(s < c); }
    else                    { assert(m > s); assert(s > c); }
  } else {
    if(h == INULL || h < m) { assert(m < c); }
    else                    { assert(m > c); }
  }
  return 0;
}

double part_test::
sibs(const idx_t h, const idx_t m, const idx_t c, const idx_t s,
     const lab_t l_hm, const lab_t l_hc, const lab_t l_hs) const {
  assert(_sibparser);
  assert(h != m);
  assert(h == INULL || h < _N);
  assert(m < _N);
  assert(c < _N);
  assert(s == INULL || s < _N);
  assert(l_hm < _L);
  assert(l_hc < _L);
  if(s != INULL) {
    assert(l_hs < _L);
    if(h == INULL) { assert(s < c); assert(c < m); }
    else if(h < m) { assert(h < s); assert(s < c); assert(c < m); }
    else           { assert(h > s); assert(s > c); assert(c > m); }
  } else {
    if(h == INULL) { assert(c < m); }
    else if(h < m) { assert(h < c); assert(c < m); }
    else           { assert(h > c); assert(c > m); }
  }
  return 0;
}

