/* -*-c++-*- */
#ifndef TIMER_H
#define TIMER_H
/************************************************************
 *
 *   Timer : utilities for timing runs
 *
 *   Author : Terry Koo
 *            maestro@csail.mit.edu
 *
 ************************************************************/


#include <time.h>
#include <sys/time.h>
#include <assert.h>

/* macro that starts a timer (also allocates local vars) */
#define timer_start()				\
  struct timeval t0;				\
  gettimeofday(&t0, NULL)
/* macro that stores elapsed seconds in ret */
#define timer_split(ret) {						\
    struct timeval tvsplit;						\
    gettimeofday(&tvsplit, NULL);					\
    ret = (double)tvsplit.tv_sec   - (double)t0.tv_sec +		\
      ((double)tvsplit.tv_usec - (double)t0.tv_usec)/1000000.0;         \
  }
/* macro that stores elapsed seconds in timediff */
#define timer_stop()						\
  struct timeval tv2;						\
  gettimeofday(&tv2, NULL);					\
  const double timediff =					\
    (double)tv2.tv_sec   - (double)t0.tv_sec +			\
    ((double)tv2.tv_usec - (double)t0.tv_usec)/1000000.0;	\
  assert(timediff > 0)


#define ticker64_start()                                        \
  const int ticker64_mask = (0x3f << TICKER_REPINT_LOG);	\
  timer_start()

#define ticker64(ii, nn) 						\
  if(((ii) & TICKER_REPINT) == 0) {                                     \
    if(((ii) & ticker64_mask) == 0) {					\
      cerr << "\r             [                                "	\
	"                                ]\r";				\
      /* print percentages and ETC */					\
      const double ratio = (double)(ii)/(double)(nn);			\
      int percent = (int)(100.0*ratio);					\
      if(percent >= 100) { percent = 99; }				\
      char buf[16];							\
      sprintf(buf, "%2d%% ", percent);					\
      cerr << buf;							\
      if(ratio != 0) {							\
	double split = 0;						\
	timer_split(split);						\
	cerr << timestr((1 - ratio)*(split/ratio));                     \
      } else {								\
	cerr << "??:??:??";						\
      }									\
      cerr << " [";							\
    }									\
    cerr << '.' << flush;						\
  }

#define ticker64_finish(nn) {                   \
    double split;                               \
    timer_split(split);                         \
    cerr << "\rEND " << timestr(split) << endl; \
  }

/* variables controlling how often ticks are printed */
extern int TICKER_REPINT_LOG;
extern int TICKER_REPINT;
/* adjust the reporting intervals by 2^{log_factor} (both positive and
   negative adjustments possible) */
void ticker_adjust_repint(int log_factor);
/* pretty-print a time interval */
const char* timestr(double seconds);

#endif /* TIMER_H */
