/* -*-c++-*- */
#ifndef TREE_FUNCTORS_H
#define TREE_FUNCTORS_H
/* TreeFunctors : Utility classes for implementing brute-force
   inference. */

#include "O3Part.h"
#include "MathUtil.h"

#include <vector>

using std::vector;

/* functor that memorizes the best tree */
class treef_argmax {
public:
  treef_argmax();
  ~treef_argmax();

  void operator()(const vector<int>& restrict p,
                  const vector<int>& restrict l,
                  const o3part::vec& restrict y,
                  const double score);
  const vector<int>& best_p() const;
  const vector<int>& best_l() const;
  double best_sc() const;

private:
  double _score;
  vector<int> _p;
  vector<int> _l;
};


/* functor that log-adds all tree scores */
class treef_partition {
public:
  treef_partition();
  ~treef_partition();

  void operator()(const vector<int>& restrict p,
                  const vector<int>& restrict l,
                  const o3part::vec& restrict y,
                  const double score);
  double logZ() const;

private:
  double _logZ;
};


/* functor that stores an array of (non-logspace) marginals */
class treef_marginals {
public:
  treef_marginals(o3part::dmap& restrict M);
  ~treef_marginals();

  void operator()(const vector<int>& restrict p,
                  const vector<int>& restrict l,
                  const o3part::vec& restrict y,
                  const double score);
  void mkprobs();
  double logZ() const;

private:
  double _logZ;
  o3part::dmap& restrict _M;
};



/*** inlined implementations ***/

/* argmax tree-functor */
inline const vector<int>& treef_argmax::best_p() const { return _p; }
inline const vector<int>& treef_argmax::best_l() const { return _l; }
inline double treef_argmax::best_sc() const            { return _score; }
inline void treef_argmax::operator()(const vector<int>& restrict p,
                                     const vector<int>& restrict l,
                                     const o3part::vec& restrict y,
                                     const double score) {
  if(_p.empty() || _score < score) { _p = p; _l = l; _score = score; }
  assert(!_p.empty()); assert(!_l.empty());
}

/* partition-function tree-functor */
inline double treef_partition::logZ() const { return _logZ; }
inline void treef_partition::operator()(const vector<int>& restrict p,
                                        const vector<int>& restrict l,
                                        const o3part::vec& restrict y,
                                        const double score) {
  _logZ = log_add(_logZ, score);
}

/* marginals tree-functor */
inline double treef_marginals::logZ() const { return _logZ; }

#endif /* TREE_FUNCTORS_H */
