#!/usr/bin/perl -w

if(scalar(@ARGV) != 3) {
    print STDERR <<END;
usage: $0 <stdin-col> <file> <file-col>
    replace column number <stdin-col> of stdin with column number
    <file-col> of <file>, printing the result to stdout.  column
    numbers are counted 1-origin.  blank (whitespace-only) lines of
    stdin or <file> are ignored as long as they are aligned with each
    other.  <file> may be gzipped as long as it has the ".gz" suffix.
END
    exit 1;
}

my $icol = shift;
my $fname = shift;
my $fcol = shift;

# convert to 0-origin
--$icol;
--$fcol;

open(FF, ($fname =~ m/\.gz$/io ? "gunzip -c '$fname' |" : "< $fname"))
    or die "couldn't open '$fname'";
while(my $iline = <STDIN>) {
    my $fline = <FF>;
    die "file ended early" if(!defined($fline));
    my @ielts = split(' ', $iline);
    my @felts = split(' ', $fline);
    # check blank lines
    if((scalar(@ielts) == 0) ||
       (scalar(@felts) == 0)) {
        die "blank lines not aligned" if((scalar(@ielts) != 0) ||
                                         (scalar(@felts) != 0));
        print "\n"; next;
    }
    die "not enough columns in stdin" if($icol >= scalar(@ielts));
    die "not enough columns in file"  if($fcol >= scalar(@felts));
    $ielts[$icol] = $felts[$fcol];
    print join("\t", @ielts)."\n";
}
close(FF);
