#include "O3Perceptron.h"
#include "O3MMSSGD.h"
#include "O3Decoder.h"
#include "O3AuxPerc.h"
#include "Macros.h"

#include <stdlib.h>
#include <assert.h>

#include <iostream>
#include <string>
#include <exception>

using std::exception;
using std::cerr;
using std::endl;

int main(const int argc, const char* argv[]) {
  cerr << "dpo3 v" VERSION "  Copyright (C) 2010  Terry Y. Koo" << endl
       << "This program comes with ABSOLUTELY NO WARRANTY.  This is " << endl
       << "free software and you are welcome to redistribute it." << endl
       << endl;

  if(argc < 2) { /* basic options: the first arg is the training alg */
    cerr << "usage: " << argv[0] << " (perc|mmssgd) [trainer-options]"
         << endl << endl;
    exit(1);
  }

  /* the training algorithm */
  const string alg(argv[1]);
  vector<string> args;
  { /* populate args with command-line args */
    args.push_back(argv[0]);
    for(int i = 2; i < argc; ++i) { args.push_back(argv[i]); }
  }

  /* dispatch based on training algorithm */
  o3trainer* T = NULL;
  if(alg == "perc") {
    args.push_back("--params=avg");
    T = new o3perceptron();
  } else if(alg == "mmssgd") {
    args.push_back("--params=lin");
    T = new o3mmssgd();
  } else if(alg == "decoder") {
    args.push_back("--params=avg");
    T = new o3decoder();
  } else if(alg == "auxperc") {
    args.push_back("--params=avg");
    T = new o3auxperc();
  } else {
    cerr << "unknown training algorithm: " << alg << endl; exit(1);
  }

  /* invoke the trainer (with exception-handling) */
  try {
    T->init(args);
    T->train();
  } catch(exception& e) {
    cerr << "dpo3 : exception:" << e.what() << endl;
    assert(0);
  } catch(...) {
    cerr << "dpo3 : unknown throw()" << endl;
    assert(0);
  }

  /* clean up */
  delete T;
  return 0;
}
