#!/bin/bash

SCRIPTDIR=$(dirname $0)
EXEC=$SCRIPTDIR/dpo3
DPDIR=/data/nlp4/maestro/dp
MODEL=$DPDIR/models/dpo3

VERSION=$(grep VERSION $SCRIPTDIR/Macros.h | cut -d\" -f2)

paramsopt=--noparams
seedopt=--seed=1
ablateopt=
marginopt=
commaopt=
truthmargopt=
parallelopt=
timedopt=
COVERAGE=0
SMALLDATA=
TRUTHMARG=marginals
FULLPARSING=0
DEBUG=0
GRIND=0
JOBS=0
LABELED=0
RANDPERM=0
TESTDATA=0
TRIALS=
DECODE=
abeam=0.0001
while [ $# -ge 0 ] ; do
    if [ "$1" = '-gdb' ] ; then
        DEBUG=1
        EXEC=$EXEC-gdb
    elif [ "$1" = '-grind' ] ; then
        GRIND=1
        EXEC=$EXEC-gdb
    elif [ "$1" = '-opt' ] ; then
        EXEC=$EXEC-opt
    elif [ "$1" = '-cov' ] ; then
        COVERAGE=1
        SMALLDATA=1
    elif [ "$1" = '-test' ] ; then
        TESTDATA=1
    elif [ "$1" = '-perm' ] ; then
        RANDPERM=1
    elif [ "$1" = '-small' ] ; then
        SMALLDATA=1k
    elif [ "$1" = '-small100' ] ; then
        SMALLDATA=100
    elif [ "$1" = '-small10' ] ; then
        SMALLDATA=10
    elif [ "$1" = '-copt' ] ; then
        EXEC=$EXEC-copt
    elif [ "$1" = '-full' ] ; then
        FULLPARSING=1
    elif [ "$1" = '-commas' ] ; then
        commaopt=--fix-commas
    elif [ "$1" = '-labeled' ] ; then
        LABELED=1
    elif [ "$1" = '-save' ] ; then
        paramsopt=
    elif [ "$1" = '-rand' ] ; then
        seedopt=
    elif [ "$1" = '-truthmarg' ] ; then
        truthmargopt=--prune-train-truth
    elif echo $1 | grep -q '^-seed=' ; then
        seedopt="-"$1
    elif echo $1 | grep -q '^-ablate=' ; then
        ablateopt="-"$1
    elif echo $1 | grep -q '^-margin=' ; then
        marginopt="-"$1
    elif echo $1 | grep -q '^-decode=' ; then
        DECODE=$(echo $1 | cut -c 9-)
    elif echo $1 | grep -q '^-beam=' ; then
        abeam=$(echo $1 | cut -c 7-)
    elif echo $1 | grep -q '^-timed=' ; then
        timedopt=-$1
        TRIALS=$(echo $1 | cut -d= -f2)
    elif echo $1 | grep -q '^-j' ; then
        JOBS=$(echo $1 | cut -c 3-)
        RANDPERM=1
        parallelopt="--parallel=$JOBS"
    else
        if [ $# -lt 5 ] ; then
            echo "usage: $0 <tag> (perc|mmssgd<C>) <fconf> (sib|gch) (eng|cze|chi)"
            exit 1
        fi
        tag=$1
        alg=$2
        fconf=$3
        sibarg=$4
        lang=$5
        break
    fi
    shift
done

mropt=
if [ $lang = cze ] ; then mropt=--multiroot ; fi

if [ ! -z "$ablateopt" ] ; then
    tag=$tag-abl$(echo $ablateopt | cut -d= -f2)
else
    tag=$tag-ablnone
fi
if [ $COVERAGE -eq 1 ] ; then    tag=$tag-cov ; fi
if [ $FULLPARSING -eq 1 ] ; then tag=$tag-full ; fi

if [ ! -z "$commaopt" ] ; then
    tag=$tag-fixcommas
fi

if [ ! -z "$truthmargopt" ] ; then
    tag=$tag-tr$abeam
else
    tag=$tag-pr$abeam
fi

algname=
Copt=
if [ $alg = 'perc' ] ; then
    algname=perc
elif echo $alg | grep -q ^mmssgd ; then
    Copt="--C="$(echo $alg | cut -c7-)
    algname=mmssgd
else
    echo "unknown alg: $alg"
    exit 1
fi

L=1
if [ $LABELED -eq 1 ] ; then
    if [ $lang = cze ] ; then
        echo "czech dataset is unlabeled!"
        exit 1
    elif [ $lang = eng ] ; then
        L=12
    elif [ $lang = chi ] ; then
        L=69
    else
        echo "unknown language: $lang"
    fi
fi

sibopt=
if [ $sibarg = sib ] ; then sibopt=--sibparser ; fi


T=30
size=
sizetag=
if [ ! -z "$SMALLDATA" ] ; then
    size=$SMALLDATA
    sizetag=-$size
    T=20
fi

valtype=val
if [ $TESTDATA -eq 1 ] ; then valtype=test ; fi

trdata=$DPDIR/data/$lang/$lang.train$size.mapped
valdata=$DPDIR/data/$lang/$lang.$valtype.mapped
tadeps=$DPDIR/data/$lang/$lang.marginals.L1.010.train$size.gz
vadeps=$DPDIR/data/$lang/$lang.marginals.L1.010.$valtype.gz
ordering=unif
if [ $RANDPERM -eq 1 ] ; then ordering=perm ; fi

if [ $lang = chi ] ; then abeam=1.52587890625e-05 ; fi

mdir=$MODEL/$lang/$tag-$alg-$ordering-$sibarg-$fconf-L$L$sizetag-j$JOBS

relabelopt=--relabel
orig_mdir=$mdir
decodeopt=
if [ ! -z "$DECODE" ] ; then
    if [ ! -e $orig_mdir ] ; then
        echo "decoder requires training first!"
        exit 1
    fi
    decodeopt="--range=$DECODE --stem=labels --noparams --pdir=$orig_mdir"
    algname=decoder
    mdir=$orig_mdir/decode-$valtype$TRIALS
    relabelopt=
    ordering=det
    trdata=$valdata
    valdata=
    tadeps=$vadeps
    vadeps=
    rm -rf $mdir
fi
if [ ! -e $mdir ] ; then mkdir $mdir ; fi
for i in ftags.map f2ctag.map words.map wordftag.map taginfo wordinfo ; do
    rm -f $mdir/$i
    ln -s $DPDIR/data/$lang/$i $mdir/$i
done
rm -f $mdir/log*
# copy the source code to the dir
make dist && cp -i dpo3-$VERSION.tar.gz $mdir

if [ ! -e $EXEC ] ; then
    echo "couldn't find $EXEC"
    exit 1
fi

# set the screen window title!
echo -ne "\033k"$lang" "$(basename $mdir)"\033\134"

if [ $DEBUG -eq 1 ] ; then
    if [ -e $mdir/options ] ; then
        rm -f $mdir/run.gdb
        echo "set args $algname $mdir/options" >> $mdir/run.gdb
        echo "break main" >> $mdir/run.gdb
        echo "run" >> $mdir/run.gdb
        exec gdb $EXEC -x $mdir/run.gdb
    fi
fi

if [ $GRIND -eq 1 ] ; then
    if [ -e $mdir/options ] ; then
        exec valgrind --leak-check=yes $EXEC $algname $mdir/options
    fi
fi

pruneopts="--topdown --prune-tmarg=$tadeps --prune-beam=$abeam"
valdataopt=
if [ ! -z "$valdata" ] ; then
    valdataopt="--valdata=$valdata"
    pruneopts="$pruneopts --prune-vmarg=$vadeps"
fi
if [ $FULLPARSING -eq 1 ] ; then pruneopts= ; fi

if ! $EXEC $algname \
    --L=$L \
    --T=$T \
    --trdata=$trdata \
    $valdataopt \
    $pruneopts \
    --ordering=$ordering \
    --dir=$mdir \
    --fconf=$fconf \
    $truthmargopt \
    $parallelopt \
    $relabelopt \
    $ablateopt \
    $marginopt \
    $paramsopt \
    $decodeopt \
    $commaopt \
    $timedopt \
    $seedopt \
    $sibopt \
    $mropt \
    $Copt \
    2>&1 | tee -a $mdir/log ; then
    echo "dpo3 failed!"
    exit 1
fi
