#!/bin/bash

SCRIPTDIR=$(dirname $0)
SCORE=$SCRIPTDIR/scoresents.pl
DPDIR=/data/nlp4/maestro/dp
ENGORIG=$DPDIR/data/eng/eng.val.goldtags.original
CZEORIG=$DPDIR/data/cze/cze.val.original
CHIORIG=$DPDIR/data/chi/chi.val.original

ERRORS=0
for dir in "$@" ; do
    puncopt=
    valorig=$ENGORIG
    if cat $dir/options | grep valdata | grep -q cze ; then
        valorig=$CZEORIG
        puncopt=-p
    elif cat $dir/options | grep valdata | grep -q chi ; then
        valorig=$CHIORIG
        puncopt=-p
    fi
    # handle test data
    if cat $dir/options | grep valdata | grep -q test ; then
        valorig=$(echo $valorig | sed 's/\.val\./.test./')
    fi
    # grep mistakes
    cat $dir/log | grep mistakes | awk '{print $NF}' > $dir/mistakes
    for pre in max mbr val ; do
        for ty in .reg .avg '' ; do
            maxlab=$pre-labels$ty
            out=$dir/$pre-scores$ty
            files=($(echo $dir/$maxlab.[0-9][0-9][0-9]))
            if [ ! -e ${files[0]} ] ; then
                #echo "no $maxlab in $dir ...?"
                continue
            fi
            rm -f $out
            if ! $SCORE $puncopt $valorig \
                $dir/$maxlab.[0-9][0-9][0-9] > $out ; then
                echo "score $maxlab $dir failed!"
                rm -f $out
                ERRORS=1
                continue
            fi
            best=$dir/best-score-$pre$ty
            rm -f $best
            sort -nk2 $out | tail -1 > $best
            cat $best
        done
        # labeled scoring
        L=$(cat $dir/options | grep '^L[= ]' | cut -d\" -f2)
        if [ "$L" -le 0 ] ; then
            echo "problem extracting L (= $L) from $dir"
            exit 1
        fi
        if [ "$L" -gt 1 ] ; then
            for ty in .reg .avg '' ; do
                maxlab=$pre-labels$ty
                out=$dir/$pre-lscores$ty
                files=($(echo $dir/$maxlab.[0-9][0-9][0-9]))
                if [ ! -e ${files[0]} ] ; then
                    echo "no $maxlab in $dir ...?"
                    continue
                fi
                rm -f $out
                if ! $SCORE $puncopt -l $VALORIG \
                    $dir/$maxlab.[0-9][0-9][0-9] > $out ; then
                    echo "score $maxlab $dir failed!"
                    rm -f $out
                    ERRORS=1
                    continue
                fi
                best=$dir/best-lscore-$pre$ty
                rm -f $best
                sort -nk2 $out | tail -1 > $best
                cat $best
            done
        fi
    done
done

if [ $ERRORS -ne 0 ] ; then
    echo "*** there were errors!"
    exit 1
fi
