#!/bin/bash

scriptdir=$(dirname $0)

if [ $# -lt 2 ] ; then
    echo "usage: $0 [-p] [-l] <goldfile> <testfile> [<testfile> ...]"
    exit 1
fi

USEPUNC=0
LABELED=0
while [ 1 ] ; do
    if [ $1 = '-p' ] ; then
	USEPUNC=1 ; shift
    elif [ $1 = '-l' ] ; then
	LABELED=1 ; shift
    else
	break
    fi
done



if [ -z "$lang" ] ; then
    lang=$(basename $1 | cut -d. -f1)
fi

val=$1
shift

if ! echo $val | grep -iq 'original$' ; then
    echo "the original (unmapped) gold test file is required"
    exit 1
fi

if [ $LABELED -ne 0 ] ; then
    if [ -z "$DEPMAP" ] ; then
	# default setting, but can also be set via environment
	DEPMAP=$(dirname $val)/dependencies.map
    fi
    if [ ! -e $DEPMAP ] ; then
	echo "no such file: $DEPMAP"
	echo "require dependency mapping for labeled evaluation"
	exit 1
    fi
fi


for tst in $* ; do
    vallines=$(cat $val | wc -l)
    tstlines=$(cat $tst | wc -l)

    valcols=$(head -1 $val | wc -w)
    tstcols=$(head -1 $tst | wc -w)

    if [ $valcols -ne 8 ] ; then
	echo "malformed gold file $val"
	exit 1
    fi

    headcol=
    if [ $tstcols -eq 4 ] ; then
	headcol=11
    elif [ $tstcols -eq 8 ] ; then
	headcol=15
    else
	echo "malformed file $tst"
	exit 1
    fi
    labelcol=$(($headcol + 1))

    if [ $vallines -ne $tstlines ] ; then
	echo "line mismatch between $val and $tst"
	exit 1
    fi

    if [ $USEPUNC -ne 0 ] ; then
	# evaluate all tokens
	total=$(cat $val | awk 'NF>0' | wc -l)
	match=-1000
	if [ $LABELED -eq 0 ] ; then
	    match=$(paste $val $tst \
		| awk 'NF>0 && $7==$'$headcol \
		| wc -l)
	else
	    match=$(paste $val $tst \
		| perl -e 'open(MAP,$ARGV[0]); @dmap = map {@x=split;$x[1]} <MAP>; close(MAP); while(<STDIN>) {@x=split; if(scalar(@x)==0) { print; next; } die "unknown dependency label" if(!defined($dmap[$x[$#x]])); $x[$#x] = $dmap[$x[$#x]]; print "@x\n"}' $DEPMAP \
		| awk 'NF>0 && $7==$'$headcol' && $8==$'$labelcol \
		| wc -l)
	fi
    else
	if [ $lang = eng ] ; then
            # english Penn Treebank punctuation tags are:
            # , : `` '' .
	    # note that -LRB-/-RRB-/$/#/... are "not punctuation"
	    # this matches with Yamada and Matsumoto (2003) and
	    # McDonald (2005; 2006; thesis)  (the token count on Section
	    # 23 is 49,892 using this punctuation filter)
	    total=$(cat $val \
		| awk 'NF>0' \
		| awk '$5!=":"&&$5!=","&&$5!="``"&&$5!="'\'\''"&&$5!="."' \
		| wc -l)
	    match=-1000
	    if [ $LABELED -eq 0 ] ; then
		match=$(paste $val $tst \
		    | awk 'NF>0 && $7==$'$headcol \
		    | awk '$5!=":"&&$5!=","&&$5!="``"&&$5!="'\'\''"&&$5!="."' \
		    | wc -l)
	    else
		match=$(paste $val $tst \
		    | perl -e 'open(MAP,$ARGV[0]); @dmap = map {@x=split;$x[1]} <MAP>; close(MAP); while(<STDIN>) {@x=split; if(scalar(@x)==0) { print; next; } die "unknown dependency label" if(!defined($dmap[$x[$#x]])); $x[$#x] = $dmap[$x[$#x]]; print "@x\n"}' $DEPMAP \
		    | awk 'NF>0 && $7==$'$headcol' && $8==$'$labelcol \
		    | awk '$5!=":"&&$5!=","&&$5!="``"&&$5!="'\'\''"&&$5!="."' \
		    | wc -l)
	    fi
	else
	    total=$(cat $val | awk 'NF>0' | wc -l)
	    match=-1000
	    if [ $LABELED -eq 0 ] ; then
		match=$(paste $val $tst \
		    | awk 'NF>0 && $7==$'$headcol \
		    | wc -l)
	    else
		match=$(paste $val $tst \
		    | perl -e 'open(MAP,$ARGV[0]); @dmap = map {@x=split;$x[1]} <MAP>; close(MAP); while(<STDIN>) {@x=split; if(scalar(@x)==0) { print; next; } die "unknown dependency label" if(!defined($dmap[$x[$#x]])); $x[$#x] = $dmap[$x[$#x]]; print "@x\n"}' $DEPMAP \
		    | awk 'NF>0 && $7==$'$headcol' && $8==$'$labelcol \
		    | wc -l)
	    fi
	fi
    fi

    echo -n $tst' '
    perl -e 'printf("%.16f\n", 100*$ARGV[0]/$ARGV[1])' $match $total
#    echo "($match / $total)"
done
