#!/usr/bin/perl -w

if(scalar(@ARGV) != 2) {
    print STDERR "usage: $0 <num-slices> <trdata>\n";
    exit 1;
}

my $n = shift;
my $trdata = shift;

my $GZ = ($trdata =~ m/\.gz$/io);

my $digits = length(sprintf("%d", $n));

open(TR, ($GZ ? "gunzip -c $trdata |" : "< $trdata"))
    or die "can't open trdata \"$trdata\"";
my $Ntr = 0;
my @split_sizes = map {0} (1 .. $n);
my @examples = ();
my $ex = "";
while(<TR>) {
    $ex .= $_;
    if(m/^\s*$/o) { # blank line; finished example
        push(@examples, $ex); $ex = "";
        ++($split_sizes[$Ntr % $n]);
        ++$Ntr;
    }
}
close(TR);



my @split_starts = map {0} (1 .. $n);
foreach my $i (1 .. ($n - 1)) {
    $split_starts[$i] = $split_starts[$i - 1] + $split_sizes[$i - 1];
}
die "???" if(($split_starts[$n - 1] + $split_sizes[$n - 1]) != $Ntr);

foreach my $i (0 .. ($n - 1)) {
    print STDERR "slice ".($i + 1).": ".$split_sizes[$i].
        " [".$split_starts[$i]."-".
        ($split_starts[$i] + $split_sizes[$i] - 1)."]\n";
}
print STDERR "total: $Ntr\n\n";



foreach my $i (0 .. ($n - 1)) {
    my $padnum = sprintf('%0'.$digits.'d', ($i + 1));
    my $tag = "xv$n-$padnum";
    my $slicetr = $trdata;
    $slicetr =~ s/\.train\./.$tag.train./;
    my $sliceval = $trdata;
    $sliceval =~ s/\.train\./.$tag.held./;
    print STDERR "mkslice ".($i + 1).": $slicetr\n";
    print STDERR "mkslice ".($i + 1).": $sliceval\n";
    open(SV, ($GZ ? "| gzip -c > $sliceval" : "> $sliceval"))
        or die "couldn't open slice-val \"$sliceval\"";
    open(ST, ($GZ ? "| gzip -c > $slicetr"  : "> $slicetr"))
        or die "couldn't open slice-tr \"$slicetr\"";
    foreach my $j (0 .. ($n - 1)) {
        my $beg = $split_starts[$j];
        my $end = $beg + $split_sizes[$j] - 1;
        foreach my $k ($beg .. $end) {
            my $ex = $examples[$k];
            if($j == $i) { print SV $ex; }
            else         { print ST $ex; }
        }
    }
    close(SV);
    close(ST);
}
